package libmath;

/**
 * Java version of MINPACK
 *
 * See: http://devernay.free.fr/hacks/cminpack/index.html
 *
 * Equivalence table between MINPACK and NAG, NPL, SLATEC, levmar and GSL:
 * MINPACK: hybrd, NAG: C05NCF, SLATEC: DNSQ, GSL: gsl_multiroot_fsolver_hybrids.
 * MINPACK: hybrd1, NAG: C05NBF, SLATEC: DNSQE
 */
public final class JMinPack {

	/** Interface with the function to compute */
	public interface IFunction {

		/**
		 * The function to compute
		 *
		 * fcn is the name of the user-supplied subroutine which
		 * calculates the functions. fcn must be declared
		 * in an external statement in the user calling
		 * program, and should be written as follows.
		 *
		 * Calculate the functions at x and return this vector in fvec.
		 *
		 * @param x input values
		 * @param fvec results of computations.
		 * @return flag. < 0 means "terminate"
		 */
		int fcn_nn(double[] x, double[] fvec, int iflag);
	}

	/**
	 * Subroutine qform
	 *
	 * This subroutine proceeds from the computed qr factorization of
	 * an m by n matrix a to accumulate the m by m orthogonal matrix
	 * q from its factored form.
	 *
	 * Argonne national laboratory. MINPACK project. march 1980.
	 * Burton S. garbow, Kenneth E. Hillstrom, Jorge J. More
	 *
	 * where
	 *
	 * @param m is a positive integer input variable set to the number
	 *     of rows of a and the order of q.
	 *
	 * @param n is a positive integer input variable set to the number
	 *     of columns of a.
	 *
	 * @param q is an m by m array. on input the full lower trapezoid in
	 *     the first min(m,n) columns of q contains the factored form.
	 *     on output q has been accumulated into a square matrix.
	 *
	 * @param ldq is a positive integer input variable not less than m
	 *     which specifies the leading dimension of the array q.
	 *
	 * @param wa is a work array of length m.
	 */
	private static void qform(final int m, final int n, final double[] q, final int ldq, final double[] wa) {

		// zero out upper triangle of q in the first min(m,n) columns.

		final int minmn = Math.min( m, n );
		if( minmn >= 2 ) {
			for( int j = 1, jm1 = ldq; j < minmn; ++j, jm1 += ldq ) {
				for( int i = jm1, end = i + j; i < end; ++i ) {
					q[ i ] = 0.;
				}
			}
		}

		// initialize remaining columns to those of the identity matrix.

		if( m > n ) {
			for( int j = n, k = j * ldq; j < m; ++j, k += ldq ) {
				for( int i = k, end = i + m; i < end; ++i ) {
					q[ i ] = 0.;
				}
				q[ j + k ] = 1.;
			}
		}

		// accumulate q from its factored form.

		for( int k = minmn - 1, k_ldq = k * ldq; k >= 0; --k, k_ldq -= ldq ) {
			for( int i = k, qi = i + k_ldq; i < m; ++i, ++qi ) {
				wa[ i ] = q[ qi ];
				q[ qi ] = 0.;
			}
			q[ k + k_ldq ] = 1.;
			if( wa[ k ] != 0. ) {
				for( int j = k * ldq, me = m * ldq; j < me; j += ldq ) {
					double sum = 0.;
					for( int i = k, qi = i + j; i < m; ++i, ++qi ) {
						sum += q[ qi ] * wa[ i ];
					}
					final double temp = sum / wa[ k ];
					for( int i = k, qi = i + j; i < m; ++i, ++qi ) {
						q[ qi ] -= temp * wa[ i ];
					}
				}
			}
		}
	}

	/**
	 * Subroutine r1mpyq
	 *
	 * Given an m by n matrix a, this subroutine computes a*q where
	 * q is the product of 2*(n - 1) transformations
	 * <pre>
	 *       gv(n-1)*...*gv(1)*gw(1)*...*gw(n-1)
	 * </pre>
	 * and gv(i), gw(i) are givens rotations in the (i,n) plane which
	 * eliminate elements in the i-th and n-th planes, respectively.
	 * q itself is not given, rather the information to recover the
	 * gv, gw rotations is supplied.
	 *
	 * Argonne national laboratory. MINPACK project. March 1980.
	 * Burton S. Garbow, Kenneth E. Hillstrom, Jorge J. More
	 *
	 * @param m is a positive integer input variable set to the number
	 *     of rows of a.
	 *
	 * @param n is a positive integer input variable set to the number
	 *     of columns of a.
	 *
	 * @param a is an m by n array. on input a must contain the matrix
	 *     to be postmultiplied by the orthogonal matrix q
	 *     described above. on output a*q has replaced a.
	 *
	 * @param lda is a positive integer input variable not less than m
	 *     which specifies the leading dimension of the array a.
	 *
	 * @param v is an input array of length n. v(i) must contain the
	 *     information necessary to recover the givens rotation gv(i)
	 *     described above.
	 *
	 * @param w is an input array of length n. w(i) must contain the
	 *     information necessary to recover the givens rotation gw(i)
	 *     described above.
	 */
	private static void r1mpyq(final int m, final int n, final double[] a, final int lda, final double[] v, final double[] w) {

		// Local variables
		double cos, sin;

		if( n < 2 ) {
			return;
		}

		// apply the first set of givens rotations to a.

		final int nm1 = n - 1;

		final int nm1_lda = nm1 * lda;
		for( int j = nm1 - 1, j_lda = j * lda; j >= 0; --j, j_lda -= lda ) {
			if( Math.abs( v[ j ] ) > 1. ) {
				cos = 1. / v[ j ];
				sin = Math.sqrt( 1. - cos * cos );
			} else {
				sin = v[ j ];
				cos = Math.sqrt( 1. - sin * sin );
			}
			for( int i_j_lda = j_lda, i_nm1_lda = nm1_lda, me = m + j_lda; i_j_lda < me; ++i_j_lda, ++i_nm1_lda ) {
				final double temp = cos * a[ i_j_lda ] - sin * a[ i_nm1_lda ];
				a[ i_nm1_lda ] = sin * a[ i_j_lda ] + cos * a[ i_nm1_lda ];
				a[ i_j_lda ] = temp;
			}
		}

		// apply the second set of givens rotations to a.

		for( int j = 0, j_lda = 0; j < nm1; ++j, j_lda += lda ) {
			if( Math.abs( w[ j ] ) > 1. ) {
				cos = 1 / w[ j ];
				sin = Math.sqrt( 1. - cos * cos );
			} else {
				sin = w[ j ];
				cos = Math.sqrt( 1. - sin * sin );
			}
			for( int i_j_lda = j_lda, i_nm1_lda = nm1_lda, me = m + j_lda; i_j_lda < me; ++i_j_lda, ++i_nm1_lda ) {
				final double temp = cos * a[ i_j_lda ] + sin * a[ i_nm1_lda ];
				a[ i_nm1_lda ] = cos * a[ i_nm1_lda ] - sin * a[ i_j_lda ];
				a[ i_j_lda ] = temp;
			}
		}
	}

/*
  About the values for rdwarf and rgiant.

  The original values, both in single-precision FORTRAN source code and in double-precision code were:
#define rdwarf 3.834e-20
#define rgiant 1.304e19
  See for example:
    http://www.netlib.org/slatec/src/denorm.f
    http://www.netlib.org/slatec/src/enorm.f
  However, rdwarf is smaller than sqrt(FLT_MIN) = 1.0842021724855044e-19, so that rdwarf**2 will
  underflow. This contradicts the constraints expressed in the comments below.

  We changed these constants to those proposed by the
  implementation found in MPFIT http://cow.physics.wisc.edu/~craigm/idl/fitting.html

 cmpfit-1.2 proposes the following definitions:
  rdwarf = sqrt(dpmpar(2)*1.5) * 10
  rgiant = sqrt(dpmpar(3)) * 0.1

 The half version does not really worked that way, so we use for half:
  rdwarf = sqrt(dpmpar(2)) * 2
  rgiant = sqrt(dpmpar(3)) * 0.5
 Any suggestion is welcome. Half CMINPACK is really only a
 proof-of-concept anyway.

 See the example/tenorm*c, which computes these values
*/
	private static final double DOUBLE_DWARF = (1.82691291192569e-153);
	private static final double DOUBLE_GIANT = (1.34078079299426e+153);

	// private static final double LONG_DOUBLE_DWARF = (2.245696932951581572e-2465l);
	// private static final double LONG__DOUBLE_GIANT = (1.090748135619415929e+2465l);

	// private static final double FLOAT_DWARF = (1.327871072777421e-18f);
	// private static final double FLOAT_GAIN = (1.844674297419792e+18f);


	/**
	 * Function enorm
	 *
	 * Fiven an n-vector x, this function calculates the
	 * euclidean norm of x.
	 *
	 * The euclidean norm is computed by accumulating the sum of
	 * squares in three different sums. The sums of squares for the
	 * small and large components are scaled so that no overflows
	 * occur. non-destructive underflows are permitted. Underflows
	 * and overflows do not occur in the computation of the unscaled
	 * sum of squares for the intermediate components.
	 * The definitions of small, intermediate and large components
	 * depend on two constants, rdwarf and rgiant. The main
	 * restrictions on these constants are that rdwarf^2 not
	 * underflow and rgiant^2 not overflow. the constants
	 * given here are suitable for every known computer.
	 *
	 * Argonne national laboratory. MINPACK project. march 1980.
	 * Burton S. Garbow, Kenneth E. Hillstrom, Jorge J. More
	 *
	 * @param n is a positive integer input variable.
	 * @param x is an input array of length n.
	 * @param xOff an offset in the x-array.
	 */
	private static double enorm(int n, final double[] x, int xOff) {
		double s1 = 0.;
		double s2 = 0.;
		double s3 = 0.;
		double x1max = 0.;
		double x3max = 0.;
		final double agiant = DOUBLE_GIANT / (double)n;
		n += xOff;
		for( ; xOff < n; ++xOff ) {
			final double xabs = Math.abs( x[ xOff ] );
			if( xabs >= agiant ) {
				// sum for large components.
				if( xabs > x1max ) {
					// Computing 2nd power
					final double d1 = x1max / xabs;
					s1 = 1. + s1 * (d1 * d1);
					x1max = xabs;
				} else {
					// Computing 2nd power
					final double d1 = xabs / x1max;
					s1 += d1 * d1;
				}
			} else if( xabs <= DOUBLE_DWARF ) {
				// sum for small components.
				if( xabs > x3max ) {
					// Computing 2nd power
					final double d1 = x3max / xabs;
					s3 = 1. + s3 * (d1 * d1);
					x3max = xabs;
				} else if( xabs != 0. ) {
					// Computing 2nd power
					final double d1 = xabs / x3max;
					s3 += d1 * d1;
				}
			} else {
				// sum for intermediate components.
				// Computing 2nd power
				s2 += xabs * xabs;
			}
		}

		// calculation of norm.

		if( s1 != 0. ) {
			return x1max * Math.sqrt( s1 + (s2 / x1max) / x1max );
		}
		if( s2 != 0. ) {
			if( s2 >= x3max ) {
				return Math.sqrt( s2 * (1. + (x3max / s2) * (x3max * s3)) );
			}
			return Math.sqrt( x3max * ((s2 / x3max) + (x3max * s3)) );
		}
		return x3max * Math.sqrt( s3 );
	}

	/** b^(1 - t), DPMPAR( realm, EPSILON(1) );// 2^-52 = 2.2204460492503131e-16 | 2^-23 = 1.19209290e-07F | long double: 2−63 ≈ 1.08e-19 */
	private static final double MACHINE_PRECISION = 2.2204460492503131e-16;

	/* b^(emin - 1), DPMPAR( realm, MIN(2) );// 2.2250738585072014e-308 | 1.17549435e-38F */
	// private static final double SMALEST_MAGNITUDE = 2.2250738585072014e-308;

	/** b^emax*(1 - b^(-t)), DPMPAR( realm, MAX(3) );// 1.7976931348623157e+308 | 3.40282347e+38F */
	private static final double LAGEST_MAGNITUDE = 1.7976931348623157e+308;

	/** 0.05 */
	private static final double P05 = .05;

	/**
	 * Subroutine qrfac
	 *
	 * This subroutine uses householder transformations with column
	 * pivoting (optional) to compute a qr factorization of the
	 * m by n matrix a. That is, qrfac determines an orthogonal
	 * matrix q, a permutation matrix p, and an upper trapezoidal
	 * matrix r with diagonal elements of nonincreasing magnitude,
	 * such that a*p = q*r. The householder transformation for
	 * column k, k = 1,2,...,min(m,n), is of the form
	 * <pre>
	 *                   t
	 *    i - (1/u(k))*u*u
	 * </pre>
	 * where u has zeros in the first k-1 positions. The form of
	 * this transformation and the method of pivoting first
	 * appeared in the corresponding LINPACK subroutine.
	 *
	 * Argonne national laboratory. MINPACK project. march 1980.
	 * Burton S. Garbow, Kenneth E. Hillstrom, Jorge J. More
	 *
	 * @param m is a positive integer input variable set to the number
	 *  of rows of a.
	 *
	 * @param n is a positive integer input variable set to the number
	 *  of columns of a.
	 *
	 * @param a is an m by n array. on input a contains the matrix for
	 *  which the qr factorization is to be computed. On output
	 *  the strict upper trapezoidal part of a contains the strict
	 *  upper trapezoidal part of r, and the lower trapezoidal
	 *  part of a contains a factored form of q (the non-trivial
	 *  elements of the u vectors described above).
	 *
	 * @param lda is a positive integer input variable not less than m
	 *  which specifies the leading dimension of the array a.
	 *
	 *   pivot is a logical input variable. If pivot is set true,
	 *  then column pivoting is enforced. If pivot is set false,
	 *  then no column pivoting is done.
	 *
	 * @param ipvt is an integer output array of length lipvt. ipvt
	 *  defines the permutation matrix p such that a*p = q*r.
	 *  column j of p is column ipvt(j) of the identity matrix.
	 *  If pivot is false, ipvt is not referenced.
	 *
	 * @param rdiag is an output array of length n which contains the
	 *  diagonal elements of r.
	 *
	 * @param acnorm is an output array of length n which contains the
	 *  norms of the corresponding columns of the input matrix a.
	 *  If this information is not needed, then acnorm can coincide
	 *  with rdiag.
	 *
	 * @param wa is a work array of length n. If pivot is false, then wa
	 *  can coincide with rdiag.
	 */
	private static void qrfac(final int m, final int n, final double[] a,
			final int lda, final boolean pivot, final int[] ipvt,// final int lipvt,
			final double[] rdiag, final double[] acnorm, final double[] wa)
	{
		// compute the initial column norms and initialize several arrays.

		for( int j = 0, k = 0; j < n; ++j, k += lda ) {
			acnorm[ j ] = enorm( m, a, k );
			rdiag[ j ] = acnorm[ j ];
			wa[ j ] = rdiag[ j ];
			if( pivot ) {
				ipvt[ j ] = j + 1;
			}
		}

		// reduce a to r with householder transformations.

		final int minmn = Math.min( m, n );
		for( int j = 0, j_lda = 0; j < minmn; ++j, j_lda += lda ) {

			if( pivot ) {

				// bring the column of largest norm into the pivot position.

				int kmax = j;
				for( int k = j; k < n; ++k ) {
					if( rdiag[ k ] > rdiag[ kmax ] ) {
						kmax = k;
					}
				}
				if( kmax != j ) {
					final int kmax_lda = kmax * lda;
					for( int i_j_lda = j_lda, i_kmax_lda = kmax_lda, me = m + j_lda; j_lda < me; ++i_j_lda, ++i_kmax_lda ) {
						final double temp = a[ i_j_lda ];
						a[ i_j_lda ] = a[ i_kmax_lda ];
						a[ i_kmax_lda ] = temp;
					}
					rdiag[ kmax ] = rdiag[ j ];
					wa[ kmax ] = wa[ j ];
					final int k = ipvt[ j ];
					ipvt[ j ] = ipvt[ kmax ];
					ipvt[ kmax ] = k;
				}
			}

			// compute the householder transformation to reduce the
			// j-th column of a to a multiple of the j-th unit vector.

			final int j_j_lda = j + j_lda;
			double ajnorm = enorm( m - j, a, j_j_lda );
			if( ajnorm != 0. ) {
				if( a[ j_j_lda ] < 0. ) {
					ajnorm = -ajnorm;
				}
				for( int i = j + j_lda, me = m + j_lda; i < me; ++i ) {
					a[ i ] /= ajnorm;
				}
				a[ j_j_lda ] += 1.;

				// apply the transformation to the remaining columns
				// and update the norms.

				final int jp1 = j + 1;
				if( n > jp1 ) {
					for( int k = jp1, k_lda = k * lda; k < n; ++k, k_lda += lda ) {
						double sum = 0.;
						for( int i = j; i < m; ++i ) {
							sum += a[ i + j_lda ] * a[ i + k_lda ];
						}
						double temp = sum / a[ j_j_lda ];
						for( int i1 = j + j_lda, me = m + j_lda, i2 = j + k_lda; i1 < me; ++i1, ++i2 ) {
							a[ i2 ] -= temp * a[ i1 ];
						}
						if( pivot && rdiag[ k ] != 0. ) {
							temp = a[ j + k_lda ] / rdiag[ k ];
							// Computing MAX
							double d1 = 1. - temp * temp;
							rdiag[ k ] *= Math.sqrt( Math.max( 0., d1 ) );
							// Computing 2nd power */
							d1 = rdiag[ k ] / wa[ k ];
							if( P05 * (d1 * d1) <= MACHINE_PRECISION ) {
								rdiag[ k ] = enorm( m - jp1, a, jp1 + k_lda );
								wa[ k ] = rdiag[ k ];
							}
						}
					}
				}
			}
			rdiag[ j ] = -ajnorm;
		}
	}

	/** 0.5 */
	private static final double P5 = .5;

	/** 0.25 */
	private static final double P25 = .25;

	/**
	 * Subroutine r1updt
	 *
	 * Given an m by n lower trapezoidal matrix s, an m-vector u,
	 * and an n-vector v, the problem is to determine an
	 * orthogonal matrix q such that
	 * <pre>
	 *               t
	 *       (s + u*v )*q
	 * </pre>
	 * is again lower trapezoidal.
	 *
	 * This subroutine determines q as the product of 2*(n - 1)
	 * transformations
	 * <pre>
	 *       gv(n-1)*...*gv(1)*gw(1)*...*gw(n-1)
	 * </pre>
	 * where gv(i), gw(i) are givens rotations in the (i,n) plane
	 * which eliminate elements in the i-th and n-th planes,
	 * respectively. q itself is not accumulated, rather the
	 * information to recover the gv, gw rotations is returned.
	 *
	 * Argonne national laboratory. MINPACK project. march 1980.
	 * Burton S. garbow, Kenneth E. Hillstrom, Jorge J. More,
	 * John L. Nazareth
	 *
	 * @param m is a positive integer input variable set to the number
	 *     of rows of s.
	 *
	 * @param n is a positive integer input variable set to the number
	 *     of columns of s. n must not exceed m.
	 *
	 * @param s is an array of length ls. On input s must contain the lower
	 *     trapezoidal matrix s stored by columns. On output s contains
	 *     the lower trapezoidal matrix produced as described above.
	 *
	 * @param ls is a positive integer input variable not less than
	 *     (n*(2*m-n+1))/2.
	 *
	 * @param u is an input array of length m which must contain the
	 *     vector u.
	 *
	 * @param v is an array of length n. On input v must contain the vector
	 *     v. On output v(i) contains the information necessary to
	 *     recover the givens rotation gv(i) described above.
	 *
	 * @param w is an output array of length m. w(i) contains information
	 *     necessary to recover the givens rotation gw(i) described
	 *     above.
	 *
	 * @return sign is a logical output variable. Sign is set true if any
	 *     of the diagonal elements of the output s are zero. Otherwise
	 *     sign is set false.
	 */
	private static boolean r1updt(final int m, final int n,
			final double[] s,
			final double[] u, final double[] v, final double[] w)
	{
		// giant is the largest magnitude.
		// final double giant = dpmpar( 3 );

		// initialize the diagonal element pointer.

		int jj = n * ((m << 1) - n + 1) / 2 - (m - n) - 1;

		// move the nontrivial part of the last column of s into w.

		final int nm1 = n - 1;
		int k = jj;
		for( int i = nm1; i < m; ++i ) {
			w[ i ] = s[ k ];
			++k;
		}

		// rotate the vector v into a multiple of the n-th unit vector
		// in such a way that a spike is introduced into w.

		if( nm1 >= 1 ) {
			for( int nmj = 0; nmj < nm1; ++nmj ) {
				final int j = nm1 - nmj - 1;
				jj -= m - j;
				w[ j ] = 0.;
				if( v[ j ] != 0. ) {

					// determine a givens rotation which eliminates the
					// j-th element of v.
					double sin, cos, tau;
					if( Math.abs( v[ nm1 ] ) < Math.abs( v[ j ] ) ) {
						final double cotan = v[ nm1 ] / v[ j ];
						sin = P5 / Math.sqrt( P25 + P25 * (cotan * cotan) );
						cos = sin * cotan;
						tau = 1.;
						if( Math.abs( cos ) * LAGEST_MAGNITUDE > 1. ) {
							tau = 1. / cos;
						}
					} else {
						final double tan = v[ j ] / v[ nm1 ];
						cos = P5 / Math.sqrt( P25 + P25 * (tan * tan) );
						sin = cos * tan;
						tau = sin;
					}

					// apply the transformation to v and store the information
					// necessary to recover the givens rotation.

					v[ nm1 ] = sin * v[ j ] + cos * v[ nm1 ];
					v[ j ] = tau;

					// apply the transformation to s and extend the spike in w.

					k = jj;
					for( int i = j; i < m; ++i ) {
						final double temp = cos * s[ k ] - sin * w[ i ];
						w[ i ] = sin * s[ k ] + cos * w[ i ];
						s[ k ] = temp;
						++k;
					}
				}
			}
		}

		// add the spike from the rank 1 update to w.

		for( int i = 0; i < m; ++i ) {
			w[ i ] += v[ nm1 ] * u[ i ];
		}

		// eliminate the spike.

		boolean sign = false;
		if( nm1 >= 1 ) {
			for( int j = 0; j < nm1; ++j ) {
				if( w[ j ] != 0. ) {

					// determine a givens rotation which eliminates the
					// j-th element of the spike.
					double sin, cos, tau;
					if( Math.abs( s[ jj ] ) < Math.abs( w[ j ] ) ) {
						final double cotan = s[ jj ] / w[ j ];
						sin = P5 / Math.sqrt( P25 + P25 * (cotan * cotan) );
						cos = sin * cotan;
						tau = 1.;
						if( Math.abs( cos ) > 1. / LAGEST_MAGNITUDE ) {
							tau = 1. / cos;
						}
					} else {
						final double tan = w[ j ] / s[ jj ];
						cos = P5 / Math.sqrt( P25 + P25 * (tan * tan) );
						sin = cos * tan;
						tau = sin;
					}

					// apply the transformation to s and reduce the spike in w.

					k = jj;
					for( int i = j; i < m; ++i ) {
						final double temp = cos * s[ k ] + sin * w[ i ];
						w[ i ] = cos * w[ i ] - sin * s[ k ];
						s[ k ] = temp;
						++k;
					}

					// store the information necessary to recover the
					// givens rotation.

					w[ j ] = tau;
				}

				// test for zero diagonal elements in the output s.

				if( s[ jj ] == 0. ) {
					sign = true;
				}
				jj += m - j;
			}
		}

		// move w back into the last column of the output s.

		k = jj;
		for( int i = nm1; i < m; ++i ) {
			s[ k ] = w[ i ];
			++k;
		}
		if( s[ jj ] == 0. ) {
			sign = true;
		}
		return sign;
	}

	/**
	 * Subroutine fdjac1
	 *
	 * This subroutine computes a forward-difference approximation
	 * to the n by n jacobian matrix associated with a specified
	 * problem of n functions in n variables. If the jacobian has
	 * a banded form, then function evaluations are saved by only
	 * approximating the nonzero terms.
	 *
	 * Argonne national laboratory. MINPACK project. March 1980.
	 * Burton S. Garbow, Kenneth E. Hillstrom, Jorge J. More
	 *
	 * @param fcn is the name of the user-supplied subroutine which
	 *     calculates the functions. fcn must be declared
	 *     in an external statement in the user calling
	 *     program, and should be written as follows.
	 * <pre>
	 *     int fcn(double[] x,double[] fvec,int iflag) {
	 *     ----------
	 *     calculate the functions at x and
	 *     return this vector in fvec and iflag
	 *     ----------
	 *     }
	 * </pre>
	 *     The value of iflag should not be changed by fcn unless
	 *     the user wants to terminate execution of fdjac1.
	 *     In this case set iflag to a negative integer.
	 *
	 * @param n is a positive integer input variable set to the number
	 *     of functions and variables.
	 *
	 * @param x is an input array of length n.
	 *
	 * @param fvec is an input array of length n which must contain the
	 *     functions evaluated at x.
	 *
	 * @param fjac is an output n by n array which contains the
	 *     approximation to the jacobian matrix evaluated at x.
	 *
	 * @param ldfjac is a positive integer input variable not less than n
	 *     which specifies the leading dimension of the array fjac.
	 *
	 * @param iflag is an integer variable which can be used to terminate
	 *     the execution of fdjac1. See description of fcn.
	 *
	 * @param ml is a nonnegative integer input variable which specifies
	 *     the number of subdiagonals within the band of the
	 *     jacobian matrix. If the jacobian is not banded, set
	 *     ml to at least n - 1.
	 *
	 * @param epsfcn is an input variable used in determining a suitable
	 *     step length for the forward-difference approximation. This
	 *     approximation assumes that the relative errors in the
	 *     functions are of the order of epsfcn. If epsfcn is less
	 *     than the machine precision, it is assumed that the relative
	 *     errors in the functions are of the order of the machine
	 *     precision.
	 *
	 * @param mu is a nonnegative integer input variable which specifies
	 *     the number of superdiagonals within the band of the
	 *     jacobian matrix. If the jacobian is not banded, set
	 *     mu to at least n - 1.
	 *
	 * @param wa1 and wa2 are work arrays of length n. If ml + mu + 1 is at
	 *     least n, then the jacobian is considered dense, and wa2 is
	 *     not referenced.
	 */
	private static int fdjac1(final IFunction fun, final int n,
			final double[] x, final double[] fvec, final double[] fjac,
			final int ldfjac, final int ml, final int mu, final double epsfcn,
			final double[] wa1, final double[] wa2)
	{
		final double eps = Math.sqrt( Math.max( epsfcn, MACHINE_PRECISION ) );
		final int msum = ml + mu + 1;
		if( msum >= n ) {

			// computation of dense approximate jacobian.

			for( int j = 0; j < n; ++j ) {
				final double temp = x[ j ];
				double h = eps * Math.abs( temp );
				if( h == 0. ) {
					h = eps;
				}
				x[ j ] = temp + h;
				// the last parameter of fcn_nn() is set to 2 to differentiate
				// calls made to compute the function from calls made to compute
				// the Jacobian (see fcn() in examples/hybdrv.c, and how njev
				// is used to compute the number of Jacobian evaluations)
				final int iflag = fun.fcn_nn( x, wa1, 2 );
				if( iflag < 0 ) {
					return iflag;
				}
				x[ j ] = temp;
				for( int i = 0, k = j * ldfjac; i < n; ++i, ++k ) {
					fjac[ k ] = (wa1[ i ] - fvec[ i ]) / h;
				}
			}
			return 0;
		}

		// computation of banded approximate jacobian.

		for( int k = 0; k < msum; ++k ) {
			for( int j = k; j < n; j += msum ) {
				wa2[ j ] = x[ j ];
				double h = eps * Math.abs( wa2[ j ] );
				if( h == 0. ) {
					h = eps;
				}
				x[ j ] = wa2[ j ] + h;
			}
			final int iflag = fun.fcn_nn( x, wa1, 1 );
			if( iflag < 0 ) {
				return iflag;
			}
			for( int j = k; j < n; j += msum ) {
				x[ j ] = wa2[ j ];
				double h = eps * Math.abs( wa2[ j ] );
				if( h == 0. ) {
					h = eps;
				}
				for( int i = 0, m = j * ldfjac; i < n; ++i, ++m ) {
					fjac[ m ] = 0.;
					if( (i >= j - mu) && (i <= j + ml) ) {
						fjac[ m ] = (wa1[ i ] - fvec[ i ]) / h;
					}
				}
			}
		}
		return 0;
	}

	/**
	 * Subroutine dogleg
	 *
	 * Given an m by n matrix a, an n by n nonsingular diagonal
	 * matrix d, an m-vector b, and a positive number delta, the
	 * problem is to determine the convex combination x of the
	 * gauss-newton and scaled gradient directions that minimizes
	 * (a*x - b) in the least squares sense, subject to the
	 * restriction that the euclidean norm of d*x be at most delta.
	 *
	 * This subroutine completes the solution of the problem
	 * if it is provided with the necessary information from the
	 * qr factorization of a. That is, if a = q*r, where q has
	 * orthogonal columns and r is an upper triangular matrix,
	 * then dogleg expects the full upper triangle of r and
	 * the first n components of (q transpose)*b.
	 *
	 * Argonne national laboratory. MINPACK project. march 1980.
	 * Burton D. Garbow, Kenneth E. Hillstrom, Jorge J. More
	 *
	 * @param n is a positive integer input variable set to the order of r.
	 *
	 * @param r is an input array of length lr which must contain the upper
	 *     triangular matrix r stored by rows.
	 *
	 * @param lr is a positive integer input variable not less than
	 *     (n*(n+1))/2.
	 *
	 *   diag is an input array of length n which must contain the
	 *     diagonal elements of the matrix d.
	 *
	 * @param qtb is an input array of length n which must contain the first
	 *     n elements of the vector (q transpose)*b.
	 *
	 * @param delta is a positive input variable which specifies an upper
	 *     bound on the euclidean norm of d*x.
	 *
	 * @param x is an output array of length n which contains the desired
	 *     convex combination of the gauss-newton direction and the
	 *     scaled gradient direction.
	 *
	 * @param wa1 work array of length n.
	 * @param wa2 work array of length n.
	 */
	private static void dogleg(final int n, final double[] r,
			final double[] diag, final double[] qtb, final double delta,
			final double[] x, final double[] wa1, final double[] wa2)
	{
		// first, calculate the gauss-newton direction.

		int jj = (n * (n + 1)) >>> 1;
		for( int k = 1; k <= n; ++k ) {
			final int j = n - k;
			final int jp1 = j + 1;
			jj -= k;
			double sum = 0.;
			if( n > jp1 ) {
				int m = jj;
				for( int i = jp1; i < n; ++i ) {
					++m;
					sum += r[ m ] * x[ i ];
				}
			}
			double temp = r[ jj ];
			if( temp == 0. ) {
				int m = j;
				for( int i = 0; i <= j; ++i ) {
					/* Computing MAX */
					final double d2 = Math.abs( r[ m ] );
					temp = Math.max( temp, d2 );
					m += n - (i + 1);
				}
				temp *= MACHINE_PRECISION;
				if( temp == 0. ) {
					temp = MACHINE_PRECISION;
				}
			}
			x[ j ] = (qtb[ j ] - sum) / temp;
		}

		// test whether the gauss-newton direction is acceptable.

		for( int j = 0; j < n; ++j ) {
			wa1[ j ] = 0.;
			wa2[ j ] = diag[ j ] * x[ j ];
		}
		final double qnorm = enorm( n, wa2, 0 );
		if( qnorm <= delta ) {
			return;
		}

		// the gauss-newton direction is not acceptable.
		// next, calculate the scaled gradient direction.

		int m = 0;
		for( int j = 0; j < n; ++j ) {
			final double temp = qtb[ j ];
			for( int i = j; i < n; ++i ) {
				wa1[ i ] += r[ m ] * temp;
				++m;
			}
			wa1[ j ] /= diag[ j ];
		}

		// calculate the norm of the scaled gradient and test for
		// the special case in which the scaled gradient is zero.

		final double gnorm = enorm( n, wa1, 0 );
		double sgnorm = 0.;
		double alpha = delta / qnorm;
		if( gnorm != 0. ) {

			// calculate the point along the scaled gradient
			// at which the quadratic is minimized.

			for( int j = 0; j < n; ++j ) {
				wa1[ j ] = wa1[ j ] / gnorm / diag[ j ];
			}
			m = 0;
			for( int j = 0; j < n; ++j ) {
				double sum = 0.;
				for( int i = j; i < n; ++i ) {
					sum += r[ m ] * wa1[ i ];
					++m;
				}
				wa2[ j ] = sum;
			}
			double temp = enorm( n, wa2, 0 );
			sgnorm = gnorm / temp / temp;

			// test whether the scaled gradient direction is acceptable.

			alpha = 0.;
			if( sgnorm < delta ) {

				// the scaled gradient direction is not acceptable.
				// finally, calculate the point along the dogleg
				// at which the quadratic is minimized.

				final double bnorm = enorm( n, qtb, 0 );
				temp = bnorm / gnorm * (bnorm / qnorm) * (sgnorm / delta);
				/* Computing 2nd power */
				double d1 = sgnorm / delta;
				/* Computing 2nd power */
				final double d2 = temp - delta / qnorm;
				/* Computing 2nd power */
				final double d3 = delta / qnorm;
				/* Computing 2nd power */
				final double d4 = sgnorm / delta;
				temp = temp - delta / qnorm * (d1 * d1) + Math.sqrt( d2 * d2 + (1. - d3 * d3) * (1. - d4 * d4) );
				/* Computing 2nd power */
				d1 = sgnorm / delta;
				alpha = delta / qnorm * (1. - d1 * d1) / temp;
			}
		}

		// form appropriate convex combination of the gauss-newton
		// direction and the scaled gradient direction.

		final double temp = (1. - alpha) * Math.min( sgnorm, delta );
		for( int j = 0; j < n; ++j ) {
			x[ j ] = temp * wa1[ j ] + alpha * x[ j ];
		}
	}

	/** Improper input parameters. */
	public static final int INFO_STATUS_IMPROPER_INPUT_PARAMETERS = 0;

	/** Relative error between two consecutive iterates is at most xtol. */
	public static final int INFO_STATUS_NO_IMPROVEMENTS_BETWEEN_ITERATIONS = -1;

	/** Number of calls to fcn has reached or exceeded maxfev. */
	public static final int INFO_STATUS_MAX_CALLS_REACHED = -2;

	/** xtol is too small. No further improvement in the approximate solution x is possible. */
	public static final int INFO_STATUS_TOLERANCE_TOO_SMALL = -3;

	/** Iteration is not making good progress, as
	 * measured by the improvement from the last
	 * five jacobian evaluations. */
	public static final int INFO_STATUS_NO_PROGRESS_5 = -4;

	/** Iteration is not making good progress, as
	 * measured by the improvement from the last
	 * ten iterations.. */
	public static final int INFO_STATUS_NO_PROGRESS_10 = -5;

	private static int terminate(final IFunction fun, final double[] x, final double[] fvec, final int nprint, final int iflag, final int info) {
		// termination, either normal or user imposed.

		if( nprint > 0 ) {
			fun.fcn_nn( x, fvec, 0 );
		}
		if( iflag < 0 ) {
			return iflag;
		}
		return info;
	}

	/** 0.1 */
	private static final double P1 = .1;

	/** 1e-3 */
	private static final double P001 = .001;

	/** 1e-4 */
	private static final double P0001 = 1e-4;

	/**
	 * Subroutine hybrd
 	 *
	 * The purpose of hybrd is to find a zero of a system of
	 * n nonlinear functions in n variables by a modification
	 * of the Powell hybrid method. The user must provide a
	 * subroutine which calculates the functions. The jacobian is
	 * then calculated by a forward-difference approximation.
	 *
	 * Argonne national laboratory. MINPACK project. March 1980.
	 * Burton S. Garbow, Kenneth E. Hillstrom, Jorge J. More
 	 *
	 * @param func is the name of the user-supplied subroutine which
	 *     calculates the functions. fcn must be declared
	 *     in an external statement in the user calling
	 *     program, and should be written as follows.
	 * <pre>
	 *     int fcn(double[] x,double[] fvec,int iflag) {
	 *     ----------
	 *     calculate the functions at x and
	 *     return this vector in fvec and iflag
	 *     ----------
	 *     }
	 * </pre>
	 *     The value of iflag should not be changed by fcn unless
	 *     the user wants to terminate execution of fdjac1.
	 *     In this case set iflag to a negative integer.
	 *
	 * @param n is a positive integer input variable set to the number
	 *     of functions and variables.
	 * @param x is an array of length n. On input x must contain
	 *     an initial estimate of the solution vector. On output x
	 *     contains the final estimate of the solution vector.
	 * @param fvec is an output array of length n which contains
	 *     the functions evaluated at the output x.
	 * @param xtol is a nonnegative input variable. Termination
	 *     occurs when the relative error between two consecutive
	 *     iterates is at most xtol.
	 * @param maxfev is a positive integer input variable. Termination
	 *     occurs when the number of calls to fcn is at least maxfev
	 *     by the end of an iteration.
	 * @param ml is a nonnegative integer input variable which specifies
	 *     the number of subdiagonals within the band of the
	 *     jacobian matrix. If the jacobian is not banded, set
	 *     ml to at least n - 1.
	 * @param mu is a nonnegative integer input variable which specifies
	 *     the number of superdiagonals within the band of the
	 *     jacobian matrix. If the jacobian is not banded, set
	 *     mu to at least n - 1.
	 * @param epsfcn is an input variable used in determining a suitable
	 *     step length for the forward-difference approximation. This
	 *     approximation assumes that the relative errors in the
	 *     functions are of the order of epsfcn. If epsfcn is less
	 *     than the machine precision, it is assumed that the relative
	 *     errors in the functions are of the order of the machine
	 *     precision.
	 * @param diag is an array of length n. If mode = 1 (see
	 *     below), diag is internally set. If mode = 2, diag
	 *     must contain positive entries that serve as
	 *     multiplicative scale factors for the variables.
	 * @param mode is an integer input variable. If mode = 1, the
	 *     variables will be scaled internally. If mode = 2,
	 *     the scaling is specified by the input diag. Other
	 *     values of mode are equivalent to mode = 1.
	 * @param factor is a positive input variable used in determining the
	 *     initial step bound. This bound is set to the product of
	 *     factor and the euclidean norm of diag*x if nonzero, or else
	 *     to factor itself. In most cases factor should lie in the
	 *     interval (.1,100.). 100. Is a generally recommended value.
	 * @param nprint is an integer input variable that enables controlled
	 *     printing of iterates if it is positive. In this case,
	 *     fcn is called with iflag = 0 at the beginning of the first
	 *     iteration and every nprint iterations thereafter and
	 *     immediately prior to return, with x and fvec available
	 *     for printing. If nprint is not positive, no special calls
	 *     of fcn with iflag = 0 are made.
	 * @param fjac is an output n by n array which contains the
	 *     orthogonal matrix q produced by the qr factorization
	 *     of the final approximate jacobian.
	 * @param ldfjac is a positive integer input variable not less than n
	 *     which specifies the leading dimension of the array fjac.
	 * @param r is an output array of length not less than (n*(n+1))/2 which contains the
	 *     upper triangular matrix produced by the qr factorization
	 *     of the final approximate jacobian, stored rowwise.
	 * @param qtf is an output array of length n which contains
	 *        the vector (q transpose)*fvec.
	 * @param wa1 work array of length n
	 * @param wa2 work array of length n
	 * @param wa3 work array of length n
	 * @param wa4 work array of length n
	 * @return info is an integer output variable. If the user has
	 *     terminated execution, info is set to the (negative)
	 *     value of iflag. See description of fcn. Otherwise,
	 *     info is set as follows.
	 *
	 *     info = 0   improper input parameters.
	 *
	 *     info = 1   relative error between two consecutive iterates
	 *                is at most xtol.
	 *
	 *     info = 2   number of calls to fcn has reached or exceeded
	 *                maxfev.
	 *
	 *     info = 3   xtol is too small. No further improvement in
	 *                the approximate solution x is possible.
	 *
	 *     info = 4   iteration is not making good progress, as
	 *                measured by the improvement from the last
	 *                five jacobian evaluations.
	 *
	 *     info = 5   iteration is not making good progress, as
	 *                measured by the improvement from the last
	 *                ten iterations.
	 *
	 * nfev is an integer output variable set to the number of calls to fcn.
	 */
	public static int hybrd(final IFunction func, final int n,
			final double[] x, final double[] fvec,
			final double xtol, final int maxfev, final int ml, final int mu,
			final double epsfcn,
			final double[] diag,
			final int mode, final double factor, final int nprint,// final int[] nfev,
			final double[] fjac, final int ldfjac,
			final double[] r, final double[] qtf,
			final double[] wa1, final double[] wa2, final double[] wa3, final double[] wa4)
	{
		// Local variables
		double delta = 0.;
		double xnorm = 0.;

		int info = INFO_STATUS_IMPROPER_INPUT_PARAMETERS;
		int iflag = 0;
		int nfev = 0;// TODO java: computed but not returned

		// check the input parameters for errors.

		if( (n <= 0) || (xtol < 0.) || (maxfev <= 0) || (ml < 0) || (mu < 0) ||
				(factor <= 0.) || (ldfjac < n) || (r.length < (n * (n + 1) / 2)) )
		{
			return terminate( func, x, fvec, nprint, iflag, info );
		}
		if( mode == 2 ) {
			for( int j = 0; j < n; ++j ) {
				if( diag[ j ] <= 0. ) {
					return terminate( func, x, fvec, nprint, iflag, info );
				}
			}
		}

		// evaluate the function at the starting point
		// and calculate its norm.

		iflag = func.fcn_nn( x, fvec, 1 );
		nfev = 1;
		if( iflag < 0 ) {
			return terminate( func, x, fvec, nprint, iflag, info );
		}
		double fnorm = enorm( n, fvec, 0 );

		// determine the number of calls to fcn needed to compute
		// the jacobian matrix.

		// Computing MIN
		final int i1 = ml + mu + 1;
		final int msum = Math.min( i1, n );

		// initialize iteration counter and monitors.

		int iter = 1;
		int ncsuc = 0;
		int ncfail = 0;
		int nslow1 = 0;
		int nslow2 = 0;

		// beginning of the outer loop.

		for( ; ; ) {
			boolean jeval = true;

			// calculate the jacobian matrix.

			iflag = fdjac1( func, n, x, fvec, fjac, ldfjac,
							ml, mu, epsfcn, wa1, wa2 );
			nfev += msum;
			if( iflag < 0 ) {
				return terminate( func, x, fvec, nprint, iflag, info );
			}

			// compute the qr factorization of the jacobian.

			qrfac( n, n, fjac, ldfjac, false, null, wa1, wa2, wa3 );

			// on the first iteration and if mode is 1, scale according
			// to the norms of the columns of the initial jacobian.

			if( iter == 1 ) {
				if( mode != 2 ) {
					for( int j = 0; j < n; ++j ) {
						diag[ j ] = wa2[ j ];
						if( wa2[ j ] == 0. ) {
							diag[ j ] = 1.;
						}
					}
				}

				// on the first iteration, calculate the norm of the scaled x
				// and initialize the step bound delta.

				for( int j = 0; j < n; ++j ) {
					wa3[ j ] = diag[ j ] * x[ j ];
				}
				xnorm = enorm( n, wa3, 0 );
				delta = factor * xnorm;
				if( delta == 0. ) {
					delta = factor;
				}
			}

			// form (q transpose)*fvec and store in qtf.

			for( int i = 0; i < n; ++i ) {
				qtf[ i ] = fvec[ i ];
			}

			for( int j = 0, j_ldfjac = 0; j < n; ++j, j_ldfjac += ldfjac ) {
				if( fjac[ j + j_ldfjac ] != 0. ) {
					double sum = 0.;
					for( int i = j, fi = i + j_ldfjac; i < n; ++i, ++fi ) {
						sum += fjac[ fi ] * qtf[ i ];
					}
					final double temp = -sum / fjac[ j + j_ldfjac ];
					for( int i = j, fi = i + j_ldfjac; i < n; ++i, ++fi ) {
						qtf[ i ] += fjac[ fi ] * temp;
					}
				}
			}

			// copy the triangular factor of the qr factorization into r.

			// boolean sign = false;// not used
			for( int j = 0, j_ldfjac = -1; j < n; ++j, j_ldfjac += ldfjac ) {
				int k = j;
				if( j >= 1 ) {
					for( int i = 1, fi = i + j_ldfjac; i <= j; ++i, ++fi ) {
						r[ k ] = fjac[ fi ];
						k += n - i;
					}
				}
				r[ k ] = wa1[ j ];
				// if( wa1[ j ] == 0. ) {
				//	sign = true;
				//}
			}

			// accumulate the orthogonal factor in fjac.

			qform( n, n, fjac, ldfjac, wa1 );

			// rescale if necessary.

			if( mode != 2 ) {
				for( int j = 0; j < n; ++j ) {
					// Computing MAX
					final double d1 = diag[ j ];
					final double d2 = wa2[ j ];
					diag[ j ] = Math.max( d1, d2 );
				}
			}

			// beginning of the inner loop.

			for( ; ; ) {

				// if requested, call fcn to enable printing of iterates.

				if( nprint > 0 ) {
					iflag = 0;
					if( (iter - 1) % nprint == 0 ) {
						iflag = func.fcn_nn( x, fvec, 0 );
					}
					if( iflag < 0 ) {
						return terminate( func, x, fvec, nprint, iflag, info );
					}
				}

				// determine the direction p.

				dogleg( n, r, diag, qtf, delta, wa1, wa2, wa3 );

				// store the direction p and x + p. calculate the norm of p.

				for( int j = 0; j < n; ++j ) {
					wa1[ j ] = -wa1[ j ];
					wa2[ j ] = x[ j ] + wa1[ j ];
					wa3[ j ] = diag[ j ] * wa1[ j ];
				}
				final double pnorm = enorm( n, wa3, 0 );

				// on the first iteration, adjust the initial step bound.

				if( iter == 1 ) {
					delta = Math.min( delta, pnorm );
				}

				// evaluate the function at x + p and calculate its norm.

				iflag = func.fcn_nn( wa2, wa4, 1 );
				++nfev;
				if( iflag < 0 ) {
					return terminate( func, x, fvec, nprint, iflag, info );
				}
				final double fnorm1 = enorm( n, wa4, 0 );

				// compute the scaled actual reduction.

				double actred = -1.;
				if( fnorm1 < fnorm ) {
					// Computing 2nd power
					final double d1 = fnorm1 / fnorm;
					actred = 1 - d1 * d1;
				}

				// compute the scaled predicted reduction.

				int k = 0;
				for( int i = 0; i < n; ++i ) {
					double sum = 0.;
					for( int j = i; j < n; ++j ) {
						sum += r[ k ] * wa1[ j ];
						++k;
					}
					wa3[ i ] = qtf[ i ] + sum;
				}
				final double temp = enorm( n, wa3, 0 );
				double prered = 0.;
				if( temp < fnorm ) {
					// Computing 2nd power
					final double d1 = temp / fnorm;
					prered = 1 - d1 * d1;
				}

				// compute the ratio of the actual to the predicted
				// reduction.

				double ratio = 0.;
				if( prered > 0. ) {
					ratio = actred / prered;
				}

				// update the step bound.

				if( ratio < P1 ) {
					ncsuc = 0;
					++ncfail;
					delta = P5 * delta;
				} else {
					ncfail = 0;
					++ncsuc;
					if( ratio >= P5 || ncsuc > 1 ) {
						// Computing MAX
						final double d1 = pnorm / P5;
						delta = Math.max( delta, d1 );
					}
					if( Math.abs( ratio - 1 ) <= P1 ) {
						delta = pnorm / P5;
					}
				}

				// test for successful iteration.

				if( ratio >= P0001 ) {

					// successful iteration. update x, fvec, and their norms.

					for( int j = 0; j < n; ++j ) {
						x[ j ] = wa2[ j ];
						wa2[ j ] = diag[ j ] * x[ j ];
						fvec[ j ] = wa4[ j ];
					}
					xnorm = enorm( n, wa2, 0 );
					fnorm = fnorm1;
					++iter;
				}

				// determine the progress of the iteration.

				++nslow1;
				if( actred >= P001 ) {
					nslow1 = 0;
				}
				if( jeval ) {
					++nslow2;
				}
				if( actred >= P1 ) {
					nslow2 = 0;
				}

				// test for convergence.

				if( delta <= xtol * xnorm || fnorm == 0. ) {
					info = INFO_STATUS_NO_IMPROVEMENTS_BETWEEN_ITERATIONS;
				}
				if( info != 0 ) {
					return terminate( func, x, fvec, nprint, iflag, info );
				}

				// tests for termination and stringent tolerances.

				if( nfev >= maxfev ) {
					info = INFO_STATUS_MAX_CALLS_REACHED;
				}
				// Computing MAX
				final double d1 = P1 * delta;
				if( P1 * Math.max( d1, pnorm ) <= MACHINE_PRECISION * xnorm ) {
					info = INFO_STATUS_TOLERANCE_TOO_SMALL;
				}
				if( nslow2 == 5 ) {
					info = INFO_STATUS_NO_PROGRESS_5;
				}
				if( nslow1 == 10 ) {
					info = INFO_STATUS_NO_PROGRESS_10;
				}
				if( info != 0 ) {
					return terminate( func, x, fvec, nprint, iflag, info );
				}

				// criterion for recalculating jacobian approximation
				// by forward differences.

				if( ncfail == 2 ) {
					break;// TERMINATE_INNER_LOOP
				}

				// calculate the rank one modification to the jacobian
				// and update qtf if necessary.

				for( int j = 0, j_ldfjac = 0; j < n; ++j, j_ldfjac += ldfjac ) {
					double sum = 0.;
					for( int i = 0, fi = j_ldfjac; i < n; ++i, ++fi ) {
						sum += fjac[ fi ] * wa4[ i ];
					}
					wa2[ j ] = (sum - wa3[ j ]) / pnorm;
					wa1[ j ] = diag[ j ] * (diag[ j ] * wa1[ j ] / pnorm);
					if( ratio >= P0001 ) {
						qtf[ j ] = sum;
					}
				}

				// compute the qr factorization of the updated jacobian.

				r1updt( n, n, r, wa1, wa2, wa3 );
				r1mpyq( n, n, fjac, ldfjac, wa2, wa3 );
				r1mpyq( 1, n, qtf, 1, wa2, wa3 );

				// end of the inner loop.

				jeval = false;
			}// for( ; ; )
		}// for( ; ; )
	}

	/**
	 * Subroutine hybrd1
	 *
	 * The purpose of hybrd1 is to find a zero of a system of
	 * n nonlinear functions in n variables by a modification
	 * of the Powell hybrid method. This is done by using the
	 * more general nonlinear equation solver hybrd. The user
	 * must provide a subroutine which calculates the functions.
	 * The jacobian is then calculated by a forward-difference
	 * approximation.
	 *
	 * Argonne national laboratory. MINPACK project. March 1980.
	 * Burton S. Garbow, Kenneth E. Hillstrom, Jorge J. More

	 * @param func is the name of the user-supplied subroutine which
	 *     calculates the functions. fcn must be declared
	 *     in an external statement in the user calling
	 *     program, and should be written as follows.
	 * <pre>
	 *     int fcn(double[] x,double[] fvec,int iflag) {
	 *     ----------
	 *     calculate the functions at x and
	 *     return this vector in fvec and iflag
	 *     ----------
	 *     }
	 * </pre>
	 *     The value of iflag should not be changed by fcn unless
	 *     the user wants to terminate execution of fdjac1.
	 *     In this case set iflag to a negative integer.
	 *
	 * @param n is a positive integer input variable set to the number
	 *     of functions and variables.
	 *
	 * @param x is an array of length n. on input x must contain
	 *     an initial estimate of the solution vector. On output x
	 *     contains the final estimate of the solution vector.
	 *
	 * @param fvec is an output array of length n which contains
	 *     the functions evaluated at the output x.
	 *
	 * @param tol is a nonnegative input variable. Termination occurs
	 *     when the algorithm estimates that the relative error
	 *     between x and the solution is at most tol.
	 *
	 * @param info is an integer output variable. If the user has
	 *     terminated execution, info is set to the (negative)
	 *     value of iflag. See description of fcn. Otherwise,
	 *     info is set as follows.
	 *
	 *     info = 0   improper input parameters.
	 *
	 *     info = 1   algorithm estimates that the relative error
	 *                between x and the solution is at most tol.
	 *
	 *     info = 2   number of calls to fcn has reached or exceeded
	 *                200*(n+1).
	 *
	 *     info = 3   tol is too small. No further improvement in
	 *                the approximate solution x is possible.
	 *
	 *     info = 4   iteration is not making good progress.
	 */
	public static int hybrd1(final IFunction func, final double[] x, final double[] fvec, final double tol) {
		/* check the input parameters for errors. */

		if( (fvec.length < x.length) || (tol < 0.) ) {
			return 0;
		}

		/* call hybrd. */
		final int n = x.length;
		final double factor = 100.;
		// final int[] nfev = new int[ 1 ];
		final int maxfev = (n + 1) * 200;
		final int ml = n - 1;
		final int mu = n - 1;
		final double epsfcn = 0.;
		final int mode = 2;

		final int nprint = 0;
		final int lr = (n * (n + 1)) >>> 1;

		final double[] diag = new double[ n ];// an array of length n to scale vars
		for( int j = 0; j < n; ++j ) {
			diag[ j ] = 1.;
		}
		final double[] fjac = new double[ n * n ];// output jacobian
		final double[] r = new double[ lr ];// is an output array of length lr which contains the upper triangular matrix
		final double[] qtf = new double[ n ];// is an output array of length n which contains the vector (q transpose)*fvec.
		final double[] wa1 = new double[ n ];// work array of length n
		final double[] wa2 = new double[ n ];// work array of length n
		final double[] wa3 = new double[ n ];// work array of length n
		final double[] wa4 = new double[ n ];// work array of length n
		int info = hybrd( func, n, x, fvec, tol, maxfev, ml, mu, epsfcn,
				diag, mode, factor, nprint,
				// nfev,
				fjac, n, r, qtf, wa1, wa2, wa3, wa4 );
		if( info == INFO_STATUS_NO_PROGRESS_10 ) {
			info = INFO_STATUS_NO_PROGRESS_5;
		}
		return info;
	}
}
