package libmath;

import java.util.ArrayList;

/**
 * Java version of HOMPACK.
 *
 * HOMPACK is a suite of FORTRAN subroutines for solving nonlinear
 * systems of equations by homotopy methods.
 *
 * It includes subroutines from LAPACK, BLAS to make it complete.
 *
 * https://netlib.org/hompack/
 * http://www.netlib.no/netlib/hompack/index.html
 */
public final class JHomPack {

	/**
	 * if IFLG2(M) = -2, then the M-th path is tracked.
	 * Otherwise the M-th path is skipped.
	 *   Thus, to find all solutions set IFLG2(M) = -2 for M=1,...,TOTDG.
	 *   selected paths can be rerun by setting IFLG2(M) = -2 for
	 *   the paths to be rerun and IFLG(M) != -2 for the others.
	 */
	public static final int FOLLOW_PATH = -2;

	//--------

	//  D1MACH( 1 ) = B^(EMIN-1), the smallest positive magnitude.
	//  D1MACH( 2 ) = B^EMAX*(1 - B^(-T)), the largest magnitude.
	//  D1MACH( 3 ) = B^(-T), the smallest relative spacing.
	//  D1MACH( 4 ) = B^(1-T), the largest relative spacing.
	//  D1MACH( 5 ) = LOG10(B)

	/** b^(emin - 1), DPMPAR( realm, MIN(2) );// 2.2250738585072014e-308 | 1.17549435e-38F */
	// private static final double SMALEST_MAGNITUDE      = 2.2250738585072014e-308;// ( 1 )

	/** b^emax*(1 - b^(-t)), DPMPAR( realm, MAX(3) );// 1.7976931348623157e+308 | 3.40282347e+38F */
	private static final double LAGEST_MAGNITUDE       = 1.7976931348623157e+308;// ( 2 )

	// private static final double HALF_MACHINE_PRECISION = 1.1102230246251565404236316680908e-16;// ( 3 )

	/** b^(1 - t), DPMPAR( realm, EPSILON(1) );// 2^-52 = 2.2204460492503131e-16 | 2^-23 = 1.19209290e-07F */
	private static final double MACHINE_PRECISION      = 2.2204460492503131e-16;// ( 4 )

	/** log10(2) */
	// private static final double LOG10_2 = 0.301029995663981195213738894724;// ( 5 )
/*
	static double d1mach(final int i) {
		switch( i ) {

			case 1: return SMALEST_MAGNITUDE;

			case 2: return LAGEST_MAGNITUDE;

			// = FLT_RADIX  ^ - DBL_MANT_DIG for IEEE:  = 2^-53 = 1.110223e-16 = .5*DBL_EPSILON
			case 3: return 0.5 * MACHINE_PRECISION;

			// = FLT_RADIX  ^ (1- DBL_MANT_DIG) = for IEEE:  = 2^-52 = DBL_EPSILON
			case 4:  return MACHINE_PRECISION;

			case 5: return LOG10_2;

			default: return 0.0;
	    }
	}
*/
	/*****************************************/
	/**
	 *  Evaluate  f(x)  and return in the vector  V.
	 *
	 * @param x variable values
	 * @param v solution
	 */
	private static void f(final double[] x, final int xOff, final double[] v, final int vOff) {
	}

	/**
	 * Return in  v  the Kth column of the jacobian matrix of
	 * f(x) evaluated at  X .
	 *
	 * diff( f, variable )
	 *
	 * @param x variable values
	 * @param v solution
	 * @param k a column (index of a variable). Starts from 0.
	 */
	private static void fjac(final double[] x, final int xOff, final double[] v, final int vOff, final int k) {
	}
	/*****************************************/

	/**
	 * DCOPY copies a vector, x, to a vector, y.
	 * uses unrolled loops for increments equal to 1.
	 *
	 * 	Authors:
	 *  ========
	 *
	 * Univ. of Tennessee
	 * Univ. of California Berkeley
	 * Univ. of Colorado Denver
	 * NAG Ltd.
	 * Further Details:
	 *     jack dongarra, linpack, 3/11/78.
	 *     modified 12/3/93, array(1) declarations changed to array(*)
	 *
	 * @param n number of elements in input vector(s)
	 *
	 * @param dx is double precision array, dimension (1 + (N - 1)*abs( INCX ))
	 *
	 * @param incX storage spacing between elements of DX
	 *
	 * @param dy is double precision array, dimension (1 + (N - 1)*abs( INCY ))
	 *
	 * @param incY is storage spacing between elements of DY
	 */
	private static void dcopy(final int n, final double[] dx, int dxOff, final int incX, final double[] dy, int dyOff, final int incY) {
		//if( n <= 0 ) {
		//	return;
		//}
		if( (incX == 1) && (incY == 1) ) {
			// code for both increments equal to 1
			// clean-up loop
			final int m = n % 7;
			if( m != 0 ) {
				for( int i = 0; i < m; ++i ) {
					dy[ dyOff++ ] = dx[ dxOff++ ];
				}
				if( n < 7 ) {
					return;
				}
			}
			for( int i = m; i < n; i += 7, dxOff += 7, dyOff += 7 ) {
				dy[ dyOff     ] = dx[ dxOff ];
				dy[ dyOff + 1 ] = dx[ dxOff + 1 ];
				dy[ dyOff + 2 ] = dx[ dxOff + 2 ];
				dy[ dyOff + 3 ] = dx[ dxOff + 3 ];
				dy[ dyOff + 4 ] = dx[ dxOff + 4 ];
				dy[ dyOff + 5 ] = dx[ dxOff + 5 ];
				dy[ dyOff + 6 ] = dx[ dxOff + 6 ];
			}
			return;
		}

		// code for unequal increments or equal increments
		// not equal to 1

		if( incX < 0 ) {
			dxOff += (1 - n) * incX;
		}
		if( incY < 0 ) {
			dyOff += (1 - n) * incY;
		}
		for( int i = 0; i < n; ++i ) {
			dy[ dyOff ] = dx[ dxOff ];
			dxOff += incX;
			dyOff += incY;
		}
	}

	/**
	 * Finds the index of element having max. absolute value.
	 * Jack Dongarra, LINPACK, 3/11/78.
	 *   modified 3/93 to return if incx <= 0.
	 *   modified 12/3/93, array(1) declarations changed to array(*)
	 *
	 * @param n number of elements in input vector
	 * @param dx is double precision array
	 * @param incX storage spacing between elements of DX
	 * @return an index of a max absolute value, starts form 0
	 */
	private static int idamax(final int n, final double[] dx, final int incX) {

		if( (n < 1) || (incX <= 0) ) {
			return 0;
		}
		int ret_val = 0;
		if( n == 1 ) {
			return ret_val;
		}
		if( incX == 1 ) {
			/* code for increment equal to 1 */
			double dmax = Math.abs( dx[ 0 ] );
			for( int i = 1; i < n; ++i ) {
				if( Math.abs( dx[ i ] ) > dmax ) {
					ret_val = i;
					dmax = Math.abs( dx[ i ] );
				}
			}
			return ret_val;
		}

		// code for increment not equal to 1

		int ix = 1;
		double dmax = Math.abs( dx[ 0 ] );
		ix += incX;
		for( int i = 1; i < n; ++i ) {
			if( Math.abs( dx[ ix ] ) > dmax ) {
				ret_val = i;
				dmax = Math.abs( dx[ ix ] );
			}
			ix += incX;
		}
		return ret_val;
	} /* idama */

	/**
	 * This subroutine performs multiplication of complex numbers:
	 * ZZZZ = XXXX*YYYY
	 *
	 * NOTE:  In the calling routine, ZZZZ should not be the same
	 * as XXXX or YYYY.  However, XXXX may be the same as YYYY.
	 * Thus, "CALL MULP(X,X,Z)" is ok, but "CALL MULP(X,Y,X)" is not.
	 *
	 * ON INPUT:
	 *
	 * @param xxxx XXXX  is an array of length two representing the first complex
	 *       number, where XXXX(1) = real part of XXXX and XXXX(2) =
	 *       imaginary part of XXXX.
	 *
	 * @param yyyy YYYY  is an array of length two representing the second complex
	 *       number, where YYYY(1) = real part of YYYY and YYYY(2) =
	 *       imaginary part of YYYY.
	 *
	 * ON OUTPUT:
	 *
	 * @param zzzz ZZZZ  is an array of length two representing the result of
	 *       the multiplication, ZZZZ = XXXX*YYYY, where ZZZZ(1) =
	 *       real part of ZZZZ and ZZZZ(2) = imaginary part of ZZZZ.
	 */
	private static void mulp(final double[] xxxx, final int xOff, final double[] yyyy, final int yOff, final double[] zzzz, final int zOff) {
		final double x_re = xxxx[ xOff ];
		final double x_im = xxxx[ xOff + 1 ];
		final double y_re = yyyy[ yOff ];
		final double y_im = yyyy[ yOff + 1 ];
		zzzz[ zOff     ] = x_re * y_re - x_im * y_im;
		zzzz[ zOff + 1 ] = x_re * y_im + x_im * y_re;
	} /* mulp */

	/**
	 * Purpose : To compute
	 * <pre>
	 * dy = da * dx + dy
	 * </pre>
	 * constant times a vector plus a vector.
	 * uses unrolled loops for increments equal to one.
	 * Jack Dongarra, LINPACK, 3/11/78.
	 * modified 12/3/93, array(1) declarations changed to array(*)
	 *
	 * --- Input ---
	 * @param n    : number of elements in input vector(s)
	 * @param da   : double scalar multiplier
	 * @param dx   : double vector with n+1 elements, dx[0] is not used
	 * @param incx : storage spacing between elements of dx
	 * @param dy   : double vector with n+1 elements, dy[0] is not used
	 * @param incy : storage spacing between elements of dy
	 * --- Output ---
	 * dy = da * dx + dy, unchanged if n <= 0
	 *    For i = 0 to n-1, replace dy[ly+i*incy] with
	 *    da*dx[lx+i*incx] + dy[ly+i*incy], where lx = 1
	 *    if  incx >= 0, else lx = (-incx)*(n-1)+1 and ly is
	 *    defined in a similar way using incy.
	 */
	private static void daxpy(final int n, final double da, final double[] dx, int dxOff, final int incx, final double[] dy, int dyOff, final int incy) {
		if( n <= 0 ) {
			return;
		}
		if( da == 0.) {
			return;
		}
		if( (incx == 1) && (incy == 1) ) {
			// code for both increments equal to 1   clean-up loop
			final int m = n % 4;
			if( m != 0 ) {
				for( int i = 0; i < m; ++i ) {
					dy[ dyOff++ ] += da * dx[ dxOff++ ];
				}
				if( n < 4 ) {
					return;
				}
			}
			for( int i = m; i < n; i += 4, dxOff += 4, dyOff += 4 ) {
				dy[ dyOff     ] += da * dx[ dxOff ];
				dy[ dyOff + 1 ] += da * dx[ dxOff + 1 ];
				dy[ dyOff + 2 ] += da * dx[ dxOff + 2 ];
				dy[ dyOff + 3 ] += da * dx[ dxOff + 3 ];
			}
			return;
		}

		// code for unequal increments or equal increments not equal to 1
		if( incx < 0 ) {
			dxOff += (1 - n) * incx;
		}
		if( incy < 0 ) {
			dyOff += (1 - n) * incy;
		}
		for( int i = 0; i < n; ++i ) {
			dy[ dyOff ] += da * dx[ dxOff ];
			dxOff += incx;
			dyOff += incy;
		}
	} /* daxpy */

	/**
	 * This subroutine performs division  of complex numbers:
	 * ZZZZ = XXXX/YYYY
	 *
	 * ON INPUT:
	 *
	 * @param xxxx XXXX  is an array of length two representing the first complex
	 *       number, where XXXX(1) = real part of XXXX and XXXX(2) =
	 *       imaginary part of XXXX.
	 *
	 * @param yyyy YYYY  is an array of length two representing the second complex
	 *       number, where YYYY(1) = real part of YYYY and YYYY(2) =
	 *       imaginary part of YYYY.
	 *
	 * ON OUTPUT:
	 *
	 * @param zzzz ZZZZ  is an array of length two representing the result of
	 *       the division, ZZZZ = XXXX/YYYY, where ZZZZ(1) =
	 *       real part of ZZZZ and ZZZZ(2) = imaginary part of ZZZZ.
	 *
	 * @return IERR =
	 *  {@code true}   If division would have caused overflow.  in this case, the
	 *      appropriate parts of ZZZZ are set equal to the largest
	 *      floating point number, as given by function  D1MACH .
	 *
	 *  {@code false}   If division does not cause overflow.
	 */
	private static boolean divp(final double[] xxxx, final int xOff, final double[] yyyy, final int yOff, final double[] zzzz, final int zOff) {
		boolean ierr = false;
		final double x_re = xxxx[ xOff ];
		final double x_im = xxxx[ xOff + 1 ];
		final double y_re = yyyy[ yOff ];
		final double y_im = yyyy[ yOff + 1 ];
		final double denom = y_re * y_re + y_im * y_im;
		double xnum = x_re * y_re + x_im * y_im;
		if( Math.abs( denom ) >= 1. || Math.abs( denom ) < 1. && Math.abs( xnum ) / LAGEST_MAGNITUDE < Math.abs( denom ) ) {
			zzzz[ zOff ] = xnum / denom;
		} else {
			zzzz[ zOff ] = LAGEST_MAGNITUDE;
			ierr = true;
		}
		xnum = x_im * y_re - x_re * y_im;
		if( Math.abs( denom ) >= 1. || Math.abs( denom ) < 1. && Math.abs( xnum ) / LAGEST_MAGNITUDE < Math.abs( denom ) ) {
			zzzz[ zOff + 1 ] = xnum / denom;
			return ierr;
		}
			zzzz[ zOff + 1 ] = LAGEST_MAGNITUDE;
		return true;
	} /* divp */

	/**
	 * DNRM2 returns the euclidean norm of a vector via the function
	 * name, so that
	 * <pre>
	 *     DNRM2 := sqrt( x'*x )
	 * </pre>
	 * This version written on 25-October-1982.
	 *    Modified on 14-October-1993 to inline the call to SLASSQ.
	 *    Sven Hammarling, Nag Ltd.
	 *
	 * @param n N is INTEGER number of elements in input vector(s)
	 * @param x X is DOUBLE PRECISION array, dimension ( 1 + ( N - 1 )*abs( INCX ) )
	 * @param incx INCX is INTEGER, storage spacing between elements of X
	 *    If INCX > 0, X(1+(i-1)*INCX) = x(i) for 1 <= i <= n
	 *    If INCX < 0, X(1-(n-i)*INCX) = x(i) for 1 <= i <= n
	 *    If INCX = 0, x isn't a vector so there is no need to call
	 *    this subroutine.  If you call it anyway, it will count x(1)
	 *    in the vector norm N times.
	 */
	private static double dnrm2(final int n, final double[] x, final int xOff, final int incx) {
		if( (n > 0) && (incx > 0) ) {
			if( n == 1 ) {
				return Math.abs( x[ xOff ] );
			}
			double scale = 0.;
			double ssq = 1.;

			// The following loop is equivalent to this call to the LAPACK
			// auxiliary routine:   CALL SLASSQ( N, X, INCX, SCALE, SSQ )

			for( int ix = xOff + (n - 1) * incx; ix >= xOff; ix -= incx ) {
				double absxi = x[ ix ];
				if( absxi != 0. ) {
					absxi = Math.abs( absxi );
					if( scale < absxi ) {
						final double temp = scale / absxi;
						ssq = ssq * (temp * temp) + 1.;
						scale = absxi;
					} else {
						final double temp = absxi / scale;
						ssq += temp * temp;
					}
				}
			}
			return scale * Math.sqrt( ssq );
		}
		return 0.0;
	} /* dnrm2 */

	/**
	 * Replace by Math.pow ?
	 *
	 * @param x value
	 * @param n power
	 * @return result, x ^ n
	 */
	private static double pow_di(double x, int n) {
		double pow = 1.;

		if( n != 0 ) {
			if( n < 0 ) {
				n = -n;
				x = 1. / x;
			}
			for( ; ; ) {
				if( (n & 1) != 0 ) {
					pow *= x;
				}
				if( (n >>>= 1) == 0 ) {
					break;
				}
				x *= x;
			}
		}
		return pow;
	}

	/**
	 * This subroutine takes a non-negative power of a complex number:
	 * YYYY = XXXX^NNNN using de Moivre's formula:
	 * <pre>
	 *     YYYY = R^NNNN * (COS(NNNN*THETA),SIN(NNNN*THETA)),
	 * </pre>
	 * where R=DNRM2(2,XXXX,1) and THETA=ATAN2(XXXX(2),XXXX(1)).
	 *
	 * NOTE: POWP sets 0^0 equal to 1.
	 *
	 * ON INPUT:
	 *
	 * @param nnnn NNNN  is a non-negative integer.
	 *
	 * @param xxxx XXXX  is an array of length two representing a complex
	 *       number, where XXXX(1) = real part of XXXX and XXXX(2) =
	 *       imaginary part of XXXX.
	 *
	 * ON OUTPUT:
	 *
	 * @param yyyy YYYY  is an array of length two representing the result of
	 *       the power, YYYY = XXXX^NNNN, where YYYY(1) =
	 *       real part of YYYY and YYYY(2) = imaginary part of YYYY.
	 */
	private static void powp(final int nnnn, final double[] xxxx, final int xOff, final double[] yyyy, final int yOff) {
		if( nnnn == 0 ) {
			yyyy[ yOff ] = 1.;
			yyyy[ yOff + 1 ] = 0.;
			return;
		}
		if( nnnn == 1 ) {
			yyyy[ yOff ] = xxxx[ xOff ];
			yyyy[ yOff + 1 ] = xxxx[ xOff + 1 ];
			return;
		}
		final double r = dnrm2( 2, xxxx, xOff, 1 );
		if( r == 0. ) {
			yyyy[ yOff ] = 0.;
			yyyy[ yOff + 1 ] = 0.;
			return;
		}
		final double rr = pow_di( r, nnnn );
		final double tt = (double) nnnn * Math.atan2( xxxx[ xOff + 1 ], xxxx[ xOff ] );
		yyyy[ yOff ] = rr * Math.cos( tt );
		yyyy[ yOff + 1 ] = rr * Math.sin( tt );
	} /* powp */

	/**
	 * Forms the dot product of two vectors.
	 * uses unrolled loops for increments equal to one.
	 *
	 * Jack Dongarra, LINPACK, 3/11/78.
	 * modified 12/3/93, array(1) declarations changed to array(*)
	 *
	 * @param n number of elements in input vector(s)
	 * @param sx is double precision array
	 * @param incX storage spacing between elements of DX
	 * @param sy is double precision array
	 * @param incY storage spacing between elements of DY
	 */
	private static double ddot(final int n, final double[] sx, int sxOff, final int incX, final double[] sy, int syOff, final int incY) {
		int i;

		double stemp = 0.0;
		if( n > 0 ) {
			if( (incX == 1) && (incY == 1) ) {
				// code for both increments equal to 1
				final int m = n - 4;
				for( i = 0; i < m; i += 5, sxOff += 5, syOff += 5 ) {
					stemp += sx[ sxOff ] * sy[ syOff ]
							+ sx[ sxOff + 1 ] * sy[ syOff + 1 ]
							+ sx[ sxOff + 2 ] * sy[ syOff + 2 ]
							+ sx[ sxOff + 3 ] * sy[ syOff + 3 ]
							+ sx[ sxOff + 4 ] * sy[ syOff + 4 ];
				}

				for( ; i < n; i++ ) {
					// clean-up loop
					stemp += sx[ sxOff ] * sy[ syOff ];
					sxOff++;
					syOff++;
				}
				return stemp;
			}
			if( incX < 0 ) {
				sxOff += (1 - n) * incX;
			}
			if( incY < 0 ) {
				syOff += (1 - n) * incY;
			}
			for( i = 0; i < n; i++ ) {
				stemp += sx[ sxOff ] * sy[ syOff ];
				sxOff += incX;
				syOff += incY;
			}
		}

		return stemp;
	} /* ddot */

	/**
	 * Scales a vector by a constant.
	 * Uses unrolled loops for increment equal to 1.
	 *
	 * Jack Dongarra, LINPACK, 3/11/78.
	 * modified 3/93 to return if incx <= 0.
	 * modified 12/3/93, array(1) declarations changed to array(*)
	 *
	 * --Input--
     * @param n N  number of elements in input vector(s)
     * @param da DA  double precision scale factor
     * @param dx DX  double precision vector with N elements
     * @param incX INCX  storage spacing between elements of DX
     *
     * --Output--
     * DX  double precision result (unchanged if N.LE.0)
     *
     * Replace double precision DX by double precision DA*DX.
     * For I = 0 to N-1, replace DX(IX+I*INCX) with  DA * DX(IX+I*INCX),
     * where IX = 1 if INCX .GE. 0, else IX = 1+(1-N)*INCX.
	 */
	private static void dscal(int n, final double da, final double[] dx, int dxOff, final int incX) {
		if( (n > 0) && (incX > 0) ) {
			if( incX == 1 ) {
				// code for increment equal to 1
				n += dxOff;
				final int m = n - 4;
				for( ; dxOff < m; dxOff += 5 ) {
					dx[ dxOff     ] = da * dx[ dxOff ];
					dx[ dxOff + 1 ] = da * dx[ dxOff + 1 ];
					dx[ dxOff + 2 ] = da * dx[ dxOff + 2 ];
					dx[ dxOff + 3 ] = da * dx[ dxOff + 3 ];
					dx[ dxOff + 4 ] = da * dx[ dxOff + 4 ];
				}
				for( ; dxOff < n; ++dxOff ) {
					// clean-up loop
					dx[ dxOff ] = da * dx[ dxOff ];
				}
				return;
			}
			for( final int nincx = dxOff + n * incX; dxOff < nincx; dxOff += incX ) {
				dx[ dxOff ] = da * dx[ dxOff ];
			}
		}
	} /* dscal */

	/**
	 * FFUNP  evaluates the system "F(X)=0" and its partial
	 * derivatives, using the "TABLEAU" input: N,NUMT,KDEG,COEF.
	 *
	 * FFUNP  can be made more efficient by customizing it to
	 * particular system types.  for example,
	 * if X(1)^2 and X(1)^3 are used in several
	 * equations, the current  FFUNP  recomputes both of these for
	 * each equation.  But (of course) we can compute
	 * X1SQ=X(1)^2 and X1CU=XSQ(1)*X(1), and
	 * use these in each of the equations.
	 *
	 * The part of the code below labeled "BLOCK A" can be
	 * customized in this way.   (The code outside of
	 * BLOCK A concerns the projective transformation and need not
	 * be changed.)  However, BLOCK A requires the homogeneous form
	 * of the polynomials rather than the standard form.  further,
	 * the partial derivatives with respect to all N+1 projective
	 * variables must be computed.  More explicitly,
	 * the original system, F(X)=0, is given in "NON-HOMOGENEOUS FORM" as
	 * described in subroutine POLSYS.  F(X)  is
	 * represented in "HOMOGENEOUS FORM" as follows:
	 * <pre>
	 *              NUMT(J)
	 *    F(J) =     SUM   TRM(J,K)
	 *               K=1
	 * </pre>
	 * where  TRM(J,K) = COEF(J,K) * XX(J,1,K)*XX(J,2,K)* ... *XX(J,N+1,K)
	 *
	 * with XX(J,L,K) = X(L)^KDEG(J,L,K) for J=1 to N, L=1 to N, and
	 * K=1 to NUMT(J), and with XX(J,N+1,K) = XNP1^KDEG(J,N+1,K) for J=1 to
	 * N and K=1 to NUMT(J), where  XNP1  is the "HOMOGENEOUS COORDINATE,"
	 * KDEG(J,N+1,K)=IDEG(J)-(KDEG(J,1,K)+ ... + KDEG(J,N,K)),
	 * and IDEG(J) the degree of the J-th equation.   XNP1  is generated
	 * from  X  and  CL  before BLOCK A.
	 *
	 * In this discussion we have omitted, for simplicity of
	 * exposition, the leading index, which differentiates the
	 * real and imaginary parts.  However, this index must not be
	 * omitted in the code.
	 *
	 * We complete the exposition of "REPLACING BLOCK A WITH MORE EFFICIENT
	 * CODE" with an explicit example.  First, the system is described.
	 * then the code that should be used is given (commented out).
	 * in tests  POLSYS  with the more efficient  FFUNP  ran about twice as
	 * fast as with the generic  FFUNP .
	 *
	 * Here is the system to be solved:
	 * <pre>
	 *     F(1) = COEF(1,1) * X(1)^4
	 *          + COEF(1,2) * X(1)^3 * X(2)
	 *          + COEF(1,3) * X(1)^3
	 *          + COEF(1,4) * X(1)
	 *          + COEF(1,5)
	 *     F(2) = COEF(2,1) * X(1)   * X(2)^2
	 *          + COEF(2,2)            X(2)^2
	 *          + COEF(2,3)
	 * </pre>
	 * The replacement code requires the following declarations:
	 *     double precision X1SQ,X1CU,X2SQ,X3SQ,X3CU,
	 *      TEMPA,TEMPB,TEMPC,TEMPD,TEMPE,TEMPF
	 *     dimension X1SQ(2),X1CU(2),X2SQ(2),X3SQ(2),X3CU(2),
	 *      TEMPA(2),TEMPB(2),TEMPC(2),TEMPD(2),TEMPE(2),TEMPF(2)
	 *
	 * Here is code to replace BLOCK A:
	 *
	 * ******************  BEGIN BLOCK A  *******************
	 * <pre>
	 *     CALL MULP(X(1,1),X(1,1),X1SQ)
	 *     CALL MULP(X1SQ  ,X(1,1),X1CU)
	 *     CALL MULP(X(1,2),X(1,2),X2SQ)
	 *     CALL MULP(XNP1,  XNP1,  X3SQ)
	 *     CALL MULP(X3SQ  ,XNP1,  X3CU)
	 *
	 *     DO 1 I=1,2
	 *       TEMPA(I)=   COEF(1,1) * X(I,1)
	 *    &            + COEF(1,2) * X(I,2)
	 *    &            + COEF(1,3) * XNP1(I)
	 *       TEMPB(I)=   COEF(1,4) * X(I,1)
	 *    &            + COEF(1,5) * XNP1(I)
	 * 1   CONTINUE
	 *
	 *     CALL MULP(X1SQ,  TEMPA,TEMPC)
	 *     CALL MULP(X(1,1),TEMPC,TEMPD)
	 *     CALL MULP(X3SQ,  TEMPB,TEMPE)
	 *     CALL MULP(XNP1,  TEMPE,TEMPF)
	 *
	 *     DO 2 I=1,2
	 *       F(I,1)=TEMPD(I) + TEMPF(I)
	 *       DF(I,1,1)= 3. *TEMPC(I) + COEF(1,1)*X1CU(I) + COEF(1,4)*X3CU(I)
	 *       DF(I,1,2)= COEF(1,2) * X1CU(I)
	 *       DF(I,1,3)= COEF(1,3)*X1CU(I) + 3. *TEMPE(I) + COEF(1,5)*X3CU(I)
	 *
	 *       TEMPA(I) = COEF(2,1) * X(I,1) + COEF(2,2) * XNP1(I)
	 *  2  CONTINUE
	 *
	 *     CALL MULP(TEMPA,X(1,2),TEMPB)
	 *     CALL MULP(TEMPB,X(1,2),TEMPC)
	 *
	 *     DO 3 I=1,2
	 *       F(I,2) = TEMPC(I) + COEF(2,3) * X3CU(I)
	 *       DF(I,2,1) = COEF(2,1) * X2SQ(I)
	 *       DF(I,2,2) = 2. * TEMPB(I)
	 *       DF(I,2,3) = COEF(2,2) * X2SQ(I) + COEF(2,3) * 3. * X3SQ(I)
	 *  3  CONTINUE
	 * </pre>
	 * ******************  END OF BLOCK A  *******************
	 *
	 * ON INPUT:
	 *
	 * N  is the number of equations and variables.
	 *
	 * NUMT(J)  is the number of terms in the Jth equation.
	 *
	 * MMAXT  is an upper bound on NUMT(J) for J=1 to N.
	 *
	 * KDEG(J,L,K)  is the degree of the L-th variable in the K-th term
	 *   of the J-th equation.
	 *
	 * COEF(J,K)  is the K-th coefficient of the J-th equation.
	 *
	 * CL  is used to define the projective transformation.  If
	 *   the projective transformation is not specified, then  CL
	 *   contains dummy values.
	 *
	 * @param x X(1,J), X(2,J)  are the real and imaginary parts respectively of
	 *   the J-th independent variable.
	 *
	 * XX workspace variable.
	 * TRM workspace variable.
	 * DTRM workspace variable.
	 * CLX workspace variable.
	 * DXNP1 workspace variable.
	 *
	 * ON OUTPUT:
	 *
	 * F(1,J), F(2,J)  are the real and imaginary parts respectively of
	 *   the J-th equation.
	 *
	 * DF(1,J,K), DF(2,J,K)  are the real and imaginary parts respectively
	 *   of the K-th partial derivative of the J-th equation.
	 *
	 * Variables: XNP1,TEMP1,TEMP2.
	 *
	 * NOTE:  XNP1(1), XNP1(2)  are the real and imaginary parts,
	 *   respectively, of the projective variable.  XNP1  is unity
	 *   if the projective transformation is not specified.
	 */
	private static void ffunp(final PolsysSolver solver, final double[] x, final int xOff) {
		// final double[] xnp1 = new double[ 2 ];// X(N+1)
		// final double[] temp1 = new double[ 2 ];
		// final double[] temp2 = new double[ 2 ];
		final int xnp1_off = 0;
		final int temp1_off = 2;
		final int temp2_off = 4;
		final int xxx_off = 6;
		final double[] xxx = solver.mXX;

		final int n = solver.mVarAndEqNumber;
		final int np1 = n + 1;
		final int n2 = n + n;

		// GENERATE XNP1, THE PROJECTIVE COORDINATE, AND ITS DERIVATIVES.

		for( int j2 = 0; j2 < n2; j2 += 2 ) {
			mulp( solver.mCL, j2, x, xOff + j2, solver.mCLX, j2 );
		}

		for( int i = 0; i < 2; ++i ) {
			double xnp1 = solver.mCL[ i + n2 ];;
			for( int j2 = i, m = i + n2; j2 < m; j2 += 2 ) {
				xnp1 += solver.mCLX[ j2 ];
				solver.mDXNP1[ j2 ] = solver.mCL[ j2 ];
			}
			xxx[ xnp1_off + i ] = xnp1;
		}

		// ******************  BEGIN BLOCK A  *******************

		// "BLOCK A" TAKES  X  AND  XNP1  AS INPUT AND RETURNS  F
		// AND  DF  AS OUTPUT.   F  IS THE HOMOGENEOUS FORM OF THE
		// ORIGINAL  F, AND  DF  CONSISTS OF THE PARTIAL
		// DERIVATIVES OF THE HOMOGENEOUS FORM OF  F  WITH RESPECT
		// TO THE N+1 VARIABLES X(1), ... ,X(N), XNP1.

		// BEGIN "COMPUTE F"

		final int n_n = n * n;
		final int kstep = n_n + n;
		final int kstep2 = kstep + kstep;
		for( int j = 0; j < n; ++j ) {
			final int m = n2 * np1 * solver.mNumberOfTerms[ j ];
			for( int k = 0, ki = j; k < m; k += kstep2, ki += kstep ) {
				final int xi = xxx_off + k + (j << 1);
				powp( solver.mVarDegrees[ ki + n_n ],  xxx, xnp1_off,  xxx, xi + (n_n << 1) );
				for( int l = 0, nl = 0; l < n2; l += 2, nl += n ) {
					powp( solver.mVarDegrees[ ki + nl ],  x, xOff + l,  xxx, xi + (nl << 1) );
				}
			}
		}
		for( int j = 0; j < n; ++j ) {
			final int term_n = n * solver.mNumberOfTerms[ j ];
			for( int kn = 0; kn < term_n; kn += n ) {
				final int kn_j = kn + j;
				final int tri = kn_j + kn_j;
				solver.mTRM[ tri    ] = solver.mCoefficients[ kn_j ];
				solver.mTRM[ tri + 1] = 0.;
				for( int xi = xxx_off + tri + n2 * kn, xe = xi + kstep2; xi < xe; xi += n2 ) {
					mulp( xxx, xi, solver.mTRM, tri, xxx, temp1_off );
					solver.mTRM[ tri     ] = xxx[ temp1_off ];
					solver.mTRM[ tri + 1 ] = xxx[ temp1_off + 1 ];
				}
			}
		}
		for( int j = 0; j < n; ++j ) {
			final int j2 = j + j;
			double re = 0.;
			double im = 0.;
			for( int k = j2, t = k + n2 * solver.mNumberOfTerms[ j ]; k < t; k += n2 ) {
				re += solver.mTRM[ k     ];
				im += solver.mTRM[ k + 1 ];
			}
			solver.mF[ j2     ] = re;
			solver.mF[ j2 + 1 ] = im;
		}

		// END OF "COMPUTE F"

		// BEGIN "COMPUTE DF"

		for( int j = 0; j < n; ++j ) {
			final int term = solver.mNumberOfTerms[ j ];
			for( int k = 0, kn = 0; k < term; ++k, kn += n ) {
				for( int m = 0; m <= n; ++m ) {

					// IF TERM DOES NOT INCLUDE X(M), SET PARTIAL DERIVATIVE OF TERM
					//   EQUAL TO ZERO.
					final int kdi = n * (kn + k + m) + j;
					final int dti = kdi << 1;
					if( solver.mVarDegrees[ kdi ] == 0 ) {
						solver.mDTRM[ dti     ] = 0.;
						solver.mDTRM[ dti + 1 ] = 0.;
					} else {

						// IF TERM DOES INCLUDE X(M), TRY COMPUTING THE PARTIAL BY DIVIDING
						//   THE TERM BY X(M).
						final int tri = ((kn + j) << 1);
						boolean ierr = false;// java to fix "The local variable ierr may not have been initialized"
						if( m < n ) {
							ierr = divp( solver.mTRM, tri,  x, xOff + (m << 1),  solver.mDTRM, dti );
						}
						if( m == n ) {
							ierr = divp( solver.mTRM, tri,  xxx, xnp1_off,  solver.mDTRM, dti );
						}
						if( ierr ) {
							// IF DIVISION WOULD CAUSE OVERFLOW, GENERATE THE PARTIAL BY
							//   THE POLYNOMIAL FORMULA.
							solver.mDTRM[ dti ] = solver.mCoefficients[ kn + j ];
							solver.mDTRM[ dti + 1 ] = 0.;
							for( int l = 0; l < np1; ++l ) {
								if( l != m ) {
									mulp( xxx, xxx_off + ((n * (kn + l + k) + j) << 1),
											solver.mDTRM, dti, xxx, temp1_off );
									solver.mDTRM[ dti ] = xxx[ temp1_off ];
									solver.mDTRM[ dti + 1 ] = xxx[ temp1_off + 1 ];
								}
							}
							final int nnnn = solver.mVarDegrees[ kdi ] - 1;
							if( m < n ) {
								powp( nnnn, x, xOff + (m << 1), xxx, temp2_off );
							}
							if( m == n ) {
								powp( nnnn, xxx, xnp1_off, xxx, temp2_off );
							}
							mulp( xxx, temp2_off, xxx, temp1_off, solver.mDTRM, dti );

						}
						solver.mDTRM[ dti     ] = solver.mVarDegrees[ kdi ] * solver.mDTRM[ dti ];
						solver.mDTRM[ dti + 1 ] = solver.mVarDegrees[ kdi ] * solver.mDTRM[ dti + 1 ];
					}
				}
			}
		}
		for( int j = 0; j < n; ++j ) {
			for( int m = j << 1, me = m + kstep2; m < me; m += n2 ) {
				double re = 0.;
				double im = 0.;
				final int term = kstep2 * solver.mNumberOfTerms[ j ];
				for( int k = 0, dti = m; k < term; k += kstep2, dti += kstep2 ) {
					re += solver.mDTRM[ dti ];
					im += solver.mDTRM[ dti + 1 ];
				}
				solver.mDF[ m     ] = re;
				solver.mDF[ m + 1 ] = im;
			}
		}

		// END OF "COMPUTE DF"
		// *******************  END BLOCK A  ********************

		// CONVERT  DF  TO BE PARTIALS WITH RESPECT TO  X(1), ... ,X(N),
		// BY APPLYING THE CHAIN RULE WITH  XNP1  CONSIDERED A FUNCTION OF
		// OF  X(1), ... ,X(N).

		for( int j = 0, dfi = (n_n << 1); j < n2; j += 2, dfi += 2 ) {
			for( int k = 0, doff = j; k < n2; k += 2, doff += n2 ) {
				mulp( solver.mDF, dfi,  solver.mDXNP1, k,  xxx, temp1_off );
				solver.mDF[ doff     ] += xxx[ temp1_off ];
				solver.mDF[ doff + 1 ] += xxx[ temp1_off + 1 ];
			}
		}
	} /* ffunp */

	/**
	 * GFUNP  evaluates the start equation "G".
	 *
	 * ON INPUT:
	 *
	 * N  is the number of variables.
	 *
	 * IDEG(J)  is the degree of the J-th equation.
	 *
	 * PDG(1,J), PDG(2,J)  are the real and imaginary parts
	 *   of the powers of P used to define G.
	 *
	 * QDG(1,J), QDG(2,J)  are the real and imaginary parts
	 *   of the powers of Q used to define G.
	 *
	 * @param x X(1,J), X(2,J)  are the real and imaginary parts of the
	 *   J-th independent variable.
	 *
	 * XDGM1 workspace array
	 * XDG workspace array
	 * PXDGM1 workspace array
	 * PXDG workspace array
	 *
	 * ON OUTPUT:
	 *
	 * N,IDEG,PDG,QDG, and X  are unchanged.
	 *
	 * G(1,J), G(2,J)  are the real and imaginary parts of the
	 *   J-th start equation.
	 *
	 * DG(1,J),DG(2,J)  are the real and imaginary parts of the
	 *   partial derivatives of the J-th start equation with respect to the
	 *   J-th independent variable.
	 */
	private static void gfunp(final PolsysSolver solver, final double[] x, final int xOff) {
		final int n = solver.mVarAndEqNumber;

		// COMPUTE THE (IDEG-1)-TH AND IDEG-TH POWER OF X

		for( int j = 0; j < n; ++j ) {
			final int j2 = j + j;
			powp( solver.mDegrees[ j ] - 1, x, xOff + j2, solver.mXDGM1, j2 );
			mulp( x, xOff + j2, solver.mXDGM1, j2, solver.mXDG, j2 );
		}

		// COMPUTE THE PRODUCT OF PDG AND XDGM1
		final int n2 = n + n;
		for( int j2 = 0; j2 < n2; j2 += 2 ) {
			mulp( solver.mPDG, j2, solver.mXDGM1, j2, solver.mPXDGM1, j2 );
		}

		// COMPUTE THE PRODUCT OF PDG AND XDG
		for( int j2 = 0; j2 < n2; j2 += 2 ) {
			mulp( solver.mPDG, j2, solver.mXDG, j2, solver.mPXDG, j2 );
		}
		for( int j = 0; j < n; ++j ) {
			int j2 = j + j;
			solver.mG[ j2 ] = solver.mPXDG[ j2 ] - solver.mQDG[ j2 ];
			solver.mDG[ j2 ] = (double) solver.mDegrees[ j ] * solver.mPXDGM1[ j2 ];
			++j2;
			solver.mG[ j2 ] = solver.mPXDG[ j2 ] - solver.mQDG[ j2 ];
			solver.mDG[ j2 ] = (double) solver.mDegrees[ j ] * solver.mPXDGM1[ j2 ];
		}
	} /* gfunp */

	/**
	 *  HFUN1P  evaluates the continuation equation "RHO".
	 *
	 *  NOTE THAT:
	 *    DRHOX is the "realification" of DCRHOX, where
	 *    DCRHOX denotes the (complex) partial
	 *    derivative matrix of the continuation system
	 *    with respect to X,  and
	 *    DRHOL is the "realification" of DCRHOL, where
	 *    DCRHOL denotes the (complex) partial
	 *    derivative matrix of the continuation system
	 *    with respect to LAMBDA. Thus
	 *      DRHOX(2J-1,2K-1) = DCRHOX(1,J,K)
	 *      DRHOX(2J  ,2K  ) = DCRHOX(1,J,K)
	 *      DRHOX(2J-1,2K  ) =-DCRHOX(2,J,K)
	 *      DRHOX(2J  ,2K-1) = DCRHOX(2,J,K)
	 *      DRHOL(2J-1,N2P1) = DCRHOL(1,J)
	 *      DRHOL(2J  ,N2P1) = DCRHOL(2,J)
	 *       RHO(2J-1)      = CRHO(1,J)
	 *       RHO(2J  )      = CRHO(2,J)
	 *    where CRHO denotes the (complex) continuation system,
	 *    the initial "1" or "2" denotes real or imaginary parts,
	 *    respectively, "J" indexes the equation, "K" indexes the partial
	 *    derivative, and neither DCRHOX nor DCRHOL are program variables.
	 *
	 *  ON INPUT:
	 *
	 * QDG  is the "random" parameter "A".
	 *
	 * @param lambda   LAMBDA  is the continuation parameter.
	 *
	 * @param x   X    is the independent variable.
	 *
	 * PDG  is one of the parameters that defines G (see subroutine
	 *         GFUNP).
	 *
	 * CL   is one of the parameters that defines F (see subroutine
	 *         FFUNP).
	 *
	 * COEF  is one of the parameters that defines F (see subroutine
	 *         FFUNP).
	 *
	 *  ON OUTPUT:
	 *
	 * RHO    is the homotopy.
	 *
	 * DRHOX  contains the partial derivatives of RHO with respect
	 *         to X.
	 *
	 * DRHOL  contains the partial derivatives of RHO with respect
	 *         to LAMBDA.
	 *
	 *  The following are variables whose workspace is passed from HFUNP:
	 *    XDGM1, XDG, G, DG, PXDGM1, PXDG, F, DF, XX, TRM, DTRM, CLX
	 *    DXNP1, N, MMAXT, IDEG, NUMT, KDEG
	 *
	 *  Other variables:
	 *    ONEML
	 */
	private static void hfun1p(final PolsysSolver solver, final double lambda, final double[] x, final int xOff) {
		final int nn = solver.mVarAndEqNumber;// n[ nOff ];

		gfunp( solver, x, xOff );
		ffunp( solver, x, xOff );

		final double oneml = 1. - lambda;
		final int nn2 = nn + nn;
		final int ke = nn2 * nn;
		final double[] drhox = solver.mDRHOX;
		final double[] df = solver.mDF;
		for( int j2 = 0; j2 < nn2; j2 += 2 ) {
			for( int kn2 = 0; kn2 < ke; kn2 += nn2 ) {
				int idx = kn2 + kn2 + j2;
				int idr = kn2 + j2;
				drhox[ idx     ] = lambda * df[ idr     ];
				drhox[ idx + 1 ] = lambda * df[ idr + 1 ];
				idr = kn2 + kn2 + j2;
				idx = idr + nn2;
				drhox[ idx + 1 ] =  drhox[ idr     ];
				drhox[ idx     ] = -drhox[ idr + 1 ];
			}
			final int idx = j2 * nn2 + j2;
			drhox[ idx     ] += oneml * solver.mDG[ j2     ];
			drhox[ idx + 1 ] += oneml * solver.mDG[ j2 + 1 ];;
			final int idr = idx + nn2;
			drhox[ idr + 1 ] =  drhox[ idx     ];
			drhox[ idr     ] = -drhox[ idx + 1 ];
			double fv = solver.mF[ j2 ];
			double gv = solver.mG[ j2 ];
			solver.mDRHOL[ j2   ] = fv - gv;
			solver.mRHO[ j2     ] = lambda * fv + oneml * gv;
			fv = solver.mF[ j2 + 1 ];
			gv = solver.mG[ j2 + 1 ];
			solver.mDRHOL[ j2 + 1 ] = fv - gv;
			solver.mRHO[ j2 + 1   ] = lambda * fv + oneml * gv;
		}
	} /* hfun1p */

	/**
	 * PAR(1:*) and IPAR(1:*) are arrays for (optional) user parameters,
	 *    which are simply passed through to the user written subroutines
	 *    RHO, RHOJAC.
	 *
	 * Evaluate  RHO(A,LAMBDA,X)  and return in the vector  V .
	 */
	private static void rho(final PolsysSolver solver,
			final double[] a,
			double lambda,
			final double[] x, final int xOff,
			final double[] v, final int vOff
		)
	{
		// THE FOLLOWING CODE IS SPECIFICALLY FOR THE POLYNOMIAL SYSTEM DRIVER
		//  POLSYS , AND SHOULD BE USED VERBATUM WITH  POLSYS .  IF THE USER IS
		// CALLING  FIXP??  OR   STEP??  DIRECTLY, HE MUST SUPPLY APPROPRIATE
		// REPLACEMENT CODE HERE.
		// FORCE PREDICTED POINT TO HAVE  LAMBDA >= 0  .

		if( lambda < 0. ) {
			lambda = 0.;
		}
		hfun1p( solver, lambda, x, xOff );
		final int n2 = solver.mVarAndEqNumber << 1;
		System.arraycopy( solver.mRHO, 0, v, vOff, n2 );
	} /* rho */

	/**
	 * PAR(1:*) AND IPAR(1:*) are arrays for (optional) user parameters,
	 *    which are simply passed through to the user written subroutines
	 *    RHO, RHOJAC.
	 *
	 * Return in the vector  V  the Kth column of the Jacobian
	 * matrix [D RHO/D LAMBDA, D RHO/DX] evaluated at the point
	 * (A, LAMBDA, X).
	 *
	 * @param lambda lambda, the point parameter
	 * @param x X, the point parameter
	 * @param v the vector V
	 * @param k the k-th column, starts from 0
	 */
	private static void rhojac(final PolsysSolver solver,
				double lambda,
				final double[] x, final int xOff,
				final double[] v, final int vOff,
				final int k
		 )
	{
		// THE FOLLOWING CODE IS SPECIFICALLY FOR THE POLYNOMIAL SYSTEM DRIVER
		//  POLSYS , AND SHOULD BE USED VERBATUM WITH  POLSYS .  IF THE USER IS
		// CALLING  FIXP??  OR   STEP??  DIRECTLY, HE MUST SUPPLY APPROPRIATE
		// REPLACEMENT CODE HERE.

		final int n2 = solver.mVarAndEqNumber << 1;
		if( k == 0 ) {
			// FORCE PREDICTED POINT TO HAVE  LAMBDA >= 0  .
			if( lambda < 0. ) {
				lambda = 0.;
			}
			hfun1p( solver, lambda, x, xOff );

			System.arraycopy( solver.mDRHOL, 0, v, vOff, n2 );
			return;
		}
		System.arraycopy( solver.mDRHOX, n2 * (k - 1), v, vOff, n2 );
	} /* rhojac */

	/**
	 * This subroutine builds the Jacobian matrix of the homotopy map,
	 * computes a QR decomposition of that matrix, and then calculates the
	 * (unit) tangent vector and the Newton step.
	 *
	 * ON INPUT:
	 *
	 * RHOLEN < 0 if the norm of the homotopy map evaluated at
	 *    (A, LAMBDA, X) is to be computed.  If  RHOLEN >= 0  the NORM is not
	 *    computed and  RHOLEN  is not changed.
	 *
	 * Y(1:N+1) = current point (LAMBDA(S), X(S)).
	 *
	 * YPOLD(1:N+1) = unit tangent vector at previous point on the zero
	 *    curve of the homotopy map.
	 *
	 * QGG A(1:*) = parameter vector in the homotopy map.
	 *
	 * QR(1:N,1:N+2), ALPHA(1:N), TZ(1:N+1), PIVOT(1:N+1)  are work arrays
	 *    used for the QR factorization.
	 *
	 * NFE = number of Jacobian matrix evaluations = number of homotopy
	 *    function evaluations.
	 *
	 * N = dimension of X.
	 *
	 * IFLAG = -2, -1, or 0, indicating the problem type.
	 *
	 * PAR(1:*) and IPAR(1:*) are arrays for (optional) user parameters,
	 *    which are simply passed through to the user written subroutines
	 *    RHO, RHOJAC.
	 *
	 * ON OUTPUT:
	 *
	 * RHOLEN = ||RHO(A, LAMBDA(S), X(S)|| IF  RHOLEN < 0  on input.
	 *    otherwise  RHOLEN  is unchanged.
	 *
	 * Y, YPOLD, A, N  are unchanged.
	 *
	 * YP(1:N+1) = DY/DS = unit tangent vector to integral curve of
	 *    D(HOMOTOPY MAP)/DS = 0  at  Y(S) = (LAMBDA(S), X(S)) .
	 *
	 * TZ = the Newton step = -(pseudo inverse of  (D RHO(A,Y(S))/D LAMBDA ,
	 *    D RHO(A,Y(S))/DX)) * RHO(A,Y(S)) .
	 *
	 * NFE  has been incremented by 1.
	 *
	 * IFLAG  is unchanged, unless the QR factorization detects a RANK < N,
	 *    in which case the tangent and Newton step vectors are not computed
	 *    and  TANGNF  returns with  IFLAG = 4 .
	 */
	private static void tangnf(final PolsysSolver solver) {
		final int n = solver.mVarAndEqNumber << 1;

		final double lambda = solver.mTangentY[ 0 ];
		final int np1 = n + 1;
		++solver.mNumberOfMatrixEvaluations;

		// NFE CONTAINS THE NUMBER OF JACOBIAN EVALUATIONS.
		//   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *

		// COMPUTE THE JACOBIAN MATRIX, STORE IT AND HOMOTOPY MAP IN QR.

		if( solver.mFlag == FOLLOW_PATH ) {

			// QR = ( D RHO(A,LAMBDA,X)/D LAMBDA , D RHO(A,LAMBDA,X)/DX ,
			//                                             RHO(A,LAMBDA,X) )  .

			int kn = 0;
			for( int k = 0; k < np1; ++k, kn += n ) {
				rhojac( solver, lambda,  solver.mTangentY, 1, solver.mQR, kn, k );
			}
			rho( solver, solver.mQDG, lambda, solver.mTangentY, 1, solver.mQR, kn );
		} else {
			f( solver.mTangentY, 1, solver.mTZ, 0 );
			final int n_np1 = n * np1;
			if( solver.mFlag == 0 ) {

				// QR = ( A - F(X), I - LAMBDA*DF(X) ,
				//                         X - A + LAMBDA*(A - F(X)) )  .

				for( int j = 0; j < n; ++j ) {
					final double sigma = solver.mQDG[ j ];
					final double beta = sigma - solver.mTZ[ j ];
					solver.mQR[ j ] = beta;
					solver.mQR[ j + n_np1 ] = solver.mTangentY[ j + 1 ] - sigma + lambda * beta;
				}
				for( int k = 0, kn = n; k < n; ++k, kn += n ) {
					fjac( solver.mTangentY, 1, solver.mTZ, 0, k );
					for( int j = 0; j < n; ++j ) {
						solver.mQR[ j + kn ] = -lambda * solver.mTZ[ j ];
					}
					solver.mQR[ k + kn ] += 1.;
				}
			} else {

				// QR = ( F(X) - X + A, LAMBDA*DF(X) + (1 - LAMBDA)*I ,
				//                              X - A + LAMBDA*(F(X) - X + A) )  .
				for( int j = 0; j < n; ++j ) {
					final double sigma = solver.mTangentY[ j + 1 ] - solver.mQDG[ j ];
					final double beta = solver.mTZ[ j ] - sigma;
					solver.mQR[ j ] = beta;

					solver.mQR[ j + n_np1 ] = sigma + lambda * beta;
				}
				for( int k = 0, qi = n; k < n; ++k, qi += np1 ) {
					fjac( solver.mTangentY, 1, solver.mTZ, 0, k );
					for( int j = 0, kn = n; j < n; ++j, kn += n ) {
						solver.mQR[ j + kn ] = lambda * solver.mTZ[ j ];
					}
					solver.mQR[ qi ] = 1. - lambda + solver.mQR[ qi ];
				}
			}
		}

		//   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *   *
		// COMPUTE THE NORM OF THE HOMOTOPY MAP IF IT WAS REQUESTED.
		if( solver.mTangentRhoLength < 0. ) {
			solver.mTangentRhoLength = dnrm2( n, solver.mQR, n * np1, 1 );
		}

		// REDUCE THE JACOBIAN MATRIX TO UPPER TRIANGULAR FORM.

		// THE FOLLOWING CODE IS A MODIFICATION OF THE ALGOL PROCEDURE
		// DECOMPOSE  IN P. BUSINGER AND G. H. GOLUB, LINEAR LEAST
		// SQUARES SOLUTIONS BY HOUSEHOLDER TRANSFORMATIONS,
		// NUMER. MATH. 7 (1965) 269-276.

		for( int j = 0, qi = 0; j < np1; ++j, qi += n ) {
			solver.mTangentYp[ j ] = ddot( n, solver.mQR, qi, 1, solver.mQR, qi, 1 );
			solver.mPivots[ j ] = j;
		}
		final int nm1 = n - 1;
		for( int k = 0, kn = 0; k < n; ++k, kn += n ) {
			double sigma = solver.mTangentYp[ k ];
			int jbar = k;
			for( int j = k + 1; j < np1; ++j ) {
				if( sigma < solver.mTangentYp[ j ] ) {
					sigma = solver.mTangentYp[ j ];
					jbar = j;
				}
			}
			if( jbar != k ) {
				int i = solver.mPivots[ k ];
				solver.mPivots[ k ] = solver.mPivots[ jbar ];
				solver.mPivots[ jbar ] = i;
				solver.mTangentYp[ jbar ] = solver.mTangentYp[ k ];
				solver.mTangentYp[ k ] = sigma;
				jbar *= n;
				i = kn;
				for( final int ie = i + n; i < ie; ++i, ++jbar ) {
					sigma = solver.mQR[ i ];
					solver.mQR[ i ] = solver.mQR[ jbar ];
					solver.mQR[ jbar ] = sigma;
				}
				// END OF COLUMN INTERCHANGE.
			}
			final int qi = kn + k;
			sigma = ddot( n - k, solver.mQR, qi, 1, solver.mQR, qi, 1 );
			if( sigma == 0. ) {
				solver.mFlag = ERR_ALGORITHM_FAILED;
				return;
			}
			if( k != nm1 ) {
				final double qrkk = solver.mQR[ qi ];
				double alphak = -Math.sqrt( sigma );
				if( qrkk < 0. ) {
					alphak = -alphak;
				}
				solver.mAlpha[ k ] = alphak;
				final double beta = 1. / (sigma - qrkk * alphak);
				solver.mQR[ qi ] = qrkk - alphak;
				for( int j = k, qri = n + j * n; j < np1; ++j, qri += n ) {
					sigma = beta * ddot( n - k, solver.mQR, qi, 1, solver.mQR, qri + k, 1 );
					for( int i = k; i < n; ++i ) {
						solver.mQR[ i + qri ] -= solver.mQR[ i + kn ] * sigma;
					}
					if( j < n ) {
						/* Computing 2nd power */
						final double d1 = solver.mQR[ qri + k ];
						solver.mTangentYp[ j + 1 ] -= d1 * d1;
					}
				}
			}
		}
		solver.mAlpha[ nm1 ] = solver.mQR[ n * n - 1 ];

		// COMPUTE KERNEL OF JACOBIAN, WHICH SPECIFIES YP=DY/DS.
		solver.mTZ[ n ] = 1.;
		for( int i = n; --i >= 0; ) {
			double sum = 0.;
			for( int j = i + 1, qi = i + j * n; j < np1; ++j, qi += n ) {
				sum += solver.mQR[ qi ] * solver.mTZ[ j ];
			}
			solver.mTZ[ i ] = -sum / solver.mAlpha[ i ];
		}
		final double ypnorm = dnrm2( np1, solver.mTZ, 0, 1 );
		for( int k = 0; k < np1; ++k ) {
			solver.mTangentYp[ solver.mPivots[ k ] ] = solver.mTZ[ k ] / ypnorm;
		}
		if( ddot( np1, solver.mTangentYp, 0, 1, solver.mTangentYpold, 0, 1 ) < 0. ) {
			for( int yi = 0, ye = yi + np1; yi < ye; ++yi ) {
				solver.mTangentYp[ yi ] = -solver.mTangentYp[ yi ];
			}
			// YP  IS THE UNIT TANGENT VECTOR IN THE CORRECT DIRECTION.
		}

		// COMPUTE THE MINIMUM NORM SOLUTION OF [D RHO(Y(S))] V = -RHO(Y(S)).
		// V IS GIVEN BY  P - (P,Q)Q  , WHERE P IS ANY SOLUTION OF
		// [D RHO] V = -RHO  AND Q IS A UNIT VECTOR IN THE KERNEL OF [D RHO].

		for( int i = n, n_n = n * n; --i >= 0; ) {
			double sum = solver.mQR[ i + n_n ] + solver.mQR[ i + n_n + n ];
			for( int j = i + 1, qi = i + j * n; j < n; ++j, qi += n ) {
				sum += solver.mQR[ qi ] * solver.mAlpha[ j ];
			}
			solver.mAlpha[ i ] = -sum / solver.mAlpha[ i ];
		}
		for( int k = 0; k < n; ++k ) {
			solver.mTZ[ solver.mPivots[ k ] ] = solver.mAlpha[ k ];
		}
		solver.mTZ[ solver.mPivots[ n ] ] = 1.;
		// TZ NOW CONTAINS A PARTICULAR SOLUTION P, AND YP CONTAINS A VECTOR Q
		// IN THE KERNEL(THE TANGENT).
		final double sigma = ddot( np1, solver.mTZ, 0, 1, solver.mTangentYp, 0, 1 );
		for( int j = 0; j < np1; ++j ) {
			solver.mTZ[ j ] -= sigma * solver.mTangentYp[ j ];
		}
		// TZ IS THE NEWTON STEP FROM THE CURRENT POINT Y(S) = (LAMBDA(S), X(S)).
		return;
	} /* tangnf */

	/** Max function evaluations, 500 */
	private static final int MAX_FUNCTION_EVALUATIONS = 500;

	/** = 1  if F(B)*F(C) < 0 and the stopping criterion is met. */
	private static final int ROOT_STOPPING_CRITERION_MET = 1;

	/** = 2  if a value b is found such that the computed value
	 *      F(B) is exactly zero.  The interval (B,C) may not
	 *      satisfy the stopping criterion. */
	private static final int ROOT_F_B_ZERO = 2;

	/** = 3  if DABS(F(B)) exceeds the input values DABS(F(B)),
	 *      DABS(F(C)).  In this case it is likely that B is close
	 *      to A pole of F. */
	private static final int ROOT_DABS_F_B_EXCEED_INPUT = 3;

	/** = 4  if no odd order root was found in the interval.  A
	 *      local minimum may have been obtained. */
	private static final int ROOT_NO_ODD_ROOT_FOUND = 4;

	/** = 5  if too many function evaluations were made.
	 *      (as programmed, 500 are allowed.) */
	private static final int ROOT_TOO_MANY_EVALUATIONS = 5;

	/** =-3 have completed computation for new iterate B */
	private static final int HAVE_COMPLETED_COMPUTATION_FOR_NEW_ITERATE_B = -3;

	/**
	 *  Root computes a root of the nonlinear equation F(X)=0
	 *  where F(X) is a continuous real function of a single real
	 *  variable X.  The method used is a combination of bisection
	 *  and the secant rule.
	 *
	 *  Normal input consists of a continuous function F and
	 *  interval (B,C) such that F(B)*F(C) <= 0.0.  Each iteration
	 *  finds new values of B and C such that the interval (B,C) is
	 *  shrunk and F(B)*F(C) <= 0.0.  The stopping criterion is
	 * <pre>
	 *          DABS(B-C) <= 2.0*(RELERR*DABS(B)+ABSERR)
	 * </pre>
	 *  where RELERR=relative error and ABSERR=absolute error are
	 *  input quantities.  Set the FLAG, IFLAG, positive to initialize
	 *  the computation.  As B, C and IFLAG are used for both input and
	 *  output, they must be variables in the calling program.
	 *
	 *  If 0 is a possible root, one should not choose ABSERR=0.0.
	 *
	 *  The output value of B is the better approximation to a root
	 *  as B and C are always redefined so that DABS(F(B)) <= DABS(F(C)).
	 *
	 *  To solve the equation, root must evaluate F(X) repeatedly. This
	 *  is done in the calling program.  When an evaluation of F is
	 *  needed at T, root returns with IFLAG negative.  Evaluate FT=F(T)
	 *  and call root again.  Do not alter IFLAG.
	 *
	 *  When the computation is complete, root returns to the calling
	 *  program with IFLAG positive=
	 *
	 *     IFLAG=1  if F(B)*F(C) < 0 and the stopping criterion is met.
	 *
	 *          =2  if a value b is found such that the computed value
	 *              F(B) is exactly zero.  The interval (B,C) may not
	 *              satisfy the stopping criterion.
	 *
	 *          =3  if DABS(F(B)) exceeds the input values DABS(F(B)),
	 *              DABS(F(C)).  In this case it is likely that B is close
	 *              to A pole of F.
	 *
	 *          =4  if no odd order root was found in the interval.  A
	 *              local minimum may have been obtained.
	 *
	 *          =5  if too many function evaluations were made.
	 *              (as programmed, 500 are allowed.)
	 *
	 *  This code is a modification of the code zeroin which is completely
	 *  explained and documented in the text  numerical computing:  an
	 *  introduction,  by L. F. Shampine and R. C. Allen.
	 */
	/* extracted in place private static int root(final PolsysSolver solver,
			final double ft,
			final double relativeError, final double absoluteError,
			int iflag)
	{
		if( iflag >= 0 ) {// cannot be < -3, if( (iflag >= 0) || (iflag < -3) ) {
			solver.mRootRelError = Math.max( relativeError, MACHINE_PRECISION );
			solver.mRootAbsError = Math.max( absoluteError, 0. );
			solver.mRootIterationCount = 0;
			solver.mRootAbsAB = Math.abs( solver.mRootSa - solver.mRootSb );// b - c
			solver.mRootA = solver.mRootSb;// = c
			solver.mRootSout = solver.mRootA;// t = a
			return -ROOT_STOPPING_CRITERION_MET;
		}
		iflag = Math.abs( iflag );
		if( iflag == ROOT_STOPPING_CRITERION_MET ) {
			solver.mRootFa = ft;
			solver.mRootSout = solver.mRootSa;// t = b
			return -ROOT_F_B_ZERO;
		}
		if( iflag == ROOT_F_B_ZERO ) {
			solver.mRootFb = ft;
			solver.mRootFc = solver.mRootFa;
			solver.mRootCount = 2;
			// Computing MAX
			solver.mRootFx = Math.max( Math.abs( solver.mRootFb ), Math.abs( solver.mRootFc ) );
		}
		if( iflag == ROOT_DABS_F_B_EXCEED_INPUT ) {
			solver.mRootFb = ft;
			if( solver.mRootFb == 0. ) {
				return ROOT_F_B_ZERO;
			}
			++solver.mRootCount;
			if( Math.copySign( 1., solver.mRootFb ) == Math.copySign( 1., solver.mRootFc ) ) {
				solver.mRootSb = solver.mRootA;// c  = a
				solver.mRootFc = solver.mRootFa;
			}
		}

		if( Math.abs( solver.mRootFc ) < Math.abs( solver.mRootFb ) ) {
			// INTERCHANGE B AND C SO THAT ABS(F(B)) <= ABS(F(C)).

			solver.mRootA = solver.mRootSa;// a = b;
			solver.mRootFa = solver.mRootFb;
			solver.mRootSa = solver.mRootSb;// b = c;
			solver.mRootFb = solver.mRootFc;
			solver.mRootSb = solver.mRootA;// c =
			solver.mRootFc = solver.mRootFa;
		}
		final double cmb = 0.5 * (solver.mRootSb - solver.mRootSa);// (c - b) * 0.5;
		final double acmb = Math.abs( cmb );
		final double tol = solver.mRootRelError * Math.abs( solver.mRootSa ) + solver.mRootAbsError;// ( b )

		// TEST STOPPING CRITERION AND FUNCTION COUNT.

		if( acmb <= tol ) {
			// FINISHED.  SET IFLAG.
			if( Math.copySign( 1., solver.mRootFb ) == Math.copySign( 1., solver.mRootFc ) ) {
				return ROOT_NO_ODD_ROOT_FOUND;
			}
			if( Math.abs( solver.mRootFb ) > solver.mRootFx ) {
				return ROOT_DABS_F_B_EXCEED_INPUT;
			}
			return ROOT_STOPPING_CRITERION_MET;
		}
		if( solver.mRootCount >= MAX_FUNCTION_EVALUATIONS ) {
			return ROOT_TOO_MANY_EVALUATIONS;
		}

		// CALCULATE NEW ITERATE EXPLICITLY AS B+P/Q
		// WHERE WE ARRANGE P.GE.0.  THE IMPLICIT
		// FORM IS USED TO PREVENT OVERFLOW.

		double p = (solver.mRootSa - solver.mRootA) * solver.mRootFb;// (b - ...
		double q = solver.mRootFa - solver.mRootFb;
		if( p < 0. ) {
			p = -p;
			q = -q;
		}

		// UPDATE A, CHECK IF REDUCTION IN THE SIZE OF BRACKETING
		// INTERVAL IS SATISFACTORY.  IF NOT BISECT UNTIL IT IS.

		solver.mRootA = solver.mRootSa;// = b
		solver.mRootFa = solver.mRootFb;
		++solver.mRootIterationCount;
		if( solver.mRootIterationCount >= 4 ) {
			if( acmb * 8. >= solver.mRootAbsAB ) {
				// USE BISECTION.
				solver.mRootSa = 0.5 * (solver.mRootSa + solver.mRootSb);// b = (c[ 0 ] + b[ 0 ]) * 0.5;

				// HAVE COMPLETED COMPUTATION FOR NEW ITERATE B.
				solver.mRootSout = solver.mRootSa;// t = b
				return HAVE_COMPLETED_COMPUTATION_FOR_NEW_ITERATE_B;
			}
			solver.mRootIterationCount = 0;
			solver.mRootAbsAB = acmb;
		}

		// TEST FOR TOO SMALL A CHANGE.

		if( p > Math.abs( q ) * tol ) {
			// ROOT OUGHT TO BE BETWEEN B AND (C+B)/2
			if( p >= cmb * q ) {
				// USE BISECTION.
				solver.mRootSa = 0.5 * (solver.mRootSa + solver.mRootSb);// b = (c[ 0 ] + b[ 0 ]) * 0.5;
			} else {
				// USE SECANT RULE.
				solver.mRootSa += p / q;// b +=
			}

			// HAVE COMPLETED COMPUTATION FOR NEW ITERATE B.
			solver.mRootSout = solver.mRootSa;// t = b
			return HAVE_COMPLETED_COMPUTATION_FOR_NEW_ITERATE_B;
		}

		// INCREMENT BY TOLERANCE

		solver.mRootSa += Math.copySign( tol, cmb );// b += ...

		// HAVE COMPLETED COMPUTATION FOR NEW ITERATE B.
		solver.mRootSout = solver.mRootSa;// t = b
		return HAVE_COMPLETED_COMPUTATION_FOR_NEW_ITERATE_B;

	} */ /* root */

	/**
	 * ROOTNF  finds the point  YBAR = (1, XBAR)  on the zero curve of the
	 * homotopy map.  It starts with two points YOLD=(LAMBDAOLD,XOLD) and
	 * Y=(LAMBDA,X) such that  LAMBDAOLD < 1 <= LAMBDA , and alternates
	 * between hermite cubic interpolation and Newton iteration until
	 * convergence.
	 *
	 * ON INPUT:
	 *
	 * N = dimension of X and the homotopy map.
	 *
	 * NFE = number of Jacobian matrix evaluations.
	 *
	 * IFLAG = -2, -1, or 0, indicating the problem type.
	 *
	 * RELERR relative and absolute error value
	 * ABSERR absolute error value
	 *    The iteration is considered to have converged when
	 *    a point Y=(LAMBDA,X) is found such that
	 * <pre>
	 *    |Y(1) - 1| <= RELERR + ABSERR              and
	 *
	 *    ||Z|| <= RELERR*||X|| + ABSERR  ,          where
	 *
	 *    (?,Z) is the Newton step to Y=(LAMBDA,X).
	 * </pre>
	 * Y(1:N+1) = point (LAMBDA(S), X(S)) on zero curve of homotopy map.
	 *
	 * YP(1:N+1) = unit tangent vector to the zero curve of the homotopy map
	 *    at  Y .
	 *
	 * YOLD(1:N+1) = a point different from  Y  on the zero curve.
	 *
	 * YPOLD(1:N+1) = unit tangent vector to the zero curve of the homotopy
	 *    map at  YOLD .
	 *
	 * A(1:*) = parameter vector in the homotopy map.
	 *
	 * QR(1:N,1:N+2), ALPHA(1:N), TZ(1:N+1), PIVOT(1:N+1), W(1:N+1),
	 *    WP(1:N+1)  are work arrays used for the QR factorization (in the
	 *    Newton step calculation) and the interpolation.
	 *
	 * PAR(1:*) AND IPAR(1:*) are arrays for (optional) user parameters,
	 *    which are simply passed through to the user written subroutines
	 *    RHO, RHOJAC.
	 *
	 * ON OUTPUT:
	 *
	 * N , RELERR , ABSERR , A  are unchanged.
	 *
	 * NFE  has been updated.
	 *
	 * IFLAG
	 *    = -2, -1, or 0 (unchanged) on a normal return.
	 *
	 *    = 4 if a Jacobian matrix with RANK < N has occurred.  The
	 *        iteration was not completed.
	 *
	 *    = 6 if the iteration failed to converge.  Y  and  YOLD  contain
	 *        the last two points found on the zero curve.
	 *
	 * Y  is the point on the zero curve of the homotopy map at  LAMBDA = 1 .
	 */
	private static void rootnf(final PolsysSolver solver) {
		// java: initial set to avoid "The local variable qsout may not have been initialized"
		// qsout is used only when lcode is -1, -2, -3
		double qsout = 0.;// ft, f(t)

		final double relative_error = Math.max( solver.mAnswerRelError, MACHINE_PRECISION );
		final double absolute_error = Math.max( solver.mAnswerAbsError, 0. );
		final int n = solver.mVarAndEqNumber << 1;
		final int np1 = n + 1;

		//--- data for tangf
		solver.mTangentY = solver.mW;
		solver.mTangentYp = solver.mWP;
		solver.mTangentYpold = solver.mYpold;

		// root variables
		int root_iteration_count = 0, root_count = 0;
		double root_rel_error = 0., root_abs_error = 0.;
		double root_a = 0., root_fa = 0., root_fb = 0., root_fc = 0., root_fx = 0.;
		double root_abs_ab = 0., root_sout = 0.;

		// *****  MAIN LOOP.  *****

		for( int judy = 1; judy <= 20; ++judy ) {
			for( int jw = 0; jw < np1; ++jw ) {
				solver.mTZ[ jw ] = solver.mY[ jw ] - solver.mYold[ jw ];
			}
			final double dels = dnrm2( np1, solver.mTZ, 0, 1 );

			// USING TWO POINTS AND TANGENTS ON THE HOMOTOPY ZERO CURVE, CONSTRUCT
			// THE HERMITE CUBIC INTERPOLANT Q(S).  THEN USE  ROOT  TO FIND THE S
			// CORRESPONDING TO  LAMBDA = 1 .  THE TWO POINTS ON THE ZERO CURVE ARE
			// ALWAYS CHOSEN TO BRACKET LAMBDA=1, WITH THE BRACKETING INTERVAL
			// ALWAYS BEING [0, DELS].

			double root_sa = 0.;
			double root_sb = dels;

			int lcode = 1;
			do {
				// extracted here lcode = root( solver, qsout, relative_error, absolute_error, lcode );
				do {
					if( lcode >= 0 ) {// cannot be < -3, if( (iflag >= 0) || (iflag < -3) ) {
						root_rel_error = relative_error;
						root_abs_error = absolute_error;
						root_iteration_count = 0;
						root_abs_ab = Math.abs( root_sa - root_sb );// b - c
						root_a = root_sb;// = c
						root_sout = root_a;// t = a
						lcode = -ROOT_STOPPING_CRITERION_MET;
						break;
					}
					lcode = Math.abs( lcode );
					if( lcode == ROOT_STOPPING_CRITERION_MET ) {
						root_fa = qsout;
						root_sout = root_sa;// t = b
						lcode = -ROOT_F_B_ZERO;
						break;
					}
					if( lcode == ROOT_F_B_ZERO ) {
						root_fb = qsout;
						root_fc = root_fa;
						root_count = 2;
						/* Computing MAX */
						root_fx = Math.max( Math.abs( root_fb ), Math.abs( root_fc ) );
					}
					if( lcode == ROOT_DABS_F_B_EXCEED_INPUT ) {
						root_fb = qsout;
						if( root_fb == 0. ) {
							lcode = ROOT_F_B_ZERO;
							break;
						}
						++root_count;
						if( Math.copySign( 1., root_fb ) == Math.copySign( 1., root_fc ) ) {
							root_sb = root_a;// c  = a
							root_fc = root_fa;
						}
					}

					if( Math.abs( root_fc ) < Math.abs( root_fb ) ) {
						// INTERCHANGE B AND C SO THAT ABS(F(B)) <= ABS(F(C)).

						root_a = root_sa;// a = b;
						root_fa = root_fb;
						root_sa = root_sb;// b = c;
						root_fb = root_fc;
						root_sb = root_a;// c =
						root_fc = root_fa;
					}
					final double cmb = 0.5 * (root_sb - root_sa);// (c - b) * 0.5;
					final double acmb = Math.abs( cmb );
					final double tol = root_rel_error * Math.abs( root_sa ) + root_abs_error;// ( b )

					// TEST STOPPING CRITERION AND FUNCTION COUNT.

					if( acmb <= tol ) {
						// FINISHED.  SET IFLAG.
						if( Math.copySign( 1., root_fb ) == Math.copySign( 1., root_fc ) ) {
							lcode = ROOT_NO_ODD_ROOT_FOUND;
							break;
						}
						if( Math.abs( root_fb ) > root_fx ) {
							lcode = ROOT_DABS_F_B_EXCEED_INPUT;
							break;
						}
						lcode = ROOT_STOPPING_CRITERION_MET;
						break;
					}
					if( root_count >= MAX_FUNCTION_EVALUATIONS ) {
						lcode = ROOT_TOO_MANY_EVALUATIONS;
						break;
					}

					// CALCULATE NEW ITERATE EXPLICITLY AS B+P/Q
					// WHERE WE ARRANGE P.GE.0.  THE IMPLICIT
					// FORM IS USED TO PREVENT OVERFLOW.

					double p = (root_sa - root_a) * root_fb;// (b - ...
					double q = root_fa - root_fb;
					if( p < 0. ) {
						p = -p;
						q = -q;
					}

					// UPDATE A, CHECK IF REDUCTION IN THE SIZE OF BRACKETING
					// INTERVAL IS SATISFACTORY.  IF NOT BISECT UNTIL IT IS.

					root_a = root_sa;// = b
					root_fa = root_fb;
					++root_iteration_count;
					if( root_iteration_count >= 4 ) {
						if( acmb * 8. >= root_abs_ab ) {
							// USE BISECTION.
							root_sa = 0.5 * (root_sa + root_sb);// b = (c[ 0 ] + b[ 0 ]) * 0.5;

							// HAVE COMPLETED COMPUTATION FOR NEW ITERATE B.
							root_sout = root_sa;// t = b
							lcode = HAVE_COMPLETED_COMPUTATION_FOR_NEW_ITERATE_B;
							break;
						}
						root_iteration_count = 0;
						root_abs_ab = acmb;
					}

					// TEST FOR TOO SMALL A CHANGE.

					if( p > Math.abs( q ) * tol ) {
						// ROOT OUGHT TO BE BETWEEN B AND (C+B)/2
						if( p >= cmb * q ) {
							// USE BISECTION.
							root_sa = 0.5 * (root_sa + root_sb);// b = (c[ 0 ] + b[ 0 ]) * 0.5;
						} else {
							// USE SECANT RULE.
							root_sa += p / q;// b +=
						}

						// HAVE COMPLETED COMPUTATION FOR NEW ITERATE B.
						root_sout = root_sa;// t = b
						lcode = HAVE_COMPLETED_COMPUTATION_FOR_NEW_ITERATE_B;
						break;
					}

					// INCREMENT BY TOLERANCE

					root_sa += Math.copySign( tol, cmb );// b += ...

					// HAVE COMPLETED COMPUTATION FOR NEW ITERATE B.
					root_sout = root_sa;// t = b
					lcode = HAVE_COMPLETED_COMPUTATION_FOR_NEW_ITERATE_B;
					break;

				} while( false );
				if( lcode > 0 ) {
					break;
				}
				qsout = ((((solver.mYP[ 0 ] - (solver.mY[ 0 ] - solver.mYold[ 0 ]) / dels) / dels
						- ((solver.mY[ 0 ] - solver.mYold[ 0 ]) / dels - solver.mYpold[ 0 ]) / dels) / dels * ( root_sout - dels)
						+ ((solver.mY[ 0 ] - solver.mYold[ 0 ]) / dels - solver.mYpold[ 0 ]) / dels) *  root_sout
						+ solver.mYpold[ 0 ]) * root_sout + solver.mYold[ 0 ] - 1.;
			} while( true );

			// IF LAMBDA = 1 WERE BRACKETED,  ROOT  CANNOT FAIL.
			if( lcode > 2 ) {
				solver.mFlag = ERR_FAILED_TO_CONVERGE;
				return;
			}

			// CALCULATE Q(SA) AS THE INITIAL POINT FOR A NEWTON ITERATION.
			for( int jw = 0; jw < np1; ++jw ) {
				solver.mW[ jw ] = ((((solver.mYP[ jw ] - (solver.mY[ jw ] - solver.mYold[ jw ]) / dels) / dels
						- ((solver.mY[ jw ] - solver.mYold[ jw ]) / dels - solver.mYpold[ jw ]) / dels) / dels
						* (root_sa - dels) + ((solver.mY[ jw ] - solver.mYold[ jw ]) / dels
						- solver.mYpold[ jw ]) / dels) * root_sa + solver.mYpold[ jw ]) * root_sa + solver.mYold[ jw ];
			}
			// CALCULATE NEWTON STEP AT Q(SA).
			solver.mTangentRhoLength = root_sa;
			tangnf( solver );
			root_sa = solver.mTangentRhoLength;
			if( solver.mFlag > 0 ) {
				return;
			}
			// NEXT POINT = CURRENT POINT + NEWTON STEP.
			for( int jw = 0; jw < np1; ++jw ) {
				solver.mW[ jw ] += solver.mTZ[ jw ];
			}
			// GET THE TANGENT  WP  AT  W  AND THE NEXT NEWTON STEP IN  TZ .
			solver.mTangentRhoLength = root_sa;
			tangnf( solver );
			root_sa = solver.mTangentRhoLength;
			if( solver.mFlag > 0 ) {
				return;
			}
			// TAKE NEWTON STEP AND CHECK CONVERGENCE.
			for( int jw = 0; jw < np1; ++jw ) {
				solver.mW[ jw ] += solver.mTZ[ jw ];
			}
			if( Math.abs( solver.mW[ 0 ] - 1. ) <= (relative_error + absolute_error) && dnrm2( n, solver.mTZ, 1, 1 ) <= relative_error * dnrm2( n, solver.mW, 1, 1 ) + absolute_error ) {
				System.arraycopy( solver.mW, 0, solver.mY, 0, np1 );
				return;
			}
			// IF THE ITERATION HAS NOT CONVERGED, DISCARD ONE OF THE OLD POINTS
			// SUCH THAT  LAMBDA = 1  IS STILL BRACKETED.
			if( (solver.mYold[ 0 ] - 1.) * (solver.mW[ 0 ] - 1.) > 0. ) {
				System.arraycopy( solver.mW, 0, solver.mYold, 0, np1 );
				System.arraycopy( solver.mWP, 0, solver.mYpold, 0, np1 );
			} else {
				System.arraycopy( solver.mW, 0, solver.mY, 0, np1 );
				System.arraycopy( solver.mWP, 0, solver.mYP, 0, np1 );
			}
		}

		// ***** END OF MAIN LOOP. *****

		// THE ALTERNATING OSCULATORY CUBIC INTERPOLATION AND NEWTON ITERATION
		// HAS NOT CONVERGED IN  LIMIT  STEPS.  ERROR RETURN.
		solver.mFlag = ERR_FAILED_TO_CONVERGE;
		return;
	} /* rootnf */

	/**
	 *  STEPNF  takes one step along the zero curve of the homotopy map
	 * using a predictor-corrector algorithm.  The predictor uses a hermite
	 * cubic interpolant, and the corrector returns to the zero curve along
	 * the flow normal to the davidenko flow.  STEPNF  also estimates a
	 * step size H for the next step along the zero curve.  Normally
	 *  STEPNF  is used indirectly through  FIXPNF , and should be called
	 * directly only if it is necessary to modify the stepping algorithm's
	 * parameters.
	 *
	 * ON INPUT:
	 *
	 * N = dimension of X and the homotopy map.
	 *
	 * NFE = number of Jacobian matrix evaluations.
	 *
	 * IFLAG = -2, -1, or 0, indicating the problem type.
	 *
	 * START = .TRUE. on first call to  STEPNF , .FALSE. otherwise.
	 *
	 * HOLD = ||Y - YOLD||; should not be modified by the user.
	 *
	 * H = upper limit on length of step that will be attempted.  H  must be
	 *    set to a positive number on the first call to  STEPNF .
	 *    thereafter  STEPNF  calculates an optimal value for  H , and  H
	 *    should not be modified by the user.
	 *
	 * RELERR relative error value.
	 * ABSERR absolute error value.
	 *    The iteration is considered to have converged when
	 *    a point W=(LAMBDA,X) is found such that
	 * <pre>
	 *    ||Z|| <= RELERR*||W|| + ABSERR  ,          where
	 *
	 *    Z is the Newton step to W=(LAMBDA,X).
	 * </pre>
	 * S = (APPROXIMATE) arc length along the homotopy zero curve up to
	 *    Y(S) = (LAMBDA(S), X(S)).
	 *
	 * Y(1:N+1) = previous point (LAMBDA(S), X(S)) found on the zero curve of
	 *    the homotopy map.
	 *
	 * YP(1:N+1) = unit tangent vector to the zero curve of the homotopy map
	 *    at  Y .
	 *
	 * YOLD(1:N+1) = a point before  Y  on the zero curve of the homotopy map.
	 *
	 * YPOLD(1:N+1) = unit tangent vector to the zero curve of the homotopy
	 *    map at  YOLD .
	 *
	 * A(1:*) = parameter vector in the homotopy map.
	 *
	 *
	 * QR(1:N,1:N+2), ALPHA(1:N), TZ(1:N+1), PIVOT(1:N+1), W(1:N+1),
	 *    WP(1:N+1)  are work arrays used for the QR factorization (in the
	 *    Newton step calculation) and the interpolation.
	 *
	 * Z0(1:N+1), Z1(1:N+1)  are work arrays used for the estimation of the
	 *    next step size  H .
	 *
	 * SSPAR(1:8) = (LIDEAL, RIDEAL, DIDEAL, HMIN, HMAX, BMIN, BMAX, P)  is
	 *    a vector of parameters used for the optimal step size estimation.
	 *
	 * PAR(1:*) and IPAR(1:*) are arrays for (optional) user parameters,
	 *    which are simply passed through to the user written subroutines
	 *    RHO, RHOJAC.
	 *
	 * ON OUTPUT:
	 *
	 * N , A , SSPAR  are unchanged.
	 *
	 * NFE  has been updated.
	 *
	 * IFLAG
	 *    = -2, -1, or 0 (unchanged) on a normal return.
	 *
	 *    = 4 if a Jacobian matrix with RANK < N has occurred.  The
	 *        iteration was not completed.
	 *
	 *    = 6 if the iteration failed to converge.  W  contains the last
	 *        Newton iterate.
	 *
	 * START = .FALSE. on a normal return.
	 *
	 * CRASH
	 *    = .FALSE. on a normal return.
	 *
	 *    = .TRUE. if the step size  H  was too small.  H  has been
	 *      increased to an acceptable value, with which  STEPNF  may be
	 *      called again.
	 *
	 *    = .TRUE. if  RELERR  and/or  ABSERR  were too small.  They have
	 *      been increased to acceptable values, with which  STEPNF  may
	 *      be called again.
	 * <pre>
	 * HOLD = ||Y - YOLD||.
	 * </pre>
	 * H = optimal value for next step to be attempted.  Normally  H  should
	 *    not be modified by the user.
	 *
	 * RELERR, ABSERR  are unchanged on a normal return.
	 *
	 * S = (approximate) arc length along the zero curve of the homotopy map
	 *    up to the latest point found, which is returned in  Y .
	 *
	 * Y, YP, YOLD, YPOLD  contain the two most recent points and tangent
	 *    vectors found on the zero curve of the homotopy map.
	 */
	private static void stepnf( final PolsysSolver solver) {
		double lcalc = 0.;// java to avoid "The local variable lcalc may not have been initialized"
		double rcalc = 0.;// java to avoid "The local variable rcalc may not have been initialized"
		double hfail = 0.;// java to avoid "The local variable hfail may not have been initialized"

		// THE LIMIT ON THE NUMBER OF NEWTON ITERATIONS ALLOWED BEFORE REDUCING
		// THE STEP SIZE  H  MAY BE CHANGED BY CHANGING THE FOLLOWING PARAMETER
		// STATEMENT:

		// DEFINITION OF HERMITE CUBIC INTERPOLANT VIA DIVIDED DIFFERENCES.

		final int n = solver.mVarAndEqNumber << 1;
		final int np1 = n + 1;
		solver.mIsCrash = true;
		// THE ARCLENGTH  S  MUST BE NONNEGATIVE.
		if( solver.mS < 0. ) {
			return;
		}
		// IF STEP SIZE IS TOO SMALL, DETERMINE AN ACCEPTABLE ONE.
		if( solver.mH < (4. * MACHINE_PRECISION) * (solver.mS + 1.) ) {
			solver.mH = (4. * MACHINE_PRECISION) * (solver.mS + 1.);
			return;
		}
		// IF ERROR TOLERANCES ARE TOO SMALL, INCREASE THEM TO ACCEPTABLE VALUES.
		double temp = dnrm2( np1, solver.mY, 0, 1 );
		if( (solver.mStepRelError * temp + solver.mStepAbsError) * .5 < (2. * MACHINE_PRECISION) * temp ) {
			if( solver.mStepRelError != 0. ) {
				solver.mStepRelError = (4. * MACHINE_PRECISION) * ((4. * MACHINE_PRECISION) + 1.);
				solver.mStepAbsError = Math.max( solver.mStepAbsError, 0. );
			} else {
				solver.mStepAbsError = (4. * MACHINE_PRECISION) * temp;
			}
			return;
		}
		solver.mIsCrash = false;
		int itnum;
		boolean fail;
		if( ! solver.mIsStart ) {
			// ***** PREDICTOR SECTION. *****
			fail = false;
			// COMPUTE POINT PREDICTED BY HERMITE INTERPOLANT.  USE STEP SIZE  H
			// COMPUTED ON LAST CALL TO  STEPNF .

			// tangnf
			solver.mTangentY = solver.mW;
			solver.mTangentYp = solver.mWP;
			solver.mTangentYpold = solver.mYP;
DO_1:
			do {
				for( int j = 0; j < np1; ++j ) {
					final double d = solver.mHold + solver.mH;
					temp = ((((solver.mYP[ j ] - (solver.mY[ j ] - solver.mYold[ j ]) / solver.mHold) / solver.mHold
							- ((solver.mY[ j ] - solver.mYold[ j ]) / solver.mHold - solver.mYpold[ j ]) / solver.mHold)
							/ solver.mHold * (d - solver.mHold) + ((solver.mY[ j ] - solver.mYold[ j ])
									/ solver.mHold - solver.mYpold[ j ]) / solver.mHold) * d + solver.mYpold[ j ]
							) * d + solver.mYold[ j ];
					solver.mW[ j ] = temp;
					solver.mZ0[ j ] = temp;
				}

				// ***** END OF PREDICTOR SECTION. *****

				// ***** CORRECTOR SECTION. *****

				for( int judy = 1; judy <= 4; ++judy ) {
					solver.mTangentRhoLength = -1.;
					// CALCULATE THE NEWTON STEP  TZ  AT THE CURRENT POINT  W .
					tangnf( solver );
					if( solver.mFlag > 0 ) {
						return;
					}

					// TAKE NEWTON STEP AND CHECK CONVERGENCE.
					for( int j = 0; j < np1; ++j ) {
						solver.mW[ j ] += solver.mTZ[ j ];
					}
					itnum = judy;
					// COMPUTE QUANTITIES USED FOR OPTIMAL STEP SIZE ESTIMATION.
					if( judy == 1 ) {
						lcalc = dnrm2( np1, solver.mTZ, 0, 1 );
						rcalc = solver.mTangentRhoLength;
						System.arraycopy( solver.mW, 0, solver.mZ1, 0, np1 );
					} else if( judy == 2 ) {
						lcalc = dnrm2( np1, solver.mTZ, 0, 1 ) / lcalc;
						rcalc = solver.mTangentRhoLength / rcalc;
					}
					// GO TO MOP-UP SECTION AFTER CONVERGENCE.
					if( dnrm2( np1, solver.mTZ, 0, 1 ) <= solver.mStepRelError * dnrm2( np1, solver.mW, 0, 1 ) + solver.mStepAbsError ) {
						break DO_1;
					}
				}

				// NO CONVERGENCE IN  LITFH  ITERATIONS.  RECORD FAILURE AT CALCULATED  H ,
				// SAVE THIS STEP SIZE, REDUCE  H  AND TRY AGAIN.
				fail = true;
				hfail = solver.mH;
				if( solver.mH <= (4. * MACHINE_PRECISION) * (solver.mS + 1.) ) {
					solver.mFlag = ERR_FAILED_TO_CONVERGE;
					return;
				}
				solver.mH *= .5;
			} while( true );

			// ***** END OF CORRECTOR SECTION. *****
		} else {

			// *****  STARTUP SECTION(FIRST STEP ALONG ZERO CURVE.  *****

			fail = false;
			solver.mIsStart = false;
			// DETERMINE SUITABLE INITIAL STEP SIZE.
			// Computing MIN
			solver.mH = Math.min( Math.min( solver.mH, .1 ),
					Math.sqrt( Math.sqrt( solver.mStepRelError * temp + solver.mStepAbsError ) ) );
			// USE LINEAR PREDICTOR ALONG TANGENT DIRECTION TO START NEWTON ITERATION.
			solver.mYpold[ 0 ] = 1.;
			for( int j = 1; j < np1; ++j ) {
				solver.mYpold[ j ] = 0.;
			}
			//--- tangnf
			solver.mTangentRhoLength = solver.mS;
			solver.mTangentY = solver.mY;
			solver.mTangentYp = solver.mYP;
			solver.mTangentYpold = solver.mYpold;
			tangnf( solver );
			solver.mS = solver.mTangentRhoLength;
			if( solver.mFlag > 0 ) {
				return;
			}
			//--- tangnf
			solver.mTangentY = solver.mW;
			solver.mTangentYp = solver.mWP;
			solver.mTangentYpold = solver.mYpold;
DO_2:
			do {
				for( int j = 0; j < np1; ++j ) {
					temp = solver.mY[ j ] + solver.mH * solver.mYP[ j ];
					solver.mW[ j ] = temp;
					solver.mZ0[ j ] = temp;
				}
				for( int judy = 1; judy <= 4; ++judy ) {
					solver.mTangentRhoLength = -1.;
					// CALCULATE THE NEWTON STEP  TZ  AT THE CURRENT POINT  W .
					tangnf( solver );
					if( solver.mFlag > 0 ) {
						return;
					}

					// TAKE NEWTON STEP AND CHECK CONVERGENCE.
					for( int j = 0; j < np1; ++j ) {
						solver.mW[ j ] += solver.mTZ[ j ];
					}
					itnum = judy;
					// COMPUTE QUANTITIES USED FOR OPTIMAL STEP SIZE ESTIMATION.
					if( judy == 1 ) {
						lcalc = dnrm2( np1, solver.mTZ, 0, 1 );
						rcalc = solver.mTangentRhoLength;
						System.arraycopy( solver.mW, 0, solver.mZ1, 0, np1 );
					} else if( judy == 2 ) {
						lcalc = dnrm2( np1, solver.mTZ, 0, 1 ) / lcalc;
						rcalc = solver.mTangentRhoLength / rcalc;
					}
					// GO TO MOP-UP SECTION AFTER CONVERGENCE.
					if( dnrm2( np1, solver.mTZ, 0, 1 ) <=
							(solver.mStepRelError * dnrm2( np1, solver.mW, 0, 1 ) + solver.mStepAbsError) ) {
						break DO_2;
					}
				}
				// NO CONVERGENCE IN  LITFH  ITERATIONS.  REDUCE  H  AND TRY AGAIN.
				if( solver.mH <= (4. * MACHINE_PRECISION) * (solver.mS + 1.) ) {
					solver.mFlag = ERR_FAILED_TO_CONVERGE;
					return;
				}
				solver.mH *= .5;
			} while( true );

			// ***** END OF STARTUP SECTION. *****
		}
		// ***** MOP-UP SECTION. *****

		// YOLD  AND  Y  ALWAYS CONTAIN THE LAST TWO POINTS FOUND ON THE ZERO
		// CURVE OF THE HOMOTOPY MAP.  YPOLD  AND  YP  CONTAIN THE TANGENT
		// VECTORS TO THE ZERO CURVE AT  YOLD  AND  Y , RESPECTIVELY.

		System.arraycopy( solver.mY, 0, solver.mYold, 0, np1 );
		System.arraycopy( solver.mYP, 0, solver.mYpold, 0, np1 );
		System.arraycopy( solver.mW, 0, solver.mY, 0, np1 );
		System.arraycopy( solver.mWP, 0, solver.mYP, 0, np1 );
		for( int j = 0; j < np1; ++j ) {
			solver.mW[ j ] = solver.mY[ j ] - solver.mYold[ j ];
		}
		// UPDATE ARC LENGTH.
		solver.mHold = dnrm2( np1, solver.mW, 0, 1 );
		solver.mS += solver.mHold;

		// ***** END OF MOP-UP SECTION. *****

		// ***** OPTIMAL STEP SIZE ESTIMATION SECTION. *****

		// CALCULATE THE DISTANCE FACTOR  DCALC .
		for( int j = 0; j < np1; ++j ) {
			solver.mTZ[ j ] = solver.mZ0[ j ] - solver.mY[ j ];
			solver.mW[ j ] = solver.mZ1[ j ] - solver.mY[ j ];
		}
		double dcalc = dnrm2( np1, solver.mTZ, 0, 1 );
		if( dcalc != 0. ) {
			dcalc = dnrm2( np1, solver.mW, 0, 1 ) / dcalc;
		}

		// THE OPTIMAL STEP SIZE HBAR IS DEFINED BY

		//   HT=HOLD * [MIN(LIDEAL/LCALC, RIDEAL/RCALC, DIDEAL/DCALC)]^(1/P)

		//     HBAR = MIN [ MAX(HT, BMIN*HOLD, HMIN), BMAX*HOLD, HMAX ]

		// IF CONVERGENCE HAD OCCURRED AFTER 1 ITERATION, SET THE CONTRACTION
		// FACTOR  LCALC  TO ZERO.
		if( itnum == 1 ) {
			lcalc = 0.;
		}
		// FORMULA FOR OPTIMAL STEP SIZE.
		double ht;
		if( lcalc + rcalc + dcalc == 0. ) {
			ht = solver.mStepSizeParameterBmax * solver.mHold;
		} else {
			// Computing MAX
			double d2 = lcalc / solver.mStepSizeParameterLIdeal;
			double d3 = rcalc / solver.mStepSizeParameterRIdeal;
			d2 = Math.max( d2, d3 );
			d3 = dcalc / solver.mStepSizeParameterDIdeal;
			final double d1 = 1. / Math.max( d2, d3 );
			final double d4 = 1. / solver.mStepSizeParameterP;
			ht = Math.pow( d1, d4 ) * solver.mHold;
		}
		// HT  CONTAINS THE ESTIMATED OPTIMAL STEP SIZE.  NOW PUT IT WITHIN
		// REASONABLE BOUNDS.
		// Computing MIN
		// Computing MAX
		double d3 = ht;
		final double d4 = solver.mStepSizeParameterBmin * solver.mHold;
		d3 = Math.max( d3, d4 );
		double d1 = Math.max( d3, solver.mStepSizeParameterHmin );
		final double d2 = solver.mStepSizeParameterBmax * solver.mHold;
		d1 = Math.min( d1, d2 );
		solver.mH = Math.min( d1, solver.mStepSizeParameterHmax );
		if( itnum == 1 ) {
			// IF CONVERGENCE HAD OCCURRED AFTER 1 ITERATION, DON'T DECREASE  H .
			solver.mH = Math.max( solver.mH, solver.mHold );
		} else if( itnum == 4 ) {
			// IF CONVERGENCE REQUIRED THE MAXIMUM  LITFH  ITERATIONS, DON'T
			// INCREASE  H .
			solver.mH = Math.min( solver.mH, solver.mHold );
		}
		// IF CONVERGENCE DID NOT OCCUR IN  LITFH  ITERATIONS FOR A PARTICULAR
		// H = HFAIL , DON'T CHOOSE THE NEW STEP SIZE LARGER THAN  HFAIL .
		if( fail ) {
			solver.mH = Math.min( solver.mH, hfail );
		}

		return;
	} /* stepnf */

	private static final int FLAG_MATRIX_SINGULAR = 4;

	/**
	 * Subroutine  QRFAQF  computes the QR factorization of a matrix  A,
	 * where  R  is an upper triangular matrix, and  Q  is an orthogonal
	 * matrix which is the product of N-1 householder transformations
	 * <pre>
	 *   Q=H1*H2*...*H(N-1).
	 * </pre>
	 * the routine has two major steps.  First, the QR factorization
	 * of  A  is computed, resulting in defining the vector  R,  and
	 * storing information in the lower triangle of  QT  which will
	 * enable the construction of  Q transpose.
	 *
	 * The second step constructs  Q transpose from the information
	 * stored in  QT,  and places it in  QT.
	 *
	 * The information stored in the lower triangle of  QT  during the first
	 * step are the vectors  UJ, which define the householder transformations
	 * <pre>
	 *                   T
	 *   HJ = I - (UJ*UJ  / PJ),  where UJ[I]=0 for I=1...J-1,
	 *                             UJ[I]=QT[I,J], for I=J...N,
	 *                             PJ = the Jth component of UJ.
	 * </pre>
	 * ON INPUT:
	 *
	 * @param qt QT(1:N,1:N)  contains the matrix  A  to be factored.
	 *
	 * @param r R(1:N*(N+1)/2)  is undefined.
	 *
	 * @param n N  is the dimension of the matrix to be factored.
	 *
	 * ON OUTPUT:
	 *
	 * QT  contains  Q transpose.
	 *
	 * R(1:N*(N+1)/2)  contains the upper triangle of  R  stored by rows.
	 *
	 * N  is unchanged.
	 *
	 * @return IFLAG = 4  if the matrix a is singular. Otherwise, IFLAG = 0
	 *    see {@link #FLAG_MATRIX_SINGULAR}
	 */
	private static int qrfaqf(final double[] qt, final double[] r, final int n) {

		// ***** CALCULATION OF QR DECOMPOSITION, PLACING R IN THE VECTOR *****
		//       R, AND PLACING THE UJ VECTORS IN THE LOWER TRIANGLE OF
		//       QT.

		final int nm1 = n - 1;
		final int np1 = n + 1;

		int indexr = 0;
		for( int k = 0, qtkni = 0; k < nm1; ++k, qtkni += np1 ) {
			final double temp = dnrm2( n - k, qt, qtkni, 1 );
			if( temp == 0. ) {
				// MATRIX IS SINGULAR, SET  IFLAG  AND RETURN.
				return FLAG_MATRIX_SINGULAR;
			}
			final double isign = qt[ qtkni ] >= 0 ? 1. : -1.;

			// COMPUTE R(K,K).

			r[ indexr ] = -isign * temp;

			// UPDATE KTH COLUMN.

			dscal( n - k, isign / temp, qt, qtkni, 1 );
			qt[ qtkni ] += 1.;

			// UPDATE THE K+1ST - NTH COLUMNS OF  QT, AND  R.

			++indexr;
			for( int qti = k + n * (k + 1), qe = k + n * n; qti < qe; qti += n ) {
				final double tau = ddot( n - k,  qt, qtkni, 1,  qt, qti, 1 ) / qt[ qtkni ];
				r[ indexr ] = qt[ qti ] - tau * qt[ qtkni ];
				++indexr;
				daxpy( nm1 - k, -tau, qt, qtkni + 1, 1, qt, qti + 1, 1 );
			}
		}
		final int qtni = n * n - 1;
		if( qt[ qtni ] == 0. ) {

			// MATRIX IS SINGULAR, SET  IFLAG  AND RETURN.

			return FLAG_MATRIX_SINGULAR;
		}
		r[ indexr ] = qt[ qtni ];

		// ***** END OF FACTORING STEP *****

		// ***** CONSTRUCT Q TRANSPOSE IN QT *****

		// FORM Q BY MULTIPLYING ((I*H(N-1))*...)*H1.
		// THIS IS DONE IN PLACE IN  QT  BY UPDATING ONLY THE LOWER
		// RIGHT HAND CORNER OF QT  (QT(K,K) TO QT(N,N)).

		qt[ qtni ] = 1.;
		for( int k = nm1 - 1, qkn = k * np1; k >= 0; --k, qkn -= np1 ) {

			// MULTIPLY  QT  BY H(K).

			final double temp = qt[ qkn ];

			// UPDATE ROW K.

			qt[ qkn ] = 1. - qt[ qkn ];
			final int number = nm1 - k;
			dcopy( number, qt, qkn + 1, 1, qt, qkn + n, n );
			dscal( number, -1., qt, qkn + n, n );

			// UPDATE REMAINING ROWS.

			for( int qi = qkn + n, qikn = qkn - k + nm1, qe = qkn + 1; qikn >= qe; --qikn ) {
				double tau = ddot( nm1 - k, qt, qikn + n, n, qt, qi, n );
				qt[ qikn ] = tau;
				tau /= temp;
				daxpy( number, -tau, qt, qi, n, qt, qikn + n, n );
			}
		}
		return 0;
	} /* qrfaqf */

	/**
	 * Subroutine QRSLQF solves the system  R*S = QT*B  for  S.
	 *
	 * ON INPUT:
	 *
	 * @param qt QT(1:N,1:N)  contains QT in the equation above.
	 *
	 * @param r R(1:N*(N+1)/2)  contains the upper triangle of  R  in the equation
	 *     above, stored by rows.
	 *
	 * @param b B(1:N)  contains  B  in the equation above.
	 *
	 * @param n N   is the dimension of the problem.
	 *
	 * ON OUTPUT:
	 *
	 * QT  and  R  are unchanged.
	 *
	 * B   contains the solution vector S.
	 *
	 * @param x X(1:N)  is a work array which contains QT*B on output.
	 *
	 * X = QT*B.
	 */
	private static void qrslqf(
			final double[] qt,
			final double[] r,
			final double[] b,
			final double[] x,
			final int n
		)
	{
		for( int i = 0; i < n; ++i ) {
			x[ i ] = ddot( n, qt, i,  n,  b, 0, 1 );
		}

		// COMPUTE S USING BACK SUBSTITUTION.

		final int nm1 = n - 1;
		int indexr = ((n * (n + 1)) >>> 1) - 1;
		b[ nm1 ] = x[ nm1 ] / r[ indexr ];
		--indexr;
		for( int i = nm1; --i >= 0; ) {
			double tau = x[ i ];
			for( int j = nm1; j > i; --j ) {
				tau -= r[ indexr ] * b[ j ];
				--indexr;
			}
			b[ i ] = tau / r[ indexr ];
			--indexr;
		}
	} /* qrslqf */

	/**
	 * SCLGNP  scales the coefficients of a polynomial system of N
	 * equations in N unknowns, F(X)=0, where the Jth term of
	 * the ith equation looks like:
	 * <pre>
	 *    COEF(I,J) * X(1)^DEG(I,1,J) ... X(N)^DEG(I,N,J)
	 * </pre>
	 * the Ith equation is scaled by 10^FACE(I).  The Kth
	 * variable is scaled by 10^FACV(K).  In other words, X(K) =
	 * 10^FACV(K) * Y(K), where Y solves the scaled equation.
	 * The scaled equation has the same form as the original
	 * equation, except that COESCL(I,J) replaces COEF(I,J), where
	 * <pre>
	 * COESCL(I,J)=COEF(I,J)* 10^( FACE(I) + FACV(1)*DEG(I,1,J)+ ...
	 *                                       +FACV(N)*DEG(I,N,J) )
	 * </pre>
	 * The criterion for generating FACE and FACV is that of
	 * minimizing the sum of squares of the exponents of the scaled
	 * coefficients.  It turns out that this criterion reduces to
	 * solving a single linear system, ALPHA*X = BETA, as defined
	 * in the code below.  Further, the form of the polynomial
	 * system alone determines the matrix ALPHA.  Thus, in cases
	 * in which many systems of the same form, but with different
	 * coefficients, are to be scaled, the matrix ALPHA is
	 * unchanged and may be factored only once (by  QRFAQF).  When
	 * SCLGNP  is called with MODE=1,  SCLGNP  Does not recompute or
	 * refactor the matrix ALPHA.  See Meintjes and Morgan "A
	 * methodology for solving chemical equilibrium systems"
	 * (General Motors Research Laboratories Technical Report
	 * GMR-4971).
	 *
	 * ON INPUT:
	 *
	 * N  is the number of equations and the number of variables.
	 *          NN  is the declared dimension of several array indices.
	 *
	 * MMAXT  is an upper bound on the set NUMT(I), I=1 to N.
	 *
	 * @param nnumt NUMT(I)  is the number of terms in the I-th equation for I=1 to N.
	 *
	 * DEG(I,K,J)  is the degree of the K-th variable in the
	 *   J-th term of the I-th equation for I=1 to N, J=1 to NUMT(I), and
	 *   K=1 to N.
	 *
	 * @param mode MODE
	 *  ={@code true}  this is not the first call to  SCLGNP, and the form of the
	 *      system has not changed.
	 *  ={@code false}  this is the first call to  SCLGNP.
	 *
	 * @param eps0 EPS0  zero-epsilon for terms (terms less than  EPS0  in magnitude
	 *   are treated as zero by the scaling algorithm).
	 *
	 * COEF(I,J)  is the coefficient of the Jth term of the Ith equation
	 *   for I=1 to N and J=1 to NUMT(N).  (COEF(I,J) may be zero.)
	 *
	 * @param ddeg workspace
	 * @param ccoef workspace
	 * @param alpha workspace
	 * @param beta workspace
	 * @param rwork workspace
	 * @param xwork workspace
	 * NNUMT, DDEG, CCOEF, ALPHA, BETA, RWORK, and  XWORK  are workspaces.
	 *
	 * ON OUTPUT:
	 *
	 * N, NUMT, DEG, MODE, EPS0, and  COEF  are unchanged.
	 *
	 * FACV(I)  is the variable scale factor for the I-th variable, for
	 *   I=1 to N.
	 *
	 * @param face FACE(I)  is the equation scale factor for the I-th equation, for
	 *   I=1 to N.
	 *
	 * @param coescl COESCL(I,J)  is the scaled version of coefficient  COEF(I,J), for
	 *   I=1 to N, J=1 to NUMT(I), unless IERR=1.
	 *
	 * @return IERR
	 *   =0  if scaling matrix, ALPHA, is well conditioned.
	 *   =1  otherwise.  in this case, ALPHA is "repaired" and a
	 *         scaling is computed.
	 */
	private static boolean sclgnp(final PolsysSolver solver,
			final boolean mode,
			final double eps0,

			final int[] nnumt,
			final int[] ddeg,

			final double[] ccoef,
			final double[] alpha,
			final double[] beta,
			final double[] rwork,
			final double[] xwork,

			final double[] face,
			final double[] coescl)
	{
		boolean ierr = false;
		final int n = solver.mVarAndEqNumber;
		final int n_n = n * n;
		final int n_n_pn = n_n + n;
		final int lenr = n_n_pn >>> 1;

		// DELETE NEAR ZERO TERMS
		for( int i = 0; i < n; ++i ) {
			nnumt[ i ] = 0;
			final int term = n_n_pn * solver.mNumberOfTerms[ i ];
			for( int j = 0, jj = 0, ci = i; j < term; j += n_n_pn, ci += n ) {
				final double c = solver.mCoefficients[ ci ];
				if( Math.abs( c ) > eps0 ) {
					++nnumt[ i ];
					ccoef[ jj + i ] = c;
					for( int k = 0, ddegi = i + jj * n + jj, degi = i + j; k < n; ++k ) {
						ddeg[ ddegi ] = solver.mVarDegrees[ degi ];
						ddegi += n;
						degi += n;
					}
					jj += n;
				}
			}
		}
		for( int i = 0; i < n; ++i ) {
			final int term = i + n * nnumt[ i ];
			for( int j = i; j < term; j += n ) {
				coescl[ j ] = Math.log10( Math.abs( ccoef[ j ] ) );
			}
		}

		// SKIP OVER THE GENERATION AND DECOMPOSITON OF MATRIX ALPHA IF MODE=1
		if( ! mode ) {

			// GENERATE THE MATRIX ALPHA
			final int n2 = n << 1;
			final int n_n_2 = n_n << 1;
			for( int s = 0; s < n; ++s ) {
				for( int k = 0; k < n_n_2; k += n2 ) {
					alpha[ s + k ] = 0.;
				}
			}
			for( int s = 0; s < n; ++s ) {
				alpha[ n2 * s + s ] = (double) nnumt[ s ];
			}
			for( int s = 0; s < n; ++s ) {
				for( int i = 0; i < n; ++i ) {
					double sum = 0.;
					final int term = i + n * s + n_n_pn * nnumt[ i ];
					for( int j = i + n * s; j < term; j += n_n_pn ) {
						sum += ddeg[ j ];
					}
					alpha[ s + i * n2 + n ] = sum;
				}
			}
			for( int s = n, sn = 0; s < n2; ++s, sn += n ) {
				for( int k = 0; k < n_n; k += n ) {
					double sum = 0.;
					for( int i = 0; i < n; ++i ) {
						final int term = i + n_n_pn * nnumt[ i ];
						for( int jn = i; jn < term; jn += n_n_pn ) {
							sum += ddeg[ jn + sn ] * ddeg[ jn + k ];
						}
					}
					alpha[ s + ((n_n + k) << 1) ] = sum;
				}
			}
			for( int s = 0; s < n; ++s ) {
				for( int k = 0, ai = s + n_n_2; k < n_n; k += n ) {
					double sum = 0.;
					for( int j = s + k,
							term = j + n_n_pn * nnumt[ s ];
							j < term; j += n_n_pn )
					{
						sum += ddeg[ j ];
					}
					alpha[ ai + (k << 1) ] = sum;
				}
			}

			// COMPUTE QR FACTORIZATION OF MATRIX ALPHA
			qrfaqf( alpha, rwork, n2 );

			// REPAIR ILL CONDITIONED SCALING MATRIX
			final int irmax = idamax( lenr, rwork, 1 );
			final double rtol = MACHINE_PRECISION * n * rwork[ irmax ];
			int index = 0;
			for( int i = n; i >= 2; --i ) {
				if( Math.abs( rwork[ index ] ) < rtol ) {
					rwork[ index ] = LAGEST_MAGNITUDE;
					ierr = true;
				}
				index += i;
			}
			if( Math.abs( rwork[ index ] ) < rtol ) {
				rwork[ index ] = LAGEST_MAGNITUDE;
				ierr = true;
			}
		}

		// CONTROL PASSES HERE IF MODE=1

		// GENERATE THE COLUMN BETA
		for( int s = 0; s < n; ++s ) {
			double sum = 0.;
			for( int ci = s, ce = ci + n * nnumt[ s ]; ci < ce; ci += n ) {
				sum += coescl[ ci ];
			}
			beta[ s ] = -sum;
		}
		for( int s = 0, bi = n; s < n; ++s, ++bi ) {
			double sum = 0.;
			for( int i = 0; i < n; ++i ) {
				final int term = i + n * nnumt[ i ];
				for( int j = i, di = i + n * s; j < term; j += n, di += n_n_pn ) {
					sum += coescl[ j ] * ddeg[ di ];
				}
			}
			beta[ bi ] = -sum;
		}

		// SOLVE THE LINEAR SYSTEM ALPHA * X = BETA
		qrslqf( alpha,  rwork,  beta,  xwork,  n << 1 );

		// GENERATE FACE, FACV, AND THE MATRIX COESCL
		for( int i = 0; i < n; ++i ) {
			face[ i ] = beta[ i ];
			solver.mFACV[ i ] = beta[ i + n ];
		}
		for( int i = 0; i < n; ++i ) {
			final int nterm = n * solver.mNumberOfTerms[ i ];
			for( int jn = 0; jn < nterm; jn += n ) {
				final double c = solver.mCoefficients[ jn + i ];
				final double abs = Math.abs( c );
				if( abs == 0. ) {
					coescl[ jn + i ] = 0.;
				} else {
					double sum = face[ i ] + Math.log10( abs );
					for( int k = 0, di = i + jn + jn * n; k < n; ++k, di += n ) {
						sum += solver.mFACV[ k ] * solver.mVarDegrees[ di ];
					}
					coescl[ jn + i ] = Math.copySign( Math.pow( 10., sum ), c );
				}
			}
		}
		return ierr;
	} /* sclgnp */

	/*
	 * OTPUTP  postprocesses the endpoints of the paths, untransforming
	 * and unscaling them.
	 *
	 * ON INPUT:
	 *
	 * @param n N  is the number of equations and variables.
	 *
	 * NUMPAT  is the current path number.
	 *
	 * @param cl CL  is the array that defines the projective transformation.
	 *
	 * @param facv FACV  contains the variable scaling factors.
	 *
	 * @param x X  is the endpoint of the path, possibly transformed and/or scaled
	 *   depending on the  POLSYS  input flag  IFLG1.
	 *
	 * @param clx CLX  is workspace.
	 *
	 * ON OUTPUT:
	 *
	 * N, NUMPAT, CL, and  FACV  are unchanged.
	 *
	 * X  is the untransformed and unscaled version of X.
	 *
	 * @param xnp1 XNP1  is the projective coordinate "X(N+1)".  XNP1  equals unity if
	 *   the projective transformation is not specified.
	 */
	/* private static void otputp(final int n,
			final double[] cl,
			final double[] facv,
			final double[] clx,
			final double[] x, final int xOff,
			final double[] xnp1)
	{
		// Local variables
		final double[] temp = new double[ 2 ];

		// Function Body

		// COMPUTE XNP1
		final int n2 = n + n;
		for( int j2 = 0; j2 < n2; j2 += 2 ) {
			mulp( cl, j2, x, xOff + j2, clx, j2 );
		}
		double sum = cl[ n2 ];
		for( int j2 = 0; j2 < n2; j2 += 2 ) {
			sum += clx[ j2 ];
		}
		xnp1[ 0 ] = sum;
		sum = cl[ n2 + 1 ];
		for( int j2 = 1; j2 < n2; j2 += 2 ) {
			sum += clx[ j2 ];
		}
		xnp1[ 1 ] = sum;
		// UNTRANSFORM VARIABLES
		for( int j2 = 0; j2 < n2; ) {
			divp( x, xOff + j2, xnp1, 0, temp, 0 );
			x[ xOff + j2++ ] = temp[ 0 ];
			x[ xOff + j2++ ] = temp[ 1 ];
		}
		// UNSCALE VARIABLES
		for( int j = 0; j < n; ++j ) {
			final int j2 = j + j;
			final double fac = Math.pow( 10., facv[ j ] );
			int ix = xOff + j2;
			if( Math.abs( x[ ix ] ) / LAGEST_MAGNITUDE * fac < 1. ) {
				x[ ix ] *= fac;
			}
			++ix;
			if( Math.abs( x[ ix ] ) / LAGEST_MAGNITUDE * fac < 1. ) {
				x[ ix ] *= fac;
			}
		}
	} */ /* otputp */

	/* SSPAR(1:8) = (LIDEAL, RIDEAL, DIDEAL, HMIN, HMAX, BMIN, BMAX, P)
	 * Is a vector of parameters used for the optimal step size estimation. */

	/**
	 * Ideal contraction factor:  ||Z[2] - Z[1]|| / ||Z[1] - Z[0]||
	 * Default value: 0.5
	 */
	public static final int STEP_SIZE_ESTIMATION_LIDEAL = 0;

	/**
	 * Ideal residual factor:  ||RHO(A, Z[1])|| / ||RHO(A, Z[0])||
	 * Default value: 0.01
	 */
	public static final int STEP_SIZE_ESTIMATION_RIDEAL = 1;

	/** Ideal distance factor:  ||Z[1] - Y|| / ||Z[0] - Y||
	 * Default value: 0.5 */
	public static final int STEP_SIZE_ESTIMATION_DIDEAL = 2;

	/**
	 * Minimum step size  HMIN .
	 * Default value: MACHINE_PRECISION * (sqrt( n + 1) + 4.)
	 */
	public static final int STEP_SIZE_ESTIMATION_HMIN   = 3;

	/**
	 * Maximum step size  HMAX .
	 * Default value: 1.0
	 */
	public static final int STEP_SIZE_ESTIMATION_HMAX   = 4;

	/**
	 * Minimum step size reduction factor  BMIN .
	 * Default value: 0.1
	 */
	public static final int STEP_SIZE_ESTIMATION_BMIN   = 5;

	/**
	 * Maximum step size expansion factor  BMAX .
	 * Default value: 3.0
	 */
	public static final int STEP_SIZE_ESTIMATION_BMAX   = 6;

	/**
	 * Assumed operating order  P .
	 * Default value: 2.0
	 */
	public static final int STEP_SIZE_ESTIMATION_P      = 7;

	/** -2   Causes  FIXPNF  to initialize everything for the problem
	 *       RHO(A,LAMBDA,X) = 0 (use on first call). */
	public static final int ERR_RHO_INITIALIZE_PROBLEM = -2;

	/** -1   Causes  FIXPNF  to initialize everything for the problem
	 *       F(X) = 0 (use on first call). */
	public static final int ERR_F_INITIALIZE_PROBLEM = -1;

	/**  1   Normal return. */
	public static final int ERR_NO = 1;

	/**  2   Specified error tolerance cannot be met.  Some or all of
	 *       ARCRE , ARCAE , ANSRE , ANSAE  have been increased to
	 *       suitable values.  To continue, just call  FIXPNF  again
	 *       without changing any parameters. */
	public static final int ERR_EPSILON_TOO_SMALL = 2;

	/**  3   STEPNF  has been called 1000 times.  To continue, call
	 *       FIXPNF  again without changing any parameters. */
	public static final int ERR_MAXIMUM_STEPS_NUMBER_EXCEEDED = 3;

	/**  4   Jacobian matrix does not have full rank.  The algorithm
	 *       has failed (the zero curve of the homotopy map cannot be
	 *       followed any further). */
	public static final int ERR_ALGORITHM_FAILED = 4;

	/**  5   The tracking algorithm has lost the zero curve of the
	 *       homotopy map and is not making progress.  The error tolerances
	 *       ARCRE  and  ANSRE  were too lenient.  The problem should be
	 *       restarted by calling  FIXPNF  with smaller error tolerances
	 *       and  IFLAG = 0 (-1, -2). */
	public static final int ERR_NO_PROGRESS = 5;

	/**  6   The normal flow Newton iteration in  STEPNF  or  ROOTNF
	 *       failed to converge.  The error tolerances  ANSRE  may be too
	 *       stringent. */
	public static final int ERR_FAILED_TO_CONVERGE = 6;

	/** 7   Illegal input parameters, a fatal error. */
	public static final int ERR_ILLEGAL_INPUT_PARAMETERS = 7;

	/**
	 * Subroutine  FIXPNF  finds a fixed point or zero of the
	 * N-dimensional vector function F(X), or tracks a zero curve
	 * of a general homotopy map RHO(A,LAMBDA,X).  For the fixed
	 * point problem F(X) is assumed to be a C2 map of some ball
	 * into itself.  the equation  X = F(X)  is solved by
	 * following the zero curve of the homotopy map
	 * <pre>
	 *  LAMBDA*(X - F(X)) + (1 - LAMBDA)*(X - A)  ,
	 * </pre>
	 * starting from LAMBDA = 0, X = A.  The curve is parameterized
	 * by arc length s, and is followed by solving the ordinary
	 * differential equation  D(HOMOTOPY MAP)/DS = 0  for
	 * Y(S) = (LAMBDA(S), X(S)) using a hermite cubic predictor and a
	 * corrector which returns to the zero curve along the flow normal
	 * to the Davidenko flow (which consists of the integral curves of
	 * D(HOMOTOPY MAP)/DS ).
	 *
	 * For the zero finding problem F(X) is assumed to be a C2 map
	 * such that for some R > 0,  X*F(X) >= 0  whenever NORM(X) = R.
	 * the equation  F(X) = 0  is solved by following the zero curve
	 * of the homotopy map
	 * <pre>
	 *   LAMBDA*F(X) + (1 - LAMBDA)*(X - A)
	 * </pre>
	 * emanating from LAMBDA = 0, X = A.
	 *
	 *  A  must be an interior point of the above mentioned balls.
	 *
	 * for the curve tracking problem RHO(A,LAMBDA,X) is assumed to
	 * be a C2 map from E^M X [0,1) X E^N into E^N, which for
	 * almost all parameter vectors a in some nonempty open subset
	 * of E^M satisfies
	 * <pre>
	 *  RANK [D RHO(A,LAMBDA,X)/D LAMBDA , D RHO(A,LAMBDA,X)/DX] = N
	 * </pre>
	 * for all points (LAMBDA,X) such that RHO(A,LAMBDA,X)=0.  It is
	 * further assumed that
	 * <pre>
	 *           RANK [ D RHO(A,0,X0)/DX ] = N  .
	 * </pre>
	 * with a fixed, the zero curve of RHO(A,LAMBDA,X) emanating
	 * from  LAMBDA = 0, X = X0  is tracked until  LAMBDA = 1  by
	 * solving the ordinary differential equation
	 * D RHO(A,LAMBDA(S),X(S))/DS = 0  for  Y(S) = (LAMBDA(S), X(S)),
	 * where s is arc length along the zero curve.  Also the homotopy
	 * map RHO(A,LAMBDA,X) is assumed to be constructed such that
	 * <pre>
	 *              D LAMBDA(0)/DS > 0  .
	 * </pre>
	 * for the fixed point and zero finding problems, the user must supply
	 * a subroutine  F(X,V)  which evaluates F(X) at x and returns the
	 * vector F(X) in V, and a subroutine  FJAC(X,V,K)  which returns in V
	 * the Kth column of the jacobian matrix of F(X) evaluated at X.  For
	 * the curve tracking problem, the user must supply a subroutine
	 *  RHO(A,LAMBDA,X,V,PAR,IPAR)  which evaluates the homotopy map RHO at
	 * (A,LAMBDA,X) and returns the vector RHO(A,LAMBDA,X) in V, and A
	 * subroutine  RHOJAC(A,LAMBDA,X,V,K,PAR,IPAR)  which returns in V the Kth
	 * column of the N X (N+1) jacobian matrix [D RHO/D LAMBDA, D RHO/DX]
	 * evaluated at (A,LAMBDA,X).  FIXPNF  directly or indirectly uses
	 * the subroutines  STEPNF , TANGNF , ROOTNF , ROOT , F (or  RHO ),
	 * FJAC (or  RHOJAC ), D1MACH , and the BLAS functions  DDOT  and
	 * DNRM2 .  Only  D1MACH  contains machine dependent constants.
	 * no other modifications by the user are required.
	 *
	 * On INPUT:
	 *
	 * N  is the dimension of X, F(X), and RHO(A,LAMBDA,X).
	 *
	 * Y  is an array of length  N + 1.  (Y(2),...,Y(N+1)) = A  is the
	 *    starting point for the zero curve for the fixed point and
	 *    zero finding problems.  (Y(2),...,Y(N+1)) = X0  for the curve
	 *    tracking problem.
	 *
	 * @param iflag IFLAG  can be -2, -1, 0, 2, or 3.  IFLAG  should be 0 on the
	 *    first call to  FIXPNF  for the problem  X=F(X), -1 for the
	 *    problem  F(X)=0, and -2 for the problem  RHO(A,LAMBDA,X)=0.
	 *    in certain situations  IFLAG  is set to 2 or 3 by  FIXPNF,
	 *    and  FIXPNF  can be called again without changing  IFLAG.
	 *
	 * ARCRE , ARCAE  are the relative and absolute errors, respectively,
	 *    allowed the normal flow iteration along the zero curve.  if
	 *    ARCRE <= 0.0  on input it is reset to  .5*sqrt(ANSRE) .
	 *    normally  ARCRE should be considerably larger than  ANSRE .
	 *
	 * ANSRE , ANSAE  are the relative and absolute error values used for
	 *    the answer at LAMBDA = 1.  the accepted answer  Y = (LAMBDA, X)
	 *    satisfies
	 * <pre>
	 *       |Y(1) - 1|  <=  ANSRE + ANSAE           &&
	 *
	 *       ||Z||  <=  ANSRE*||X|| + ANSAE          where
	 *
	 *    (.,Z) is the newton step to Y.
	 * </pre>
	 * TRACE  is an PrintStream specifying the logical i/o unit for
	 *    intermediate output.  if  TRACE != {@code null}  the points computed on
	 *    the zero curve are written to i/o unit  TRACE .
	 *
	 * A(1:*)  contains the parameter vector  A .  For the fixed point
	 *    and zero finding problems, A  need not be initialized by the
	 *    user, and is assumed to have length  N.  for the curve
	 *    tracking problem, A  must be initialized by the user.
	 *
	 * YP(1:N+1)  is a work array containing the tangent vector to
	 *    the zero curve at the current point  Y .
	 *
	 * YOLD(1:N+1)  is a work array containing the previous point found
	 *    on the zero curve.
	 *
	 * YPOLD(1:N+1)  is a work array containing the tangent vector to
	 *    the zero curve at  YOLD .
	 *
	 * QR(1:N,1:N+2), ALPHA(1:N), TZ(1:N+1), PIVOT(1:N+1) , W(1:N+1) ,
	 *    WP(1:N+1) , Z0(1:N+1) , Z1(1:N+1)  are all work arrays used by
	 *    STEPNF  to calculate the tangent vectors and newton steps.
	 *
	 * SSPAR(1:8) = (LIDEAL, RIDEAL, DIDEAL, HMIN, HMAX, BMIN, BMAX, P)  is
	 *    a vector of parameters used for the optimal step size estimation.
	 *    if  SSPAR(J) <= 0.0  on input, it is reset to a default value
	 *    by  FIXPNF .  otherwise the input value of  SSPAR(J)  is used.
	 *    see the comments below and in  STEPNF  for more information about
	 *    these constants.
	 *
	 * PAR(1:*) and IPAR(1:*) are arrays for (optional) user parameters,
	 *    which are simply passed through to the user written subroutines
	 *    RHO, RHOJAC.
	 *
	 * ON OUTPUT:
	 *
	 * N , trace , a  are unchanged.
	 *
	 * Y(1) = LAMBDA, (Y(2),...,Y(N+1)) = X, and Y is an approximate
	 *    zero of the homotopy map.  Normally LAMBDA = 1 and X is a
	 *    fixed point(zero) of F(X).  In abnormal situations lambda
	 *    may only be near 1 and X is near a fixed point(zero).
	 *
	 * IFLAG =
	 *  -2   Causes  FIXPNF  to initialize everything for the problem
	 *       RHO(A,LAMBDA,X) = 0 (use on first call).
	 *
	 *  -1   Causes  FIXPNF  to initialize everything for the problem
	 *       F(X) = 0 (use on first call).
	 *
	 *   0   Causes  FIXPNF  to initialize everything for the problem
	 *       X = F(X) (use on first call).
	 *
	 *   1   Normal return.
	 *
	 *   2   Specified error tolerance cannot be met.  Some or all of
	 *       ARCRE , ARCAE , ANSRE , ANSAE  have been increased to
	 *       suitable values.  To continue, just call  FIXPNF  again
	 *       without changing any parameters.
	 *
	 *   3   STEPNF  has been called 1000 times.  To continue, call
	 *       FIXPNF  again without changing any parameters.
	 *
	 *   4   Jacobian matrix does not have full rank.  The algorithm
	 *       has failed (the zero curve of the homotopy map cannot be
	 *       followed any further).
	 *
	 *   5   The tracking algorithm has lost the zero curve of the
	 *       homotopy map and is not making progress.  The error tolerances
	 *       ARCRE  and  ANSRE  were too lenient.  The problem should be
	 *       restarted by calling  FIXPNF  with smaller error tolerances
	 *       and  IFLAG = 0 (-1, -2).
	 *
	 *   6   The normal flow Newton iteration in  STEPNF  or  ROOTNF
	 *       failed to converge.  the error tolerances  ANSRE  may be too
	 *       stringent.
	 *
	 *   7   Illegal input parameters, a fatal error.
	 *
	 * ARCRE , ARCAE , ANSRE , ANSAE  are unchanged after a normal return
	 *    (IFLAG = 1).  they are increased to appropriate values on the
	 *    return  IFLAG = 2 .
	 *
	 * NFE  is the number of function evaluations (= number of
	 *    jacobian evaluations).
	 *
	 * ARCLEN  is the length of the path followed.
	 */
	private static int polsys(final PolsysSolver solver, final int iflag) {//, final PrintStream trace) {
		final int n = solver.mVarAndEqNumber << 1;

		if( (n <= 0) || (solver.mAnswerRelError <= 0.) || (solver.mAnswerAbsError < 0.) ) {
			return ERR_ILLEGAL_INPUT_PARAMETERS;
		}

		// *****  INITIALIZATION BLOCK.  *****

		switch( iflag ) {
			default:
				return ERR_ILLEGAL_INPUT_PARAMETERS;
			case -2 :
			case -1 :
			case 0 : // ( iflag[ 0 ] >= -2 && iflag[ 0 ] <= 0 ) L20
				solver.mArcLength = 0.;
				if( solver.mArcRelError <= 0. ) {
					solver.mArcRelError = Math.sqrt( solver.mAnswerRelError ) * .5;
				}
				if( solver.mArcAbsError <= 0. ) {
					solver.mArcAbsError = Math.sqrt( solver.mAnswerAbsError ) * .5;
				}
				solver.mNumberOfMatrixEvaluations = 0;
				solver.mFlag = iflag;
				// SET INITIAL CONDITIONS FOR FIRST CALL TO  STEPNF .
				solver.mIsStart = true;
				solver.mIsCrash = false;
				solver.mHold = 1.;
				solver.mH = .1;
				solver.mS = 0.;
				solver.mYpold[ 0 ] = 1.;
				solver.mYP[ 0 ] = 1.;
				solver.mY[ 0 ] = 0.;
				for( int jw = 1; jw <= n; ++jw ) {
					solver.mYpold[ jw ] = 0.;
					solver.mYP[ jw ] = 0.;
				}
				// SET OPTIMAL STEP SIZE ESTIMATION PARAMETERS.
				// LET Z[K] DENOTE THE NEWTON ITERATES ALONG THE FLOW NORMAL TO THE
				// DAVIDENKO FLOW AND Y THEIR LIMIT.
				// IDEAL CONTRACTION FACTOR:  ||Z[2] - Z[1]|| / ||Z[1] - Z[0]||
				if( solver.mStepSizeParameterLIdeal <= 0. ) {// if( solver.sspar[ LIDEAL ] <= 0. ) {
					solver.mStepSizeParameterLIdeal = .5;// solver.sspar[ RIDEAL ] = .01;
				}
				// IDEAL RESIDUAL FACTOR:  ||RHO(A, Z[1])|| / ||RHO(A, Z[0])||
				if( solver.mStepSizeParameterRIdeal <= 0. ) {// if( solver.sspar[ RIDEAL ] <= 0. ) {
					solver.mStepSizeParameterRIdeal = .01;// solver.sspar[ RIDEAL ] = .01;
				}
				// IDEAL DISTANCE FACTOR:  ||Z[1] - Y|| / ||Z[0] - Y||
				if( solver.mStepSizeParameterDIdeal <= 0. ) {// if( solver.sspar[ DIDEAL ] <= 0. ) {
					solver.mStepSizeParameterDIdeal = .5;// solver.sspar[ DIDEAL ] = .5;
				}
				// MINIMUM STEP SIZE  HMIN .
				if( solver.mStepSizeParameterHmin <= 0. ) {// if( solver.sspar[ HMIN ] <= 0. ) {
					solver.mStepSizeParameterHmin = (Math.sqrt( n + 1.) + 4.) * MACHINE_PRECISION;// solver.sspar[ HMIN ] = (Math.sqrt( n + 1.) + 4.) * MACHINE_PRECISION;
				}
				// MAXIMUM STEP SIZE  HMAX .
				if( solver.mStepSizeParameterHmax <= 0. ) {// if( solver.sspar[ HMAX ] <= 0. ) {
					solver.mStepSizeParameterHmax = 1.;// solver.sspar[ HMAX ] = 1.;
				}
				// MINIMUM STEP SIZE REDUCTION FACTOR  BMIN .
				if( solver.mStepSizeParameterBmin <= 0. ) {// if( solver.sspar[ BMIN ] <= 0. ) {
					solver.mStepSizeParameterBmin = .1;// solver.sspar[ BMIN ] = .1;
				}
				// MAXIMUM STEP SIZE EXPANSION FACTOR  BMAX .
				if( solver.mStepSizeParameterBmax <= 0. ) {// if( solver.sspar[ BMAX ] <= 0. ) {
					solver.mStepSizeParameterBmax = 3.;// solver.sspar[ BMAX ] = 3.;
				}
				// ASSUMED OPERATING ORDER  P .
				if( solver.mStepSizeParameterP <= 0. ) {// if( solver.sspar[ P ] <= 0. ) {
					solver.mStepSizeParameterP = 2.;// solver.sspar[ P ] = 2.;
				}

				// LOAD  A  FOR THE FIXED POINT AND ZERO FINDING PROBLEMS.
				if( solver.mFlag >= -1 ) {
					System.arraycopy( solver.mY, 1, solver.mQDG, 0, n );
				}

			case 3 :// L90

				solver.mIterationNumberLimitation = 1000;

			case 2 :// L120

		}

		// *****  END OF INITIALIZATION BLOCK.  *****

		// *****  MAIN LOOP.  *****

		for( int iter = 1; iter <= solver.mIterationNumberLimitation; ++iter ) {
			if( solver.mY[ 0 ] < 0. ) {
				solver.mArcLength = solver.mS;
				return ERR_NO_PROGRESS;
			}

			// SET DIFFERENT ERROR TOLERANCE IF THE TRAJECTORY Y(S) HAS ANY HIGH
			// CURVATURE COMPONENTS.
			final double curr_tolerance = solver.mHold * 10.;
			solver.mStepRelError = solver.mArcRelError;
			solver.mStepAbsError = solver.mArcAbsError;
			for( int jw = 0; jw <= n; ++jw ) {
				if( Math.abs( solver.mYP[ jw ] - solver.mYpold[ jw ] ) > curr_tolerance ) {
					solver.mStepRelError = solver.mAnswerRelError;
					solver.mStepAbsError = solver.mAnswerAbsError;
					break;
				}
			}

			// TAKE A STEP ALONG THE CURVE.
			stepnf( solver );

			// PRINT LATEST POINT ON CURVE IF REQUESTED.
			// if( trace != null ) {
				final java.io.PrintStream trace = System.out;
				trace.printf( "%nSTEP%d\tNFE = %d\tRC LENGTH = %g\tLAMBDA = %g%n",
						Integer.valueOf( iter ), Integer.valueOf( solver.mNumberOfMatrixEvaluations ),
						Double.valueOf( solver.mS ), Double.valueOf( solver.mY[ 0 ] ) );

				for( int jw = 1; jw <= n; ++jw ) {
					trace.printf( "%s, ", Double.toString( solver.mY[ jw ] ) );
				}
				trace.println();
				trace.flush();

			//}
			// CHECK IF THE STEP WAS SUCCESSFUL.
			if( solver.mFlag > 0 ) {
				solver.mArcLength = solver.mS;
				return solver.mFlag;
			}
			if( solver.mIsCrash ) {

				// CHANGE ERROR TOLERANCES.
				if( solver.mArcRelError < solver.mStepRelError ) {
					solver.mArcRelError = solver.mStepRelError;
				}
				if( solver.mAnswerRelError < solver.mStepRelError ) {
					solver.mAnswerRelError = solver.mStepRelError;
				}
				if( solver.mArcAbsError < solver.mStepAbsError ) {
					solver.mArcAbsError = solver.mStepAbsError;
				}
				if( solver.mAnswerAbsError < solver.mStepAbsError ) {
					solver.mAnswerAbsError = solver.mStepAbsError;
				}
				// CHANGE LIMIT ON NUMBER OF ITERATIONS.
				solver.mIterationNumberLimitation -= iter;
				// RETURN CODE FOR ERROR TOLERANCE TOO SMALL.
				return ERR_EPSILON_TOO_SMALL;
			}

			if( solver.mY[ 0 ] >= 1. ) {

				// USE HERMITE CUBIC INTERPOLATION AND NEWTON ITERATION TO GET THE
				// ANSWER AT LAMBDA = 1.0 .

				// SAVE  YOLD  FOR ARC LENGTH CALCULATION LATER.
				System.arraycopy( solver.mYold, 0, solver.mZ0, 0, n + 1 );
				rootnf( solver );

				// CALCULATE FINAL ARC LENGTH.
				for( int jw = 0; jw <= n; ++jw ) {
					solver.mW[ jw ] = solver.mY[ jw ] - solver.mZ0[ jw ];
				}
				solver.mArcLength = solver.mS - solver.mHold + dnrm2( n + 1, solver.mW, 0, 1 );
				// SET ERROR FLAG IF  ROOTNF  COULD NOT GET THE POINT ON THE ZERO
				// CURVE AT  LAMBDA = 1.0  .
				if( solver.mFlag > 0 ) {
					return solver.mFlag;
				}
				return ERR_NO;
			}

			// FOR POLYNOMIAL SYSTEMS AND THE  POLSYS  HOMOTOPY MAP,
			// D LAMBDA/DS <= 0 NECESSARILY.  THIS CONDITION IS FORCED HERE IF
			// THE ENTRY POINT WAS  POLYNF .

			// if( polsys ) {
				if( solver.mYP[ 0 ] < 0. ) {
					// REVERSE TANGENT DIRECTION SO D LAMBDA/DS = YP(1) > 0 .
					for( int jw = 0; jw <= n; ++jw ) {
						solver.mYP[ jw ] = -solver.mYP[ jw ];
					}
					System.arraycopy( solver.mYP, 0, solver.mYpold, 0, n + 1 );
					// FORCE  STEPNF  TO USE THE LINEAR PREDICTOR FOR THE NEXT STEP ONLY.
					solver.mIsStart = true;
				}
			// }
		}

		// *****  END OF MAIN LOOP.  *****

		// LAMBDA HAS NOT REACHED 1 IN 1000 STEPS.
		solver.mArcLength = solver.mS;
		return ERR_MAXIMUM_STEPS_NUMBER_EXCEEDED;

	} /* fixpnf */

	/**
	 * Computes initial points for paths.
	 *
	 * ON INPUT:
	 *
	 * N is the number of (complex) variables.
	 *
	 * ICOUNT is a counter used to increment each
	 *   variable around the unit circle so that every
	 *   combination of start values is chosen.  ICOUNT  is
	 *   initialized in  POLYP.
	 *
	 * IDEG(J)  is the degree of the J-th equation.
	 *
	 * R(I,J)  is a (complex) array generated by subroutine  INITP.
	 *   R(1,J), and R(2,J) are the real and imaginary parts, respectively.
	 *
	 * ON OUTPUT:
	 *
	 * @param x X(I,J)  is initialized to the start values for the current path,
	 *   with X(1,J) and X(2,J) the real and imaginary parts of the
	 *   J-th variable, respectively.
	 */
	private static void strptp(final PolsysSolver solver, final double[] x, int xOff) {

		final int n = solver.mVarAndEqNumber;

		final double[] r = solver.mR;
		final int[] icount = solver.mCounts;
		final int[] ideg = solver.mDegrees;

		// Local variables
		final double[] xxxx = solver.mXX;// new double[ 2 ];

		for( int j = 0; j < n; ++j ) {
			if( icount[ j ] < ideg[ j ] ) {
				++icount[ j ];
				break;
			}
			icount[ j ] = 1;
		}
		// final double twopi = Math.atan( 1. ) * 8.;
		for( int j = 0; j < n; ++j, xOff += 2 ) {
			final double angle = (2. * Math.PI) / ideg[ j ] * icount[ j ];
			xxxx[ 0 ] = Math.cos( angle );
			xxxx[ 1 ] = Math.sin( angle );
			mulp( xxxx, 0, r, j << 1, x, xOff );
		}
	} /* strptp */

	/**
	 * 00  If the problem is to be solved without
	 *       calling POLSYS' scaling routine, SCLGNP, and
	 *       without using the projective transformation.
	 */
	public static final int FLAG_NO_SCALING_NO_PROJECTIVE = 0;

	/** 01  If scaling but no projective transformation is to be used. */
	public static final int FLAG_SCALING = 1;

	/** 10  If no scaling but projective transformation is to be used. */
	public static final int FLAG_PROJECTIVE = 10;

	/** 11  If both scaling and projective transformation are to be used. */
	public static final int FLAG_SCALING_PROJECTIVE = 11;

	/**
	 * The purpose of polyp is to alias the workspaces "WK" and
	 * "IWK" in POLSYS to the variables "PDG" through "IPAR".
	 * POLYP generates the start points for the paths and calls the
	 * path tracker  POLYNF .
	 *
	 * ON INPUT:
	 *
	 * N,NUMT,COEF,KDEG,IFLG1,IFLG2,EPSBIG,EPSSML,NUMRR,NN,
	 *   MMAXT,TTOTDG  are as described in POLSYS.
	 *
	 * TOTDG  is the total degree of the system.
	 *
	 * PDG,QDG,R,FACV,CL,Y,YP,YOLD,YPOLD,QR,ALPHA,TZ,W,WP,Z0,
	 *   Z1,SSPAR,PAR,IDEG,ICOUNT,PIVOT,IPAR  are variables.
	 *
	 * ON OUTPUT:
	 *
	 *   LAMBDA,ROOTS,ARCLEN,NFE  are as described in POLSYS.
	 */
	private static void polyp(final PolsysSolver solver) {

		final int n = solver.mVarAndEqNumber;
		final int n2 = n << 1;

		// ICOUNT IS A COUNTER USED BY "STRPTP"
		solver.mCounts[ 0 ] = 0;
		for( int i = 1; i < n; ++i ) {
			solver.mCounts[ i ] = 1;
		}

		// PATHS LOOP -- ITERATE THROUGH PATHS
		// final double[] xnp1 = new double[ 2 ];// X(N+1)

		for( int path = 0; path < solver.mTotalDegree; ++path ) {
			/*         GET A START POINT, Y, FOR THE PATH. */
			solver.mY[ 0 ] = 0.;
			strptp( solver, solver.mY, 1 );
			/*         CHECK WHETHER PATH IS TO BE FOLLOWED. */
			int iflag = solver.mFlagsIsPathTracked[ path ];
			if( iflag == FOLLOW_PATH ) {
				solver.mArcRelError = solver.mEpsilonBig;
				solver.mArcAbsError = solver.mEpsilonBig;
				solver.mAnswerRelError = solver.mEpsilonSmall;
				solver.mAnswerAbsError = solver.mEpsilonSmall;
				// final int trace = 0;
				/*         TRACK A HOMOTOPY PATH. */
				for( int i = 1; i <= solver.mTriesNumber1000x; ++i ) {
					iflag = polsys( solver, iflag );//, trace );
					if( (iflag != ERR_EPSILON_TOO_SMALL) && (iflag != ERR_MAXIMUM_STEPS_NUMBER_EXCEEDED) ) {
						break;
					}
				}
				// UNSCALE AND UNTRANSFORM COMPUTED SOLUTION.
				// unrolled otputp( n, solver.mCL, solver.mFACV, solver.mCLX, solver.mY, 1, xnp1 );

				// final double[] temp = new double[ 2 ];
				final double[] xxx = solver.mXX;
				final int temp_off = 0;
				final int xnp1_off = 2;

				// COMPUTE XNP1
				for( int j2 = 0; j2 < n2; j2 += 2 ) {
					mulp( solver.mCL, j2, solver.mY, 1 + j2, solver.mCLX, j2 );
				}
				double sum = solver.mCL[ n2 ];
				for( int j2 = 0; j2 < n2; j2 += 2 ) {
					sum += solver.mCLX[ j2 ];
				}
				xxx[ xnp1_off     ] = sum;
				sum = solver.mCL[ n2 + 1 ];
				for( int j2 = 1; j2 < n2; j2 += 2 ) {
					sum += solver.mCLX[ j2 ];
				}
				xxx[ xnp1_off + 1 ] = sum;
				// UNTRANSFORM VARIABLES
				for( int j2 = 1; j2 <= n2; ) {
					divp( solver.mY, j2, xxx, xnp1_off, xxx, temp_off );
					solver.mY[ j2++ ] = xxx[ temp_off ];
					solver.mY[ j2++ ] = xxx[ temp_off + 1 ];
				}
				// UNSCALE VARIABLES
				for( int j = 0; j < n; ++j ) {
					final double fac = Math.pow( 10., solver.mFACV[ j ] );
					int ix = 1 + j + j;
					if( Math.abs( solver.mY[ ix ] ) / LAGEST_MAGNITUDE * fac < 1. ) {
						solver.mY[ ix ] *= fac;
					}
					++ix;
					if( Math.abs( solver.mY[ ix ] ) / LAGEST_MAGNITUDE * fac < 1. ) {
						solver.mY[ ix ] *= fac;
					}
				}
				solver.mPathLambdas[ path ] = solver.mY[ 0 ];
				int j = (path * (n + 1)) << 1;
				System.arraycopy( solver.mY, 1, solver.mPathRoots, j, n2 );
				j += n2;
				solver.mPathRoots[ j     ] = xxx[ xnp1_off ];
				solver.mPathRoots[ j + 1 ] = xxx[ xnp1_off + 1 ];
				solver.mPathArcLengths[ path ] = solver.mArcLength;
				solver.mPathNumberOfMatrixEvaluations[ path ] = solver.mNumberOfMatrixEvaluations;
				solver.mFlagsIsPathTracked[ path ] = iflag;
			}
		}
	} /* polyp */


	/**
	 * COEFFICIENTS( 1:NN , 1:MAXTERMS )
	 *
	 * COEF(J,K) is the K-th coefficient of the J-th equation for J=1 to N, K=1 to NUMT(J)
	 *
	 * @param n variables and equations number.
	 * @param eq equation number, starts from 0.
	 * @param term term number, starts from 0.
	 * @return array index of ( Eq , Term ), starts from 0.
	 */
	public static int getIndexOfCoefficient(final int n, final int eq, final int term) {
		return eq + term * n;
	}

	/**
	 * COEFFICIENTS( 1:NN , 1:MAXTERMS )
	 *
	 * COEF(J,K) is the K-th coefficient of the J-th equation for J=1 to N, K=1 to NUMT(J)
	 *
	 * @param n variables and equations number.
	 * @param eq equation number, starts from 1.
	 * @param term term number, starts from 1.
	 * @return array index of ( Eq , Term ), starts from 0.
	 */
	public static int getIndexOfCoefficient1(final int n, final int eq, final int term) {
		return eq - 1 + term * n - n;
	}

	/**
	 * DEGREES( 1 : NN , 1 : NN+1 , 1 : MAXTERMS )
	 *
	 * Note that the degrees of variables in each term of each equation
	 * are defined by the following indexing scheme:
	 * <pre>
	 *     KDEG(J,  L,  K)
	 *          ^   ^   ^
	 *          E   V   T
	 *          Q   A   E
	 *          U   R   R
	 *          A   I   M
	 *          T   A
	 *          I   B
	 *          O   L
	 *          N   E
	 * </pre>
	 * @param n variables and equations number.
	 * @param eq equation number, starts from 0.
	 * @param term term number, starts from 0.
	 * @param var variable number, starts from 0.
	 * @return array index of ( Eq , Var , Term ), starts from 0.
	 */
	public static int getIndexOfDegrees(final int n, final int eq, final int term, final int var) {
		return eq + (var + term * (n + 1)) * n;
	}

	/**
	 * DEGREES( 1 : NN , 1 : NN+1 , 1 : MAXTERMS )
	 *
	 * Note that the degrees of variables in each term of each equation
	 * are defined by the following indexing scheme:
	 * <pre>
	 *     KDEG(J,  L,  K)
	 *          ^   ^   ^
	 *          E   V   T
	 *          Q   A   E
	 *          U   R   R
	 *          A   I   M
	 *          T   A
	 *          I   B
	 *          O   L
	 *          N   E
	 * </pre>
	 * @param n variables and equations number.
	 * @param eq equation number, starts from 1.
	 * @param term term number, starts from 1.
	 * @param var variable number, starts from 1.
	 * @return array index of ( Eq , Var , Term ), starts from 0.
	 */
	public static int getIndexOfDegrees1(final int n, final int eq, final int term, final int var) {
		return eq - 1 + (var - 1 + (term - 1) * (n + 1)) * n;
	}

	/**
	 * Returns index to use with "numt" array.
	 *
	 * @param n an equation number, starts from 1.
	 * @return an index in number of terms array.
	 */
	public static int getIndexOfEq1(final int n) {
		return n - 1;
	}

	/**
	 * POLSYS state.
	 *
	 * POLSYS finds all (complex) solutions to a system
	 * F(X)=0 of N polynomial equations in N unknowns
	 * with real coefficients. If IFLG=10 or IFLG=11, POLSYS
	 * returns the solutions at infinity also.
	 *
	 * The system F(X)=0 is described via the coefficients,
	 * "COEF", and the parameters "N, NUMT, KDEG", as follows.
	 * <pre>
	 *       NUMT(J)
	 * F(J) = SUM  COEF(J,K) * X(1)^KDEG(J,1,K) * ... * X(N)^KDEG(J,N,K)
	 *        K=1
	 *
	 * For J=1, ..., N.
	 * </pre>
	 * POLSYS has two main run options:  Automatic scaling and
	 * the projective transformation.  These are evoked via the
	 * flag "IFLG1", as described below.  The other input
	 * parameters are the same whether one or both of these options
	 * are specified, and the output is always returned unscaled
	 * and untransformed.
	 *
	 * If automatic scaling is specified, then the input
	 * coefficients are modified by subroutine  SCLGNP . The problem
	 * is solved with the scaled coefficients and scaled variables.
	 * the coefficients are returned scaled.
	 *
	 * If the projective transformation is specified, then
	 * essentially the system is reformulated in homogeneous
	 * coordinates, Z(1), ..., Z(n+1), and solved in complex
	 * projective space.  the resulting solutions are
	 * untransformed via
	 * <pre>
	 * X(J) = Z(J)/Z(N+1)   J=1, ..., N.
	 *
	 * on return,
	 *
	 * ROOTS(1,J,M) = real part of x(j) for the M-th path,
	 *
	 * ROOTS(2,J,M) = imaginary part of x(j) for the M-th path,
	 *
	 * For J=1, ..., N, and
	 *
	 * ROOTS(1,N+1,M) = real part of Z(N+1) for the M-th path,
	 *
	 * ROOTS(2,N+1,M) = imaginary part of Z(N+1) for the M-th path.
	 * </pre>
	 * If roots(*,N+1,M) is small, then the associated solution
	 * should be regarded as being "near infinity".  note that,
	 * when the projective transformation has been specified, the
	 * roots values have been untransformed -- that is, divided
	 * through by Z(N+1) -- unless such division would have caused
	 * overflow.  in this latter case, the affected components of
	 * roots are set to the largest floating point number (machine
	 * infinity).
	 *
	 * The code can be modified easily to solve systems with complex
	 * coefficients,  COEF .  Only the subroutines  INITP  and  FFUNP
	 * need be changed.
	 *
	 * The FORTRAN complex declaration is not used in POLSYS.
	 * Complex variables are represented by real arrays with first
	 * index dimensioned 2 and complex operations are evoked by
	 * subroutine calls.
	 *
	 * The total number of paths that will the tracked (if
	 * IFLG2(M) = -2 for all M) is equal to the "total degree" of the
	 * system, TOTDG.
	 *  TOTDG is equal to the products of the
	 * degrees of all the equations in the system.
	 *  The degree of an equation is the maximum of the degrees of its terms.
	 *  The degree of a term is the sum of the degrees of the variables.
	 * Thus, TOTDG = IDEG(1) * ... * IDEG(N)
	 *  where IDEG(J) = MAX {JDEG(J,K) | K=1,...,NUMT(J)}
	 *  where JDEG(J,K) = KDEG(J,1,K) + ... + KDEG(J,N,K).
	 *
	 * IFLG1  determines whether the system is to be automatically
	 * scaled by  POLSYS  and whether the projective transformation
	 * of the system is to be automatically evoked by POLSYS.  See
	 * "ON INPUT" below.
	 *
	 * IFLG2, EPSBIG, EPSSML, and  SSPAR  tell the path tracker
	 * POLYNF  which paths to track and set parameters for the path
	 * tracker.
	 *
	 * NUMRR  tells  POLSYS  How many multiples of 1000 steps to try
	 * before abandoning a path.
	 *
	 * NN, MMAXT, TTOTDG, LENWK, LENIWK  give the dimensions of arrays.
	 *
	 * The output consists of  IFLG1, and of  LAMBDA, ROOTS, ARCLEN, and
	 * NFE  for each path.  IFLG1  returns input data error information.
	 * Roots  gives the solutions themselves, while  LAMBDA, ARCLEN,
	 * and  NFE  give information about the associated paths.
	 */
	public static final class PolsysSolver {
		// polsys input values

		/** N  is the number of equations and variables.
		 *  NN  is the declared dimension of  NUMT  and of the
		 *  first index of  COEF  and  KDEG .  the second index of
		 *  KDEG  and  ROOTS  is dimensioned NN+1.  NN  must
		 *  be greater than or equal to n. */
		private final int mVarAndEqNumber;

		/** NUMT(1:NN)  IS AN INTEGER ARRAY.
		 *  NUMT(J)  IS THE NUMBER OF TERMS IN THE JTH EQUATION FOR J=1 TO N. */
		private final int[] mNumberOfTerms;

		/** COEF(1:NN,1:MMAXT)  IS A REAL ARRAY.  COEF(J,K)  IS
		 *  THE K-TH COEFFICIENT OF THE J-TH EQUATION FOR J=1 TO N,
		 *  K=1 TO NUMT(J). */
		private final double[] mCoefficients;

		/** KDEG(1:NN,1:NN+1,1:MMAXT)  is an integer array.
		 *  KDEG(J,L,K)  is the degree of the L-th variable int the K-th
		 *  term of the J-th equation for  J=1 to N, L=1 to N, K=1 to NUMT(J). */
		private final int[] mVarDegrees;

		/* MMAXT  is the declared dimension of the second index of
		 *  COEF  and the third index of  KDEG.  MMAXT  must be
		 *  greater than or equal to the maximum number of
		 *  terms in each equation.  in other words,
		 *  MMAXT >= MAX {NUMT(J) | J=1, ..., N} . */
		// private final int mMaxTerms;

		/** TTOTDG  is the declared dimension of  IFLG2 , LAMBDA , ARCLEN ,
		 *  NFE , and of the third index of  ROOTS.  TTOTDG
		 *  must be greater than or equal to  TOTDG, the total
		 *  degree of the system.
		 *
		 *  (max vars in a term in eq1) * (max vars in a term in eq2)
		 *   * (max vars in a term in eq3) * (max vars in a term in eq4)
		 */
		private final int mTotalDegree;

		/* IFLG1 =
		 *  00  IF THE PROBLEM IS TO BE SOLVED WITHOUT
		 *     CALLING POLSYS' SCALING ROUTINE, SCLGNP, AND
		 *     WITHOUT USING THE PROJECTIVE TRANSFORMTION.
		 *  01  IF SCALING BUT NO PROJECTIVE TRANSFORMATION IS TO BE USED.
		 *  10  IF NO SCALING BUT PROJECTIVE TRANSFORMATION IS TO BE USED.
		 *  11  IF BOTH SCALING AND PROJECTIVE TRANSFORMATION ARE TO BE USED. */
		// private final int mFlagSolvingWay;

		/** IFLG2(1:TTOTDG)  IS AN INTEGER ARRAY.  IF IFLG2(M) = -2, THEN THE
		 *  M-TH PATH IS TRACKED.  OTHERWISE THE M-TH PATH IS SKIPPED.
		 *  THUS, TO FIND ALL SOLUTIONS SET IFLG2(M) = -2 FOR M=1,...,TOTDG.
		 *  SELECTED PATHS CAN BE RERUN BY SETTING IFLG2(M)=-2 FOR
		 *  THE PATHS TO BE RERUN AND IFLG(M) != -2 FOR THE OTHERS.
		 *  -2 is JHomePack.FOLLOW_PATH  */
		private final int[] mFlagsIsPathTracked;

		/** EPSBIG  is the local error tolerance allowed the path tracker along
		 *   the path.  ARCRE and ARCAE (in  POLYNF ) are set to  EPSBIG. */
		private final double mEpsilonBig;

		/** EPSSML  is the accuracy desired for the final solution.  ANSRE and
		 *   ANSAE (in  POLYNF ) are set to  EPSSML. */
		private final double mEpsilonSmall;

		/** NUMRR  is the number of multiples of 1000 steps that will be tried
		 *   before abandoning a path. */
		private final int mTriesNumber1000x;

		/** are the real and imaginary parts of the powers of P used to define G. */
		private final double[] mPDG;

		/** the real and imaginary parts of the powers of Q used to define G. */
		private final double[] mQDG;
		/** R(I,J)  is a (complex) array generated by subroutine  INITP.
		 *  R(1,J), and R(2,J) are the real and imaginary parts, respectively. */
		private final double[] mR;
		/** the variable scale factor for the I-th variable, for
		 *   I=1 to N. */
		private final double[] mFACV;
		/** is used to define the projective transformation.
		 * If the projective transformation is not specified, then
		 * CL contains dummy values. */
		private final double[] mCL;
		/** Y(1:N+1) = point (LAMBDA(S), X(S)) on zero curve of homotopy map. */
		private final double[] mY;
		/** YP(1:N+1) = unit tangent vector to the zero curve
		 *  of the homotopy map at  Y . */
		private final double[] mYP;

		/** * YOLD(1:N+1) = a point different from  Y  on the zero curve. */
		private final double[] mYold;
		/** YPOLD(1:N+1) = unit tangent vector to the zero curve
		 *  of the homotopy map at  YOLD . */
		private final double[] mYpold;
		/** the work array used for the QR factorization. */
		private final double[] mQR;
		/** the work array used for the QR factorization. */
		private final double[] mAlpha;
		/** TZ = the Newton step = -(pseudo inverse of  (D RHO(A,Y(S))/D LAMBDA ,
		 * D RHO(A,Y(S))/DX)) * RHO(A,Y(S)) . */
		private final double[] mTZ;
		/** the work array used for the QR factorization (in the
		 *  Newton step calculation) and the interpolation. */
		private final double[] mW;
		/** the work array used for the QR factorization (in the
		 *  Newton step calculation) and the interpolation. */
		private final double[] mWP;
		/** the work array used for the estimation of the
		 *  next step size  H . */
		private final double[] mZ0;
		/** the work array used for the estimation of the
		 *  next step size  H . */
		private final double[] mZ1;

		/* SSPAR(1:8) = (LIDEAL, RIDEAL, DIDEAL, HMIN, HMAX, BMIN, BMAX, P)
		 * Is a vector of parameters used for the optimal step size estimation. */
		/** SSPAR(1:8) = (LIDEAL, RIDEAL, DIDEAL, HMIN, HMAX, BMIN, BMAX, P)  is
		 *    a vector of parameters used for the optimal step size estimation.
		 *    If  SSPAR(J) <= 0.0  on input, it is reset to a default value
		 *    by  POLYNF .  otherwise the input value of  SSPAR(J)  is used.
		 *    See the comments in  POLYNF  and in  STEPNF  for more information
		 *    about these constants. */
		final double[] mInStepSizeEstimationParameters;

		/**
		 * Ideal contraction factor:  ||Z[2] - Z[1]|| / ||Z[1] - Z[0]||
		 * Default value: 0.5
		 */
		private double mStepSizeParameterLIdeal;

		/**
		 * Ideal residual factor:  ||RHO(A, Z[1])|| / ||RHO(A, Z[0])||
		 * Default value: 0.01
		 */
		private double mStepSizeParameterRIdeal;

		/** Ideal distance factor:  ||Z[1] - Y|| / ||Z[0] - Y||
		 * Default value: 0.5 */
		private double mStepSizeParameterDIdeal;

		/**
		 * Minimum step size  HMIN .
		 * Default value: MACHINE_PRECISION * (sqrt( n + 1) + 4.)
		 */
		private double mStepSizeParameterHmin;

		/**
		 * Maximum step size  HMAX .
		 * Default value: 1.0
		 */
		private double mStepSizeParameterHmax;

		/**
		 * Minimum step size reduction factor  BMIN .
		 * Default value: 0.1
		 */
		private double mStepSizeParameterBmin;

		/**
		 * Maximum step size expansion factor  BMAX .
		 * Default value: 3.0
		 */
		private double mStepSizeParameterBmax;

		/**
		 * Assumed operating order  P .
		 * Default value: 2.0
		 */
		private double mStepSizeParameterP;

		//--- end sspar

		// private final double[] par;

		// par
		/** the homotopy */
		private final double[] mRHO;// h;

		/** contains the partial derivatives of RHO with respect to X. */
		private final double[] mDRHOX;// drhox;// dhx;
		/** contains the partial derivatives of RHO with respect to LAMBDA. */
		private final double[] mDRHOL;// drhol;// dht;
		/** workspace array */
		private final double[] mXDGM1;// xdgm1;
		/** workspace array */
		private final double[] mXDG;

		/** the real and imaginary parts of the J-th start equation. */
		private final double[] mG;

		/** the real and imaginary parts of the partial derivatives of the
		 * J-th start equation with respect to the J-th independent variable.*/
		private final double[] mDG;
		/** the workspace array */
		private final double[] mPXDGM1;
		/** the workspace array */
		private final double[] mPXDG;
		/** F(1,J), F(2,J)  are the real and imaginary parts respectively of
		 * the J-th equation. */
		private final double[] mF;
		/** DF(1,J,K), DF(2,J,K)  are the real and imaginary parts respectively
		 *  of the K-th partial derivative of the J-th equation. */
		private final double[] mDF;
		/** workspace variable. */
		private final double[] mXX;
		/** workspace variable. */
		private final double[] mTRM;
		/** workspace variable. */
		private final double[] mDTRM;
		/** workspace variable */
		private final double[] mCLX;
		/** workspace variable */
		private final double[] mDXNP1;

		/** the degree of the J-th equation */
		private final int[] mDegrees;
		/** ICOUNT is a counter used to increment each
		 *  variable around the unit circle so that every
		 *  combination of start values is chosen.  ICOUNT  is
		 *  initialized in  POLYP. */
		private final int[] mCounts;
		/** the work array used for the QR factorization. */
		private final int[] mPivots;
		// private final int[] ipar;

		//******** output ********
		// N, NUMT, COEF, KDEG, NN, MMAXT, TTOTDG, LENWK, and  LENIWK
		//   are unchanged.
		//--------
		/** LAMBDA(M)  is the final lambda value for the M-th path, M = 1, ...,
		 *   TOTDG, where LAMBDA is the continuation parameter. */
		private final double[] mPathLambdas;

		/** ROOTS(1,J,M), ROOTS(2,J,M)  are the Real and Imaginary parts
		 *   of the Jth variable respectively, for j = 1, ..., N, for
		 *   the M-th path, for M = 1, ..., TOTDG.  If  iflg1 = 10 or 11, then
		 *   roots( 1, N+1, M )  and  roots( 2, N+1, M )  are the Real and
		 *   Imaginary parts respectively of the projective
		 *   coordinate of the solution.
		 */
		private final double[] mPathRoots;

		/** ARCLEN(M)  is the ARC length of the M-th path for M = 1, ..., TOTDG. */
		private final double[] mPathArcLengths;

		/** NFE(M)  is the number of jacobian matrix evaluations required to
		 * track the M-th path for M = 1, ..., TOTDG. */
		private final int[] mPathNumberOfMatrixEvaluations;

		//--------  state  --------
		/** A flag used to transfer status information between routines */
		private int mFlag;

		/** NFE = number of Jacobian matrix evaluations */
		private int mNumberOfMatrixEvaluations;

		/** the length of the path followed. */
		private double mArcLength;

		//-------- {@link #fixpnf} state --------

		/** Limit number of iterations. */
		private int mIterationNumberLimitation;

		/** The relative errors, acre */
		private double mArcRelError;

		/** The absolute errors, arcae */
		private double mArcAbsError;

		/**
		 * The relative error value used for the answer at LAMBDA = 1.
		 * The accepted answer  Y = (LAMBDA, X)  satisfies
		 */
		private double mAnswerRelError;

		/**
		 * The absolute error value used for the answer at LAMBDA = 1.
		 * The accepted answer  Y = (LAMBDA, X)  satisfies
		 */
		private double mAnswerAbsError;

		//-------- stepnf state --------

		/** START = .TRUE. on first call to  STEPNF , .FALSE. otherwise. */
		private boolean mIsStart;

		/**
		 * CRASH
		 *    = .FALSE. on a normal return.
		 *
		 *    = .TRUE. if the step size  H  was too small.  H  has been
		 *      increased to an acceptable value, with which  STEPNF  may be
		 *      called again.
		 *
		 *    = .TRUE. if  RELERR  and/or  ABSERR  were too small.  They have
		 *      been increased to acceptable values, with which  STEPNF  may
		 *      be called again.
		 */
		private boolean mIsCrash;

		/** RELERR relative error value. */
		private double mStepRelError;

		/** ABSERR absolute error value. */
		private double mStepAbsError;

		/**
		 * H = upper limit on length of step that will be attempted.  H  must be
		 *    set to a positive number on the first call to  STEPNF .
		 *    thereafter  STEPNF  calculates an optimal value for  H , and  H
		 *    should not be modified by the user.
		 */
		private double mH;

		/** HOLD = ||Y - YOLD||; should not be modified by the user. */
		private double mHold;

		/** S = (APPROXIMATE) arc length along the homotopy zero curve up to
		 *    Y(S) = (LAMBDA(S), X(S)). */
		private double mS;

		//-------- {@link #tangnf} state --------
		private double mTangentRhoLength;
		/** Y(1:N+1) = current point (LAMBDA(S), X(S)). */
		private double[] mTangentY;
		/** YP(1:N+1) = DY/DS = unit tangent vector to integral curve of
		 * D(HOMOTOPY MAP)/DS = 0  at  Y(S) = (LAMBDA(S), X(S)) . */
		private double[] mTangentYp;
		/** YPOLD(1:N+1) = unit tangent vector at previous point on the zero curve of the homotopy map. */
		private double[] mTangentYpold;
		// private int mNfe;
		// private int mN;
		//private int mFlag;

		//-------- {@link #root} state --------

		/* Iteration count */
		// private int mRootIterationCount;
		/* Root count */
		// private int mRootCount;
		// private double mRootA;
		/* F(a) */
		// private double mRootFa;
		/* F(b) */
		// private double mRootFb;
		/* F(c) */
		// private double mRootFc;
		/* Fx = max( abs( Fb ), abs( Fc ) ) */
		// private double mRootFx;
		/* Relative error */
		// private double mRootRelError;
		/* Absolute error */
		// private double mRootAbsError;
		/* abs( a - b ) */
		// private double mRootAbsAB;
		/* Solution */
		// private double mRootSout;
		/* A point */
		// private double mRootSa;
		/* B point */
		// private double mRootSb;

		/**
		 * Create a new solver state.
		 *
		 * The system F(X)=0 is described via the coefficients,
		 * "coefficients", and the parameters "n, numberOfTerms, varDegrees", as follows:
		 * </pre>
		 *  numberOfTerms(J)
		 * F(J) = SUM  coefficients(J,K) * X(1)^varDegrees(J,1,K) * ... * X(N)^varDegrees(J,N,K)
		 *        K=1
		 *
		 * For J = 0, ..., N-1.
		 * </pre>
		 * @param n the number of equations and variables.
		 *
		 * @param numberOfTerms the number of terms int the J-th equation for J=0 to N-1.
		 *
		 * @param coefficients coefficients(J,K) is
		 *    the K-th coefficient of the J-th equation for J=0 to N-1, K=0 to numberOfTerms(J).
		 *
		 * @param varDegrees varDegrees(J,L,K) is
		 *    the degree of the L-th variable int the K-th
		 *    term of the J-th equation for  J=0 to N-1, L=0 to N-1, K=0 to numberOfTerms(J).
		 *
		 * @param epsilonBig is the local error tolerance allowed the path tracker along
		 *    the path.  ARCRE and ARCAE (in  POLYNF ) are set to  "epsilonBig".
		 *
		 * @param epsilonSmall is the accuracy desired for the final solution.
		 *    ANSRE and ANSAE (in  POLYNF ) are set to  "epsilonSmall".
		 *
		 * @param triesNumber1000x is the number of multiples of 1000 steps that will be tried
		 *    before abandoning a path.
		 *
		 * @param flagSolvingWay
		 *    00  if the problem is to be solved without
		 *           calling POLSYS' scaling routine, SCLGNP, and
		 *           without using the projective transformation.
		 *    01  if scaling but no projective transformation is to be used.
		 *    10  if no scaling but projective transformation is to be used.
		 *    11  if both scaling and projective transformation are to be used.
		 *
		 * @param stepSizeEstimationParameters
		 *    (LIDEAL, RIDEAL, DIDEAL, HMIN, HMAX, BMIN, BMAX, P)
		 *    is a vector of parameters used for the optimal step size estimation.
		 *    If SSPAR(J) <= 0.0 on input, it is reset to a default value by POLYNF .
		 *    otherwise the input value of SSPAR(J) is used.
		 *    If {@code null}, default values are used.
		 */
		private PolsysSolver(final int n, final int[] numberOfTerms,
					final double[] coefficients, final int[] varDegrees,
					final double epsilonBig, final double epsilonSmall,
					final int triesNumber1000x,
					int flagSolvingWay, final double[] stepSizeEstimationParameters
				)
		{
			mVarAndEqNumber = n;
			mNumberOfTerms = numberOfTerms;
			mCoefficients = coefficients;
			mVarDegrees = varDegrees;

			// compute max terms
			int max_terms = 0;
			for( int j = 0; j < n; ++j ) {
				if( max_terms < numberOfTerms[ j ] ) {
					max_terms = numberOfTerms[ j ];
				}
			}
			// mMaxTerms = max_terms;

			final int n2 = n + n;
			final int np1 = n + 1;
			// compute max degree
			int total_degree = 1;
			this.mDegrees = new int[ n ];

			for( int eq = 0; eq < n; ++eq ) {
				int eq_degree = 0;
				final int eq_terms = numberOfTerms[ eq ];
				for( int term = 0; term < eq_terms; ++term ) {
					int degree = 0;
					for( int var = 0; var < n; ++var ) {
						degree += varDegrees[ eq + (var + term * np1) * n ];// [eq:n:n+1]
					}
					if( degree > eq_degree ) {
						eq_degree = degree;
					}
				}
				this.mDegrees[ eq ] = eq_degree;
				total_degree *= eq_degree;
			}
			mTotalDegree = total_degree;

			for( int eq = 0; eq < n; ++eq ) {
				final int eq_terms = this.mNumberOfTerms[ eq ];
				for( int term = 0; term < eq_terms; ++term ) {
					int degree = 0;
					for( int var = 0; var < n; ++var ) {
						degree += this.mVarDegrees[ eq + (var + term * np1) * n ];
					}
					this.mVarDegrees[ eq + (n + term * np1) * n ] = this.mDegrees[ eq ] - degree;
				}
			}

			if( flagSolvingWay != FLAG_NO_SCALING_NO_PROJECTIVE
					&& flagSolvingWay != FLAG_SCALING
					&& flagSolvingWay != FLAG_PROJECTIVE
					&& flagSolvingWay != FLAG_SCALING_PROJECTIVE )
			{
				flagSolvingWay = FLAG_NO_SCALING_NO_PROJECTIVE;
			}
			// mFlagSolvingWay = flagSolvingWay;

			final int[] flags_is_path_tracked = new int[ total_degree ];
			for( int i = 0; i < flags_is_path_tracked.length; i++ ) {
				flags_is_path_tracked[ i ] = FOLLOW_PATH;
			}
			mFlagsIsPathTracked = flags_is_path_tracked;

			mEpsilonBig = epsilonBig;
			mEpsilonSmall = epsilonSmall;
			mTriesNumber1000x = Math.max( 1, triesNumber1000x );

			if( (null != stepSizeEstimationParameters) && (8 == stepSizeEstimationParameters.length) ) {
				mInStepSizeEstimationParameters = stepSizeEstimationParameters;
			} else {// defaults
				mInStepSizeEstimationParameters = new double[] { 0, 0, 0, 0, 0, 0, 0, 0 };
			}

			//final int lenwk = 21 + 61 * n + 10 * n * n + 7 * n * max_terms + 4 * n * n * max_terms;
			//mWk = new double[ lenwk ];
			//final int leniwk = 43 + 7 * n + n * (n + 1) * max_terms;
			//mIwk = new int[ leniwk ];
			mPathLambdas = new double[ total_degree ];
			mPathRoots = new double[ (np1 * total_degree) << 1 ];
			mPathArcLengths = new double[ total_degree ];
			mPathNumberOfMatrixEvaluations = new int[ total_degree ];

			// VARIABLES THAT ARE PASSED IN ARRAY WK: (LENGTHS ARE IN THE
			// INTEGER ARRAY LWK.)
			//
			//    VARIABLE NAME       LENGTH
			//
			//    1   PDG               N2
			//    2   QDG               N2
			//    3   R                 N2
			//    4   FACV              N
			//    5   CL                2*(N+1)
			//    6   Y                 N2+1
			//    7   YP                N2+1
			//    8   YOLD              N2+1
			//    9   YPOLD             N2+1
			//   10   QR                N2*(N2+2)
			//   11   ALPHA             N2
			//   12   TZ                N2+1
			//   13   W                 N2+1
			//   14   WP                N2+1
			//   15   Z0                N2+1
			//   16   Z1                N2+1
			//   17   SSPAR             8
			//   18   PAR               2 + 28*N + 6*N^2 + 7*N*MMAXT + 4*N^2*MMAXT
			mPDG   = new double[ n2 ];
			mQDG   = new double[ n2 ];
			mR     = new double[ n2 ];
			mFACV  = new double[ n ];
			mCL    = new double[ np1 << 1 ];
			mY     = new double[ n2 + 1 ];
			mYP    = new double[ n2 + 1 ];
			mYold  = new double[ n2 + 1 ];
			mYpold = new double[ n2 + 1 ];
			mQR    = new double[ n2 * (n2 + 2) ];
			mAlpha = new double[ n2 ];
			mTZ    = new double[ n2 + 1 ];
			mW     = new double[ n2 + 1 ];
			mWP    = new double[ n2 + 1 ];
			mZ0    = new double[ n2 + 1 ];
			mZ1    = new double[ n2 + 1 ];
			// sspar = new double[ 8 ];

			//     PAR INDEX     VARIABLE NAME       LENGTH
			//    ----------     -------------    -----------------
			//          1              PDG               2*N
			//          2               CL               2*(N+1)
			//          3             COEF               N*MMAXT
			//          4                H               N2
			//          5              DHX               N2*N2
			//          6              DHT               N2
			//          7            XDGM1               2*N
			//          8              XDG               2*N
			//          9              G                 2*N
			//         10             DG                 2*N
			//         11           PXDGM1               2*N
			//         12             PXDG               2*N
			//         13               F                2*N
			//         14              DF                2*N*(N+1)
			//         15               XX               2*N*(N+1)*MMAXT
			//         16              TRM               2*N*MMAXT
			//         17             DTRM               2*N*(N+1)*MMAXT
			//         18              CLX               2*N
			//         19            DXNP1               2*N
			// par = new double[ 2 + 28 * n + 6 * n * n + 7 * n * max_terms + 4 * n * n * max_terms ];
			// par is replaced by:
			mRHO = new double[ n2 ];// h
			mDRHOX = new double[ n2 * n2 ];// dhx
			mDRHOL = new double[ n2 ];// dht
			mXDGM1 = new double[ n2 ];
			mXDG = new double[ n2 ];
			mG = new double[ n2 ];
			mDG = new double[ n2 ];
			mPXDGM1 = new double[ n2 ];
			mPXDG = new double[ n2 ];
			mF = new double[ n2 ];
			mDF = new double[ n2 * np1 ];
			mXX = new double[ n2 * np1 * max_terms + 6 ];// java: 6 is added for xnp1, temp1 and temp2
			mTRM = new double[ n2 * max_terms ];
			mDTRM = new double[ n2 * np1 * max_terms ];
			mCLX = new double[ n2 ];
			mDXNP1 = new double[ n2 ];

			// VARIABLES THAT ARE PASSED IN ARRAY IWK: (LENGTHS ARE IN THE
			// INTEGER ARRAY LIWK.)
			//
			//    VARIABLE NAME       LENGTH
			//
			//    1   IDEG              N
			//    2   ICOUNT            N
			//    3   PIVOT             N2+1
			//    4   IPAR              42 + 2*N + N*(N+1)*MMAXT
			// ideg = new int[ n ];
			mCounts = new int[ n ];
			mPivots = new int[ n2 + 1 ];
			// ipar = new int[ 42 + n2 + n * (n + 1) * max_terms ];

			// System.arraycopy( mStepSizeEstimationParameters, 0, sspar, 0, 8 );
			mStepSizeParameterLIdeal = mInStepSizeEstimationParameters[ STEP_SIZE_ESTIMATION_LIDEAL ];
			mStepSizeParameterRIdeal = mInStepSizeEstimationParameters[ STEP_SIZE_ESTIMATION_RIDEAL ];
			mStepSizeParameterDIdeal = mInStepSizeEstimationParameters[ STEP_SIZE_ESTIMATION_DIDEAL ];
			mStepSizeParameterHmin = mInStepSizeEstimationParameters[ STEP_SIZE_ESTIMATION_HMIN ];
			mStepSizeParameterHmax = mInStepSizeEstimationParameters[ STEP_SIZE_ESTIMATION_HMAX ];
			mStepSizeParameterBmin = mInStepSizeEstimationParameters[ STEP_SIZE_ESTIMATION_BMIN ];
			mStepSizeParameterBmax = mInStepSizeEstimationParameters[ STEP_SIZE_ESTIMATION_BMAX ];
			mStepSizeParameterP = mInStepSizeEstimationParameters[ STEP_SIZE_ESTIMATION_P ];

			//---- init ----

			if( (flagSolvingWay == FLAG_PROJECTIVE) || (flagSolvingWay == FLAG_NO_SCALING_NO_PROJECTIVE) ) {

				// DON'T SCALE THE COEFFICIENTS.  SET  FACV  EQUAL TO NOMINAL VALUES.

				for( int i = 0; i < n; ++i ) {
					this.mFACV[ i ] = 0.;
				}
			} else {

				// SET UP THE WORKSPACE FOR SUBROUTINE  SCLGNP  AND CALL  SCLGNP  TO
				// SCALE THE COEFFICIENTS.

				// *****************************************************************
				// VARIABLES THAT ARE PASSED IN ARRAY PAR.
				//    VARIABLE NAME   LENGTH        OFFSET
				//    1   CCOEF       N*MMAXT       1
				//    2   ALPHA       4*N^2         1+N*MMAXT
				//    3   BETA        2*N           1+N*MMAXT+4*N^2
				//    4   RWORK       N*(2*N+1)     1+N*MMAXT+4*N^2+2*N
				//    5   XWORK       2*N           1+N*MMAXT+4*N^2+2*N+N*(2*N+1)
				//    6   FACE        N             1+N*MMAXT+4*N^2+4*N+N*(2*N+1)
				//    7   COESCL      N*MMAXT       1+N*MMAXT+4*N^2+5*N+N*(2*N+1)

				// *****************************************************************
				// VARIABLES THAT ARE PASSED IN ARRAY IPAR.
				//    VARIABLE NAME       LENGTH               OFFSET
				//    1   NNUMT             N                  1
				//    2   KKDEG             N*(N+1)*MMAXT      1+N
				// *****************************************************************
				// java: temp arrays
				final int[] nnumt = new int[ n ];
				final int[] kkdeg = new int[ n * np1 * max_terms ];
				final double[] ccoef  = new double[ n * max_terms ];
				final double[] alpha  = new double[ n2 * n2 ];
				final double[] beta   = new double[ n2 ];
				final double[] rwork  = new double[ n * (n2 + 1) ];
				final double[] xwork  = new double[ n2 ];
				final double[] face   = new double[ n ];
				final double[] coescl = new double[ n * max_terms ];

				final boolean err = sclgnp(
						this,

						false,// mode
						0,

						nnumt,
						kkdeg,

						ccoef,
						alpha,
						beta,
						rwork,
						xwork,
						face,
						coescl
					);

				// SET COEF EQUAL TO THE SCALED COEFFICIENTS

				if( ! err ) {
					for( int j = 0; j < n; ++j ) {
						for( int ci = j, pi = j,
								pe = pi + n * this.mNumberOfTerms[ j ];
								pi < pe; pi += n, ci += n )
						{
							// Computing 2nd power
							this.mCoefficients[ ci ] = coescl[ pi ];
						}
					}
				}
			}

			// IF THE PROJECTIVE TRANSFORMATION IS TO BE USED, THEN  CL  IS
			// SET EQUAL TO THE  CCL  VALUES.  OTHERWISE,  CL  IS SET
			// EQUAL TO NOMINAL VALUES.

			if( (flagSolvingWay == FLAG_SCALING) || (flagSolvingWay == FLAG_NO_SCALING_NO_PROJECTIVE) ) {
				for( int j = 0; j < n2; j += 2 ) {
					this.mCL[ j     ] = 0.;
					this.mCL[ j + 1 ] = 0.;
				}
				this.mCL[ n2     ] = 1.;
				this.mCL[ n2 + 1 ] = 0.;
			} else {
				final double[] ccl = new double[ /* 22 */ ]	/* was [2][11] */ {
						-0.03485644332,  0.28554634336,  0.91453454766,  0.35354566613, -0.36568737635,
						 0.45634642477, -0.89089767544,  0.34524523544,  0.13523462465,  0.43534535555,
						-0.34523544445,  0.00734522256, -0.80004678763, -0.009387123644, -0.875432124245,
						 4.5687651e-4,   0.65256352333, -0.12356777452,  0.09986798321548, -0.56753456577,
						 0.29674947394739, 0.93274302173
					};
				for( int j = 0; j < np1; ++j ) {
					final int jj = (j % 11) << 1;
					final int j2 = j << 1;
					this.mCL[ j2     ] = ccl[ jj ];
					this.mCL[ j2 + 1 ] = ccl[ jj + 1 ];
				}
			}

			/* COMPUTE POWERS OF P AND Q, AND R=Q/P */
			final double[] p = new double[/* 20 */]	/* was [2][10] */ {
					 0.12324754231,  0.76253746298,  0.9385783895,  -0.9937589281,  -0.23467908356,
					 0.39383930009,  0.83542556622, -0.10192888288, -0.55763522521, -0.83729899911,
					-0.78348738738, -0.10578234903,  0.03938347346,  0.04825184716, -0.43428734331,
					 0.93836289418, -0.99383729993, -0.40947822291,  0.09383736736,  0.26459172298
				};

			final double[] q = new double[/* 20 */]	/* was [2][10] */ {
					 0.58720452864,  0.01321964722,  0.978841347,   -0.14433009712,  0.39383737289,
					 0.41543223411, -0.03938376373, -0.61253112318,  0.39383737388, -0.26454678861,
					-0.00938376766,  0.34447867861, -0.04837366632,  0.4825273679,   0.93725237347,
					-0.54356527623,  0.39373957747,  0.65573434564, -0.39380038371,  0.98903450052
				};

			for( int j = 0; j < n; ++j ) {
				final int jj = (j % 10) << 1;
				final int deg = this.mDegrees[ j ];
				final int j2 = j << 1;
				powp( deg, p, jj, this.mPDG, j2 );
				powp( deg, q, jj, this.mQDG, j2 );
				divp( q, jj, p, jj, this.mR, j2 );
			}
		}

		/** Solve equations. */
		public final void solve() {
			polyp( this );
		}

		/**
		 * The total number of paths that will the tracked (if IFLG2(M) = -2 for all M)
		 * is equal to the "total degree" of the system, TOTDG.
		 *   TOTDG is equal to the products of the degrees of all the equations in the system.
		 *   The degree of an equation is the maximum of the degrees of its terms.
		 *   The degree of a term is the sum of the degrees of the variables.
		 * Thus, TOTDG = IDEG(1) * ... * IDEG(N)
		 *    where IDEG(J) = MAX {JDEG(J,K) | K=1,...,NUMT(J)}
		 *    where JDEG(J,K) = KDEG(J,1,K) + ... + KDEG(J,N,K).
		 *
		 * @return The total number of paths.
		 */
		public final int getTotalDegree() {
			return mTotalDegree;
		}

		/**
		 * ARCLEN(M)  is the ARC length of the M-th path for M = 1, ..., TOTDG.
		 *
		 * @param path a path index. Starts from 0.
		 * @return the ARC length of the M-th path.
		 */
		public final double getArcLength(final int path) {
			return mPathArcLengths[ path ];
		}

		/**
		 * NFE(M)  is the number of jacobian matrix evaluations required to
		 * track the M-th path for M = 1, ..., TOTDG.
		 *
		 * @param path a path index. Starts from 0.
		 * @return the number of jacobian matrix evaluations.
		 */
		public final int getNumberOfMatrixEvaluations(final int path) {
			return mPathNumberOfMatrixEvaluations[ path ];
		}

		/**
		 * IFLG2(1:TTOTDG)  is an integer array.  If IFLG2(M) = -2, then the
		 *    M-th path is tracked.  otherwise the m-th path is skipped.
		 *    thus, to find all solutions set iflg2(m) = -2 for M=1,...,totdg.
		 *    selected paths can be rerun by setting IFLG2(M) = -2 for
		 *    the paths to be rerun and IFLG2(M) != -2 for the others.
		 *
		 * @param path a path index. Starts from 0.
		 * @return the tracking status.
		 */
		public final int getFlagsIsPathTracked(final int path) {
			return mFlagsIsPathTracked[ path ];
		}

		/**
		 * LAMBDA(M)  is the final lambda value for the M-th path, M = 1, ..., TOTDG.
		 * where LAMBDA is the continuation parameter.
		 *
		 * @param path a path index. Starts from 0.
		 * @return the lambda value.
		 */
		public final double getLambda(final int path) {
			return mPathLambdas[ path ];
		}

		/**
		 * Returns real part of a root.
		 *
		 * @param path a path index. Starts from 0.
		 * @param var a variable index. Starts from 0.
		 * @return real part of a root
		 */
		public final double getRootRe(final int path, final int var) {
			return mPathRoots[ (var + path * mVarAndEqNumber + path) << 1 ];
		}

		/**
		 * Returns image part of a root.
		 *
		 * @param path a path index. Starts from 0.
		 * @param var a variable index. Starts from 0.
		 * @return image part of a root
		 */
		public final double getRootIm(final int path, final int var) {
			return mPathRoots[ ((var + path * mVarAndEqNumber + path) << 1) + 1 ];
		}

		/**
		 * Returns all found complex roots.
		 * 2 dimensional array, [ solution number ][ variable ]
		 * Variable is a complex number, even indexes are real parts,
		 * odd indexes are image parts.
		 *
		 * @return the array with complex roots.
		 */
		public final ArrayList< double[] > getRoots() {
			//ROOTS(1,J,M), ROOTS(2,J,M)  ARE THE REAL AND IMAGINARY PARTS
			//   OF THE JTH VARIABLE RESPECTIVELY, FOR J = 1,...,N, FOR
			//   THE M-TH PATH, FOR M = 1,...,TOTDG.
			final ArrayList< double[] > result = new ArrayList<>();

			// final int root_dimension1 = 2;// re + im
			final int root_dimension2 = mVarAndEqNumber + 1;
			for( int path = 0; path < mTotalDegree; path++ ) {
				if( ERR_NO == mFlagsIsPathTracked[ path ] ) {
					final double[] variables = new double[ mVarAndEqNumber << 1 ];// * root_dimension1
					final int shift = path * root_dimension2;
					for( int var = 0; var < mVarAndEqNumber; ++var ) {// mVarAndEqNumber + 1 is an internal computed value
						int index = (var + shift) << 1;// * root_dimension1
						final double re = mPathRoots[ index ];
						final double im = mPathRoots[ index + 1 ];
						index = var << 1;
						variables[ index     ] = re;
						variables[ index + 1 ] = im;
					}
					result.add( variables );
				}
			}
			return result;
		}

		/** Flag to filter all real solutions */
		public static final int ALL_REALS = 0;

		/** Flag to filter all real and negative solutions */
		public static final int NEGATIVE_REALS = -1;

		/** Flag to filter all real and positive solutions */
		public static final int POSITIVE_REALS = 1;

		/**
		 * Returns all computed real roots.
		 * 2 dimensional array, [ solution number ][ variable ]
		 * Variable is a double value.
		 *
		 * @param flag 0 - any, < 0 - all negative, > 0 - all positive.
		 * @param tolerance a factor to filter results, a positive value.
		 *    a value is real if abs( im ) < tolerance * abs( re )
		 * @return the array with real roots.
		 */
		public final ArrayList< double[] > getRealRoots(final int flag, final double tolerance) {
			//ROOTS(1,J,M), ROOTS(2,J,M)  ARE THE REAL AND IMAGINARY PARTS
			//   OF THE JTH VARIABLE RESPECTIVELY, FOR J = 1,...,N, FOR
			//   THE M-TH PATH, FOR M = 1,...,TOTDG.
			final ArrayList< double[] > result = new ArrayList<>();

			// final int root_dimension1 = 2;// re + im
			final int root_dimension2 = mVarAndEqNumber + 1;
			for( int path = 0; path < mTotalDegree; path++ ) {
				if( ERR_NO == mFlagsIsPathTracked[ path ] ) {
					final int shift = path * root_dimension2;
					boolean is_real = true;
					for( int var = 0; (var < mVarAndEqNumber) && is_real; ++var ) {
						final int index = (var + shift) << 1;// * root_dimension1
						final double re = mPathRoots[ index ];
						final double im = mPathRoots[ index + 1 ];
						is_real &= Math.abs( im ) < tolerance * Math.abs( re );
						is_real &= ((flag >= 0) || (re <= 0.));// ! ((flag < 0) && (re > 0.));
						is_real &= ((flag <= 0) || (re >= 0.));// ! ((flag > 0) && (re < 0.));
					}
					if( is_real ) {
						final double[] variables = new double[ mVarAndEqNumber ];
						for( int var = 0; var < mVarAndEqNumber; ++var ) {
							final int index = (var + shift) << 1;// * root_dimension1
							final double re = mPathRoots[ index ];
							final double im = mPathRoots[ index + 1 ];
							variables[ var ] = Math.copySign( Math.hypot( re, im ), re );
						}// [mVarAndEqNumber] is an internal computed value
						result.add( variables );
					}
				}
			}
			return result;
		}
	}

	/**
	 * Create a new solver state.
	 *
	 * The system F(X)=0 is described via the coefficients,
	 * "coefficients", and the parameters "n, numberOfTerms, varDegrees", as follows:
	 * <pre>
	 *  numberOfTerms(J)
	 * F(J) = SUM  coefficients(J,K) * X(1)^varDegrees(J,1,K) * ... * X(N)^varDegrees(J,N,K)
	 *        K=1
	 *
	 * For J = 0, ..., N-1.
	 * </pre>
	 *
	 * @param n the number of equations and variables.
	 *
	 * @param numberOfTerms the number of terms int the J-th equation for J=0 to N-1.
	 *
	 * @param coefficients coefficients(J,K) is
	 *    the K-th coefficient of the J-th equation for J=0 to N-1, K=0 to numberOfTerms(J).
	 *
	 * @param varDegrees varDegrees(J,L,K) is
	 *    the degree of the L-th variable int the K-th
	 *    term of the J-th equation for  J=0 to N-1, L=0 to N-1, K=0 to numberOfTerms(J).
	 *
	 * @param epsilonBig is the local error tolerance allowed the path tracker along
	 *    the path.  ARCRE and ARCAE (in  POLYNF ) are set to  "epsilonBig".
	 *
	 * @param epsilonSmall is the accuracy desired for the final solution.
	 *    ANSRE and ANSAE (in  POLYNF ) are set to  "epsilonSmall".
	 *
	 * @param triesNumber1000x is the number of multiples of 1000 steps that will be tried
	 *    before abandoning a path.
	 *
	 * @param flagSolvingWay
	 *    00  if the problem is to be solved without
	 *           calling POLSYS' scaling routine, SCLGNP, and
	 *           without using the projective transformation.
	 *           {@link JHomPack#FLAG_NO_SCALING_NO_PROJECTIVE}
	 *    01  if scaling but no projective transformation is to be used.
	 *           {@link JHomPack#FLAG_SCALING}
	 *    10  if no scaling but projective transformation is to be used.
	 *           {@link JHomPack#FLAG_PROJECTIVE}
	 *    11  if both scaling and projective transformation are to be used.
	 *           {@link JHomPack#FLAG_SCALING_PROJECTIVE}
	 *
	 * @param stepSizeEstimationParameters
	 *    (LIDEAL, RIDEAL, DIDEAL, HMIN, HMAX, BMIN, BMAX, P)
	 *    is a vector of parameters used for the optimal step size estimation.
	 *    If SSPAR(J) <= 0.0 on input, it is reset to a default value by POLYNF .
	 *    otherwise the input value of SSPAR(J) is used.
	 *    If {@code null}, default values are used.
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_LIDEAL},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_RIDEAL},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_DIDEAL},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_HMIN},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_HMAX},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_BMIN},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_BMAX},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_P},
	 *
	 * @return the state.
	 */
	public static PolsysSolver newSolver(final int n, final int[] numberOfTerms,
			final double[] coefficients, final int[] varDegrees,
			final double epsilonBig, final double epsilonSmall,
			final int triesNumber1000x,
			final int flagSolvingWay, final double[] stepSizeEstimationParameters)
	{
		return new PolsysSolver( n, numberOfTerms, coefficients, varDegrees,
				epsilonBig, epsilonSmall, triesNumber1000x,
				flagSolvingWay, stepSizeEstimationParameters );
	}

	/** Structure to hold input parameters for equations. */
	public static final class SolverInputParameters {
		private final int mN;
		private final int[] mNumberOfTerms;
		private final double[] mCoefficients;
		private final int[] mVarDegrees;

		/**
		 * The system F(X)=0 is described via the coefficients,
		 * "coefficients", and the parameters "n, numberOfTerms, varDegrees", as follows:
		 * <pre>
		 *  numberOfTerms(J)
		 * F(J) = SUM  coefficients(J,K) * X(1)^varDegrees(J,1,K) * ... * X(N)^varDegrees(J,N,K)
		 *        K=1
		 *
		 * For J = 0, ..., N-1.
		 * </pre>
		 *
		 * @param n the number of equations and variables.
		 *
		 * @param numberOfTerms the number of terms int the J-th equation for J=0 to N-1.
		 *
		 * @param coefficients coefficients(J,K) is
		 *    the K-th coefficient of the J-th equation for J=0 to N-1, K=0 to numberOfTerms(J).
		 *
		 * @param varDegrees varDegrees(J,L,K) is
		 *    the degree of the L-th variable int the K-th
		 *    term of the J-th equation for  J=0 to N-1, L=0 to N-1, K=0 to numberOfTerms(J).
		 */
		public SolverInputParameters(final int n, final int[] numberOfTerms,
							   final double[] coefficients, final int[] varDegrees)
		{
			mN = n;
			mNumberOfTerms = numberOfTerms;
			mCoefficients = coefficients;
			mVarDegrees = varDegrees;
		}
	}

	/**
	 * Create a new solver state.
	 *
	 * {@link JHomPack#newSolver(int, int[], double[], int[], double, double, int, int, double[])}
	 *
	 * @param epsilonBig is the local error tolerance allowed the path tracker along
	 *    the path.  ARCRE and ARCAE (in  POLYNF ) are set to  "epsilonBig".
	 *
	 * @param epsilonSmall is the accuracy desired for the final solution.
	 *    ANSRE and ANSAE (in  POLYNF ) are set to  "epsilonSmall".
	 *
	 * @param triesNumber1000x is the number of multiples of 1000 steps that will be tried
	 *    before abandoning a path.
	 *
	 * @param flagSolvingWay
	 *    00  if the problem is to be solved without
	 *           calling POLSYS' scaling routine, SCLGNP, and
	 *           without using the projective transformation.
	 *           {@link JHomPack#FLAG_NO_SCALING_NO_PROJECTIVE}
	 *    01  if scaling but no projective transformation is to be used.
	 *           {@link JHomPack#FLAG_SCALING}
	 *    10  if no scaling but projective transformation is to be used.
	 *           {@link JHomPack#FLAG_PROJECTIVE}
	 *    11  if both scaling and projective transformation are to be used.
	 *           {@link JHomPack#FLAG_SCALING_PROJECTIVE}
	 *
	 * @param stepSizeEstimationParameters
	 *    (LIDEAL, RIDEAL, DIDEAL, HMIN, HMAX, BMIN, BMAX, P)
	 *    is a vector of parameters used for the optimal step size estimation.
	 *    If SSPAR(J) <= 0.0 on input, it is reset to a default value by POLYNF .
	 *    otherwise the input value of SSPAR(J) is used.
	 *    If {@code null}, default values are used.
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_LIDEAL},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_RIDEAL},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_DIDEAL},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_HMIN},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_HMAX},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_BMIN},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_BMAX},
	 *    {@link JHomPack#STEP_SIZE_ESTIMATION_P},
	 *
	 * @return the state.
	 */
	public static PolsysSolver newSolver(final SolverInputParameters parameters,
			final double epsilonBig, final double epsilonSmall,
			final int triesNumber1000x,
			final int flagSolvingWay, final double[] stepSizeEstimationParameters)
	{
		return new PolsysSolver( parameters.mN, parameters.mNumberOfTerms,
				parameters.mCoefficients, parameters.mVarDegrees,
				epsilonBig, epsilonSmall, triesNumber1000x,
				flagSolvingWay, stepSizeEstimationParameters );
	}
}
